/*
 * I2O SCSI definitions
 * 
 * Copyright (C) 1999   Red Hat Software
 * Copyright (C) 2002-3 Wind River Systems
 * Copyright (C) 2002-3 Intel Corporation
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
 */

/* 
 *  Complications for I2O scsi
 *
 *	o	Each (bus,lun) is a logical device in I2O. We keep a map
 *		table. We spoof failed selection for unmapped units
 *	o	Request sense buffers can come back for free. 
 *	o	Scatter gather is a bit dynamic. We have to investigate at
 *		setup time.
 *	o	Some of our resources are dynamically shared. The i2o core
 *		needs a message reservation protocol to avoid swap v net
 *		deadlocking. We need to back off queue requests.
 *	
 *	In general the firmware wants to help. Where its help isn't performance
 *	useful we just ignore the aid. Its not worth the code in truth.
 *
 *	Fixes:
 *      0.0.3   Jason O'Broin   :       Larger message frame size supported
 *      0.0.2   Jason O'Broin   :       Sense data requested & returned up stack
 *      0.0.1	Steve Ralston	:	Scatter gather now works
 *
 *	To Do
 *		64bit cleanups
 *		Fix the resource management problems.
 *
 * NOTE: 11/22/02
 *
 * (C) Copyright 2002   Wind River Systems
 *
 * This code for the Intel PRO/1000 T IP Storage Adapter is1000 driver
 * was based heavily on the Red Hat i2o_scsi.c module. As such all 
 * term of the GNU General Public License Apply.
 *
 * 04aug03 eab Removed useless field in ISCSI_TASK_DATA; abort and
 *             bus reset replies handled correctly; Added macros
 *             for kernel-variant io locking
 * 31jul03 eab Got rid of useless locks, cleanup
 * 21jul03 eab Up'd scsi host max_id to i2o max: 128; fixed spinlock
 *             to allow for SMP+error reply
 * 16jun03 eab Porting to Red Hat 9 started: version 2.0.7
 * 07may03 eab Compile with new headers, add stubs
 * 07may03 eab First round support for i2o class added
 * 16apr03 eab Changed scsi host max luns to I2O_ISCSI_MAX_LUNS
 * 15apr03 eab Port of is1000-1.1-12 driver to Enterprise AS (is1000-1.2-0)
 * 07mar03 eab Modified release routine to unlock controller, since detect
 *             routine locks controller via call to i2o_find_controller.
 * 22nov02 eab Written, based on the RH 7.2 i2o_core.c file.
 * 
 * To Do: 
 * Check on feasibility of checking/handling signals and
 *     handling a canceled i/o when target is failed
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/types.h>
#include <linux/string.h>
#include <linux/ioport.h>
#include <linux/sched.h>
#include <linux/interrupt.h>
#include <linux/timer.h>
#include <linux/delay.h>
#include <linux/proc_fs.h>
#include <linux/prefetch.h>
#include <asm/dma.h>
#include <asm/system.h>
#include <asm/io.h>
#include <asm/atomic.h>
#include <linux/blk.h>
#include <linux/version.h>
#include <linux/i2o.h>
#include "scsi.h"
#include "hosts.h"
#include "sd.h"

#include "is1000_scsi.h"
#include "is1000.h"

#define I2O_HBA_BUS_EXEC 0x8a


//#define DRIVERDEBUG
//#define SCSI_DEBUG
//#define RECOVERY_DEBUG

#ifdef DRIVERDEBUG
#define dprintk(s, args...) printk(s, ## args)
#else
#define dprintk(s, args...)
#endif

extern struct iscsi_adapter_private iscsi_adapter_priv[];

static int scsi_context;
static Scsi_Host_Template driver_template = I2OSCSI;
static int i2o_scsi_hosts;
static u32 *retry[32];
static struct i2o_controller *retry_ctrl[32];
static struct timer_list retry_timer;
static int retry_ct = 0;

static atomic_t queue_depth;

/*
 *	SG Chain buffer support...
 */

#define SG_MAX_FRAGS		64

/*
 *	FIXME: we should allocate one of these per bus we find as we
 *	locate them not in a lump at boot.
 */
 
typedef struct _chain_buf
{
	u32 sg_flags_cnt[SG_MAX_FRAGS];
	u32 sg_buf[SG_MAX_FRAGS];
} chain_buf;

#define SG_CHAIN_BUF_SZ sizeof(chain_buf)

#define SG_MAX_BUFS		(i2o_num_controllers * I2O_SCSI_CAN_QUEUE)
#define SG_CHAIN_POOL_SZ	(SG_MAX_BUFS * SG_CHAIN_BUF_SZ)

void is1000_timeout_func(unsigned long data);
void is1000_failed_func(unsigned long data);

#if 0
static spinlock_t is1000_cmd_lock = SPIN_LOCK_UNLOCKED;
#define IS1000_CMD_LOCK spin_lock_irqsave(&is1000_cmd_lock,flags)
#define IS1000_CMD_UNLOCK spin_unlock_irqrestore(&is1000_cmd_lock,flags)
#else
#define IS1000_CMD_LOCK
#define IS1000_CMD_UNLOCK

#endif

static int max_sg_len = 0;
static chain_buf *sg_chain_pool = NULL;
static int sg_chain_tag = 0;
static int sg_max_frags = SG_MAX_FRAGS;

static void internal_done(Scsi_Cmnd * SCpnt)
{
        dprintk(KERN_INFO "internal_done()\n");
        //        SCpnt->SCp.Status++;
}

static void useless_timeout_function(unsigned long arg)
{
    
}

static inline void add_completion_timer(Scsi_Cmnd *sc)
{
	int rtn;

	if ((sc->eh_timeout.function == 
	     (void (*)(unsigned long)) (is1000_timeout_func)) ||
	    (sc->eh_timeout.function == 
	     (void (*)(unsigned long)) (is1000_failed_func)))
	{
#ifdef SCSI_DEBUG      
		printk(KERN_INFO "is1000: reply: adding bogus timer\n");      
#endif
		rtn = del_timer(&sc->eh_timeout);
#ifdef SCSI_DEBUG      
		if (rtn == 0)
			printk(KERN_ERR "is1000: return from del_timer is 0!\n");
#endif
		sc->eh_timeout.data = (unsigned long)sc;
		sc->eh_timeout.expires = jiffies + sc->timeout_per_command;
		sc->eh_timeout.function = useless_timeout_function;
		add_timer(&sc->eh_timeout);
	}
#ifdef SCSI_DEBUG
	else
	{
		printk(KERN_INFO "is1000: reply: not touching timer!\n");
	}
#endif
}

/*
 *	Retry congested frames. This actually needs pushing down into
 *	i2o core. We should only bother the OSM with this when we can't
 *	queue and retry the frame. Or perhaps we should call the OSM
 *	and its default handler should be this in the core, and this
 *	call a 2nd "I give up" handler in the OSM ?
 */
 
static void i2o_retry_run(unsigned long f)
{
	int i;
	unsigned long flags;
	
#ifdef RH90
	spin_lock_irqsave(&io_request_lock, flags);
#else
	save_flags(flags);
	cli();
#endif
	for(i=0;i<retry_ct;i++)
                i2o_post_message(retry_ctrl[i], virt_to_bus(retry[i]));
	retry_ct=0;
	
#ifdef RH90
	spin_unlock_irqrestore(&io_request_lock, flags);
#else
	restore_flags(flags);
#endif
}

static void flush_pending(void)
{
	int i;
	unsigned long flags;
	
#ifdef RH90
	spin_lock_irqsave(&io_request_lock, flags);
#else
	save_flags(flags);
	cli();
#endif

	for(i=0;i<retry_ct;i++)
        {
                retry[i][0]&=~0xFFFFFF;
		retry[i][0]|=I2O_CMD_UTIL_NOP<<24;
		i2o_post_message(retry_ctrl[i],virt_to_bus(retry[i]));
	}
	retry_ct=0;
	
#ifdef RH90
	spin_unlock_irqrestore(&io_request_lock, flags);
#else
	restore_flags(flags);
#endif
}


static void i2o_scsi_reply(struct i2o_handler *h, struct i2o_controller *c, struct i2o_message *msg)
{
	Scsi_Cmnd *current_command;
	u32 *m = (u32 *)msg;
	u8 as,ds,st;

        PREFETCH_SPINLOCK;

	if(m[0] & (1<<13))
	{
		printk("IOP fail.\n");
		printk("From %d To %d Cmd %d.\n",
                       (m[1]>>12)&0xFFF,
                       m[1]&0xFFF,
                       m[1]>>24);
		printk("Failure Code %d.\n", m[4]>>24);
		if(m[4]&(1<<16))
			printk("Format error.\n");
		if(m[4]&(1<<17))
			printk("Path error.\n");
		if(m[4]&(1<<18))
			printk("Path State.\n");
		if(m[4]&(1<<18))
			printk("Congestion.\n");
		
		m=(u32 *)bus_to_virt(m[7]);
		printk("Failing message is %p.\n", m);
		
		if((m[4]&(1<<18)) && retry_ct < 32)
		{
			retry_ctrl[retry_ct]=c;
			retry[retry_ct]=m;
			if(!retry_ct++)
			{
				retry_timer.expires=jiffies+1;
				add_timer(&retry_timer);
			}
		}
		else
		{
			/* Create a scsi error for this */
			current_command = (Scsi_Cmnd *)m[3];

			printk("Aborted %ld\n", current_command->serial_number);

                        REACQUIRE_MIDLAYER_LOCK(current_command->host);

			current_command->result = DID_ERROR << 16;
                        add_completion_timer(current_command);
			current_command->scsi_done(current_command);

                        RELEASE_MIDLAYER_LOCK(current_command->host);
			
			/* Now flush the message by making it a NOP */
			m[0]&=0x00FFFFFF;
			m[0]|=(I2O_CMD_UTIL_NOP)<<24;
			i2o_post_message(c,virt_to_bus(m));
		}
		return;
	}
	
	prefetchw(&queue_depth);
		

	/*
	 *	Low byte is device status, next is adapter status,
	 *	(then one byte reserved), then request status.
	 */
#ifdef RH90
	ds=(u8)le32_to_cpu(m[4]); 
	as=(u8)le32_to_cpu(m[4]>>8);
	st=(u8)le32_to_cpu(m[4]>>24);
#else
	ds=(u8)m[4]; 
	as=(u8)(m[4]>>8);
	st=(u8)(m[4]>>24);
#endif
	

        /* look for a reply from a bus reset w/o context */
        if((m[3] == 0xFFFFFFFF) && m[2]&0x80000000)
	{
#ifdef RECOVERY_DEBUG
		printk(KERN_INFO "i2o_scsi_reply: bus reset reply w/o context\n");
#endif
		return;
	}

	current_command = (Scsi_Cmnd *)m[3];
	
	/*
	 *	Is this an abort or reset reply? 
	 */
        /*	if(current_command->SCp.Status == -1)*/
        /*	if( (current_command->SCp.Status == -1) && */
	if((m[1] >> 24) == I2O_CMD_SCSI_ABORT)
	{
#ifdef RECOVERY_DEBUG
		printk(KERN_INFO "         abort reply   m[0]=%08X  m[1]=%08X\n", 
		       m[0], m[1]);
		printk(KERN_INFO "                       m[2]=%08X  m[3]=%08X\n", 
		       m[2], m[3]);
		printk(KERN_INFO "                       m[4]=%08X  m[5]=%08X\n\n", 
		       m[4], m[5]);
#endif          
		if(st)
		{
#ifdef RECOVERY_DEBUG
			printk(KERN_INFO "SCSI abort status: st %x  ds %x  as %x\n", st, ds, as);
#endif
			current_command->result = DID_ERROR << 16;
		}
		else
			current_command->result = DID_ABORT << 16;

		REACQUIRE_MIDLAYER_LOCK(current_command->host);

		/* new foo */
		//current_command->scsi_done(current_command);
		current_command->SCp.Status &= ~ISCSI_CMD_ABORT_SENT;

#ifdef RECOVERY_DEBUG
		printk(KERN_INFO "current_command->SCp.Status 0x%08x\n",
		       (int)current_command->SCp.Status);
#endif
		RELEASE_MIDLAYER_LOCK(current_command->host);

		return;
	} /* end abort reply */

        if((m[1] >> 24) == I2O_HBA_BUS_RESET)
	{
#ifdef RECOVERY_DEBUG
		printk(KERN_INFO "         reset reply   m[0]=%08X  m[1]=%08X\n", 
		       m[0], m[1]);
		printk(KERN_INFO "                       m[2]=%08X  m[3]=%08X\n", 
		       m[2], m[3]);
		printk(KERN_INFO "                       m[4]=%08X  m[5]=%08X\n\n", 
		       m[4], m[5]);
#endif          
		if(st)
                {
#ifdef RECOVERY_DEBUG
			printk(KERN_INFO "SCSI reset status: st %x  ds %x  as %x\n", st, ds, as);
#endif
			current_command->result = DID_ERROR << 16;
                }
		else
			current_command->result = DID_RESET << 16;

		REACQUIRE_MIDLAYER_LOCK(current_command->host);

		/* new foo */
		//current_command->scsi_done(current_command);
		current_command->SCp.Status &= ~ISCSI_CMD_BUS_RESET_SENT;
            
#ifdef RECOVERY_DEBUG
		printk(KERN_INFO "current_command->SCp.Status 0x%08x\n",
		       (int)current_command->SCp.Status);
#endif
		RELEASE_MIDLAYER_LOCK(current_command->host);

		return;
	}

	
#ifdef SCSI_DEBUG
	printk(KERN_INFO "i2o got a scsi reply m[0]=%08X   m[1]=%08X\n", m[0], m[1]);
	printk(KERN_INFO "                     m[2]=%08X   m[3]=%08X\n", m[2], m[3]);
	printk(KERN_INFO "                     m[4]=%08X   m[5]=%08x\n\n", m[4], m[5]);

	printk(KERN_INFO "Completed %ld         %X\n", current_command->serial_number, 
	       (int)current_command);
#endif

        /* add bogus timer for mid layer to delete */
        add_completion_timer(current_command);

        /* Queued depth only on scsi commands, so keep this here */
	atomic_dec(&queue_depth);
	
	if(st)
	{
		/* An error has occurred */
#ifdef SCSI_DEBUG
		printk(KERN_ERR "SCSI error   m[4] %08X    (ds %X)\n", m[4], ds);
#endif
                if ((st == 6) && (ds == 0x22))
		{
			switch (current_command->cmnd[0]) {
			case INQUIRY:
			case REPORT_LUNS:
			case TEST_UNIT_READY:
			case READ_CAPACITY:
			case START_STOP:
			case MODE_SENSE:

#if 1
				if (current_command->SCp.Status & ISCSI_CMD_WAS_ABORTED)
				{
#ifdef SCSI_DEBUG
					printk(KERN_ERR "NOT sending reply: cmd 0x%08x\n",
					       (int)current_command);
#endif
					return;
				}
#endif

#ifdef SCSI_DEBUG
				printk(KERN_ERR "returning cmd w/err _ERROR\n");
#endif
				current_command->result = DID_ERROR << 16;

				break;
			default:
#ifdef SCSI_DEBUG
				printk(KERN_ERR "*Special SCSI error   m[4] %08X  _DID_BUS_BUSY\n", m[4]);
#endif
				current_command->result = DID_BUS_BUSY << 16;
			}
		}
                else if (as == 0x0E) 
			/* SCSI Reset */
			current_command->result = DID_RESET << 16;
		else if (as == 0x0F)
			current_command->result = DID_PARITY << 16;
		else if (ds == 2)
		{
#ifdef SCSI_DEBUG
                        int j;
                        printk(KERN_INFO "is1000: sensebuffer:  %d bytes  ->result 0x%08x\n", m[6] & 0xffff,
                               current_command->result);
                        printk(KERN_INFO "sense data:\n");
                        for (j = 0 ; j < 28 ; j+=4)
				printk(KERN_INFO "%02x %02x %02x %02x\n", 
				       current_command->sense_buffer[j],
				       current_command->sense_buffer[j+1],
				       current_command->sense_buffer[j+2],
				       current_command->sense_buffer[j+3]);
#endif

#ifdef RH90
                        current_command->result = DRIVER_SENSE << 24;
#else                        
                        current_command->result = CHECK_CONDITION << 24;
#endif
		}
		else
			/* We want to fail this command ~fatally~ so return 0xf here
			   so that mid layer will just give up.
			*/
			/*current_command->result = 0xf << 16 | ds;*/
			current_command->result = DID_ERROR << 16 | 0x22;
	}
	else
	{
		current_command->result = DID_OK << 16 | ds;
                if (ds == 2)
#ifdef RH90
			current_command->result |= DRIVER_SENSE << 24;
#else                        
                current_command->result |= CHECK_CONDITION << 24;
#endif

	}
	
        REACQUIRE_MIDLAYER_LOCK(current_command->host);

	current_command->scsi_done(current_command);

        RELEASE_MIDLAYER_LOCK(current_command->host);

	return;
}


struct i2o_handler i2o_scsi_handler=
{
	i2o_scsi_reply,
	NULL,
	NULL,
	NULL,
	"I2O SCSI OSM",
	0,
	I2O_CLASS_BUS_ADAPTER_PORT
};

void i2o_scsi_init0(struct i2o_controller *c, struct Scsi_Host *shpnt)
{
	struct i2o_scsi_host *h =(struct i2o_scsi_host *)shpnt->hostdata;
	int target;
	
	dprintk(KERN_INFO "is1000: i2o_scsi_init0() called.\n");

	if (h == NULL)
        {
	        dprintk(KERN_ERR "is1000: No host data alloc'd in SCSI Host\n");
                return;
        }
	h->controller=c;
	
	for(target = 0 ; target < IOP_MAXIMUM_TARGETS ; target++)
	{
		h->aTidData[target].tid = -1;
		h->aTidData[target].index = -1;
		h->aTidData[target].status = -1;
                memset(h->aTidData[target].name, 0, ISCSI_TARGET_NAME_LENGTH);
		h->aTaskData[target].pTidData = NULL;
		h->aTaskData[target].inUse = 0;
	}
}


/* This routine will now be called as the dyn_lct handler. If the
 * device is already known about, do nothing, otherwise add the
 *  tid to the table, and get the device name from the hba session
 * table. When a target status ioctl is called, the tid data 
 * table will be searched, and if the target name is found, the
 * target index will be filled into the tid data index field,
 * and the pointer to the task data will be filled into the
 * tid data pTidData field.
 */

int i2o_scsi_init(struct i2o_controller *c, struct i2o_device *d, struct Scsi_Host *shpnt)
{
	struct i2o_scsi_host *h =(struct i2o_scsi_host *)shpnt->hostdata;
        int i;
	u16 limit;
        
#ifdef SCSI_DEBUG
        printk(KERN_INFO "is1000: i2o_scsi_init()\n");
#endif
        /* This is now just for error checking - in case tid is already here */
        for (i = 0 ; i < IOP_MAXIMUM_TARGETS ; i++)
	{
		if (h->aTidData[i].tid == d->lct_data.tid)
		{
#ifdef SCSI_DEBUG
			printk(KERN_INFO "is1000: found target id already %d\n", i);
#endif
			return -1;
		}
	}

        /* get target number by enumerating through hosts task array
         * and find first one that's -1 */

        for (i = 0 ; i < IOP_MAXIMUM_TARGETS ; i++)
	{
		if (h->aTidData[i].tid == -1)
		{
			int res;

			h->aTidData[i].tid = d->lct_data.tid;

			h->tagclock[i]=jiffies;

#ifdef SCSI_DEBUG
			printk(KERN_INFO "is1000: aTidData index %d gets device tid %d\n", 
			       i, d->lct_data.tid);
#endif
			/* Get the target name for this device */
			res =  i2o_query_scalar(c, d->lct_data.tid,
						ISCSI_ISM_PER_HBA_SESSION_GROUP_NO,
						ISCSI_ISM_PER_HBA_NAME,
						h->aTidData[i].name,
						ISCSI_TARGET_NAME_LENGTH - 1);

			if (res < 0) {
				printk(KERN_ERR "is1000: Failed to get target name from tid %d\n",
				       d->lct_data.tid);
			}
#ifdef SCSI_DEBUG
			printk(KERN_ERR "is1000: Adding target name <%s> from tid %d\n",
			       h->aTidData[i].name, d->lct_data.tid);
#endif
			/* Get the max fragments/request */
			i2o_query_scalar(c, d->lct_data.tid, 0xF103, 3, &limit, 2);
                
			/* sanity */
			if ( limit == 0 )
			{
				printk(KERN_WARNING "i2o_scsi: Ignoring unreasonable SG limit of 0 from IOP!\n");
				limit = 1;
			}
                
			shpnt->sg_tablesize = limit;

			dprintk(KERN_INFO "i2o_scsi: set scatter-gather to %d.\n", shpnt->sg_tablesize);

			return i;
		}
	}
        return -1;
}

struct i2o_scsi_host * is1000_get_hostdata(u32 unit)
{
	struct Scsi_Host * pScsiHost = 
		(struct Scsi_Host *)iscsi_adapter_priv[unit].shpnt;

	if (pScsiHost == NULL)
	{
		printk(KERN_ERR "is1000: is1000_get_hostdata found NULL Scsi_Host *\n");
		return NULL;
	}

	return (struct i2o_scsi_host *) 
		((struct Scsi_Host *)iscsi_adapter_priv[unit].shpnt)->hostdata;
}

/* Returns tid number of targetId target and a lot more! */
int is1000_tid_match(u8 *pTargetId, char *targetName, int unit)
{
	struct i2o_scsi_host *h;
	int i, targetId = -1;

	h = is1000_get_hostdata(unit);

	if (h == NULL)
	{
		printk(KERN_ERR "is1000: is1000_tid_match found NULL i2o_scsi_host *\n");
		return 0;
	}

	/* If targetId not valid, find next available target id */
	if (*pTargetId == IS_TARGETID_INVALID)
	{
		/* Otherwise, give it the first available target Id */
		for (i = 0 ; i < IOP_MAXIMUM_TARGETS ; i++)
		{
			if (h->aTaskData[i].inUse == 0)
			{
				targetId = i;
				h->aTaskData[i].inUse = 1;
#ifdef SCSI_DEBUG
				printk(KERN_ERR "is1000: is1000_tid_match: assigning targetId %d\n",
				       i);
#endif
				break;
			}
		}

		/* check here to see if we really got a valid target id */
		if (targetId == -1)
		{
			printk(KERN_ERR "is1000: No target Ids left, reached max targets?\n");
			return 0;
		}
		else
			*pTargetId = targetId;
	}
	else
	{
		targetId = *pTargetId;

		/* Otherwise check that targetId is already 
		 * reserved for this target Id
		 */

#ifdef SCSI_DEBUG
		printk(KERN_ERR "is1000: tid_match called with targetId %d\n",
		       targetId);
		if (h->aTaskData[targetId].inUse == 0)
		{
			printk(KERN_ERR "is1000: Something wrong with targetId and tid mgmt!\n");
			return 0;
		}
#endif
	}


	/* Search aTidData[].name for this target name. Set aTidData[].index to
	   targetId, and set aTaskData[].pTidData to this tid data struct 
	*/

	for (i = 0; i < IOP_MAXIMUM_TARGETS ; i++)
	{
#ifdef SCSI_DEBUG
		printk(KERN_ERR "is1000: targetName <%s>  h->aTidData[i].name <%s>\n",
		       targetName, h->aTidData[i].name);             
#endif
		if (!strncmp(targetName, h->aTidData[i].name,
			     ISCSI_TARGET_NAME_LENGTH - 1))
		{
			h->aTaskData[targetId].pTidData = &h->aTidData[i];
			h->aTidData[i].index = targetId;
#ifdef SCSI_DEBUG
			printk(KERN_ERR "is1000: _tid_match returning %d\n",
			       h->aTidData[i].tid);
#endif

			h->aTidData[i].status = ISCSI_TARGET_STATUS_CONNECTED;

			return h->aTidData[i].tid;
		}
	}

#ifdef SCSI_DEBUG
	printk(KERN_ERR "is1000: Probably don't have LCT update yet...\n");
#endif

	return -1;
}



int i2o_scsi_init_wrapper(struct i2o_controller *c, struct i2o_device *d, u32 hp)
{
	return i2o_scsi_init(c, d, (struct Scsi_Host *) hp);
}


int i2o_scsi_detect(Scsi_Host_Template * tpnt)
{
	unsigned long flags;
	struct Scsi_Host *shpnt = NULL;
	int i;
	int count;

	printk(KERN_INFO "is1000: SCSI detection started\n");

	if(i2o_install_handler(&i2o_scsi_handler)<0)
	{
		printk(KERN_ERR "is1000: Unable to install OSM handler.\n");
		return -ENODEV;
	}
	scsi_context = i2o_scsi_handler.context;
	if((sg_chain_pool = kmalloc(SG_CHAIN_POOL_SZ, GFP_KERNEL)) == NULL)
	{
		printk("is1000: Unable to alloc %d byte SG chain buffer pool.\n", SG_CHAIN_POOL_SZ);
		printk("is1000: SG chaining DISABLED!\n");
		sg_max_frags = 11;
	}
	else
	{
		dprintk("  chain_pool: %d bytes @ %p\n", SG_CHAIN_POOL_SZ, sg_chain_pool);
		dprintk("  (%d byte buffers X %d can_queue X %d i2o controllers)\n",
			SG_CHAIN_BUF_SZ, I2O_SCSI_CAN_QUEUE, i2o_num_controllers);
                printk(KERN_INFO "is1000: Initializing %d iSCSI Contoller%s\n",
                       i2o_num_controllers, (i2o_num_controllers==1) ? "" : "s");
		sg_max_frags = SG_MAX_FRAGS;    // 64
	}
	
	init_timer(&retry_timer);
	retry_timer.data = 0UL;
	retry_timer.function = i2o_retry_run;

	for (count = 0, i = 0; i < MAX_I2O_CONTROLLERS; i++)
	{
		struct i2o_controller *c=i2o_find_controller(i);
		
		if(c==NULL)
                        continue;

		/*
		 *	Fixme - we need some altered device locking. This
		 *	is racing with device addition in theory. Easy to fix.
		 */
		
                shpnt = scsi_register(tpnt, sizeof(struct i2o_scsi_host));

                if (!shpnt)
                {
                        dprintk(KERN_INFO "scsi_register %x failed\n",
                                (int)tpnt);
                        continue;
                }
                save_flags(flags);
                cli();
                shpnt->unique_id = (u32)c;
                shpnt->io_port = 0;
                shpnt->n_io_port = 0;
                shpnt->irq = 0;
                shpnt->max_lun = I2O_ISCSI_MAX_LUNS;

                shpnt->max_id = IOP_MAXIMUM_TARGETS;

#if (defined(CONFIG_HIGHIO) || defined(AS)) && defined(IS1000_HIGHMEMIO)
		if(c->type == I2O_TYPE_PCI)
#ifdef RH90
			shpnt->pci_dev = c->pdev;
#else
		shpnt->pci_dev = c->bus.pci.pdev;
#endif
#endif

                //                shpnt->max_channel;
#ifdef SCSI_DEBUG
                printk(KERN_ERR "is1000: need to know: shpnt->max_channel = %d\n",
                       shpnt->max_channel);
#endif
                shpnt->this_id = I2O_SCSI_ID;
                restore_flags(flags);

                dprintk(KERN_INFO "is1000: scsi_register returned host %d\n", shpnt->host_no);

                iscsi_adapter_priv[i].shpnt = (u32)shpnt;
                iscsi_adapter_priv[i].hostAdapterNum = shpnt->host_no;

                i2o_scsi_init0(c,
                               (struct Scsi_Host *) iscsi_adapter_priv[i].shpnt);

                dprintk(KERN_INFO "                   after calling i2o_scsi_init0()  shpnt 0x%08x  host_no %d\n\n",
                        (int)shpnt, (int)shpnt->host_no);

                count++;
        }

	i2o_scsi_hosts = count;
	
	if(count==0)
	{
		if(sg_chain_pool!=NULL)
		{
			kfree(sg_chain_pool);
			sg_chain_pool = NULL;
		}
		flush_pending();
		del_timer(&retry_timer);

		i2o_remove_handler(&i2o_scsi_handler);

	}

	dprintk(KERN_INFO "is1000: i2o_scsi_detect exiting\n");

	return count;
}

int i2o_scsi_release(struct Scsi_Host *host)
{
        struct i2o_controller *c = (struct i2o_controller *)host->unique_id;
        dprintk(KERN_INFO "is1000: i2o_scsi_hosts before decr: %d on controller: 0x%x\n",
		i2o_scsi_hosts, (int)c);
        i2o_unlock_controller(c);

	if(--i2o_scsi_hosts==0)
	{
		if(sg_chain_pool!=NULL)
		{
			kfree(sg_chain_pool);
			sg_chain_pool = NULL;
		}
		flush_pending();
		del_timer(&retry_timer);
		i2o_remove_handler(&i2o_scsi_handler);

	}
	return 0;
}


const char *i2o_scsi_info(struct Scsi_Host *SChost)
{
	struct i2o_scsi_host *hostdata;

	hostdata = (struct i2o_scsi_host *)SChost->hostdata;

	return(&hostdata->controller->name[0]);
}


/*
 * From the wd93 driver:
 * Returns true if there will be a DATA_OUT phase with this command, 
 * false otherwise.
 * (Thanks to Joerg Dorchain for the research and suggestion.)
 *
 */
static int is_dir_out(Scsi_Cmnd *cmd)
{
	switch (cmd->cmnd[0]) 
	{
	case WRITE_6:           case WRITE_10:          case WRITE_12:
	case WRITE_LONG:        case WRITE_SAME:        case WRITE_BUFFER:
	case WRITE_VERIFY:      case WRITE_VERIFY_12:      
	case COMPARE:           case COPY:              case COPY_VERIFY:
	case SEARCH_EQUAL:      case SEARCH_HIGH:       case SEARCH_LOW:
	case SEARCH_EQUAL_12:   case SEARCH_HIGH_12:    case SEARCH_LOW_12:      
	case FORMAT_UNIT:       case REASSIGN_BLOCKS:   case RESERVE:
	case MODE_SELECT:       case MODE_SELECT_10:    case LOG_SELECT:
	case SEND_DIAGNOSTIC:   case CHANGE_DEFINITION: case UPDATE_BLOCK:
	case SET_WINDOW:        case MEDIUM_SCAN:       case SEND_VOLUME_TAG:
	case 0xea:
		return 1;
	default:
		return 0;
	}
}


void is1000_add_timer(Scsi_Cmnd * SCpnt,
                      int timeout,
                      void (*complete) (unsigned long))
{

	SCpnt->eh_timeout.data = (unsigned long) SCpnt;
	SCpnt->eh_timeout.expires = jiffies + timeout;
	SCpnt->eh_timeout.function = (void (*)(unsigned long)) complete;

#ifdef SCSI_DEBUG
	dprintk(KERN_INFO "is1000: is1000_add_timer() called\n");
#endif
	add_timer(&SCpnt->eh_timeout);
}

void is1000_timeout_func(unsigned long data)
{
	/* No need to delete the timer, if we are here then
	 * the timer went off and was deleted already
	 timer_status = scsi_delete_timer(SCpnt); */

#ifdef IOCTLDEBUG
	printk(KERN_INFO "is1000: is1000_timeout_func() adding timer again data 0x%x  %d\n",
	       (int)data, (int)((Scsi_Cmnd *)data)->serial_number );
#endif

	IS1000_CMD_LOCK;

	is1000_add_timer((Scsi_Cmnd *)data, 30 * HZ,
			 (void (*)(unsigned long)) (is1000_timeout_func));

	IS1000_CMD_UNLOCK;
}

void is1000_failed_func(unsigned long data)
{
        PISCSI_TID_DATA pTidData;
	struct Scsi_Host *host;
	struct i2o_scsi_host *hostdata;
        Scsi_Cmnd * SCpnt  = (Scsi_Cmnd *) data;

        IS1000_CMD_LOCK;

	host = SCpnt->host;
	hostdata = (struct i2o_scsi_host *)host->hostdata;
	
        if (hostdata == NULL)
	{
		printk(KERN_ERR "is1000: scsi host data NULL, failing scsi cmd\n");
		SCpnt->result = DID_ERROR << 16;
		add_completion_timer(SCpnt);
		SCpnt->scsi_done(SCpnt);
		IS1000_CMD_UNLOCK;
		return;
	}

        pTidData = hostdata->aTaskData[SCpnt->target].pTidData;

        if (pTidData == NULL)
	{
		printk(KERN_ERR "is1000: scsi target data NULL, failing scsi cmd, 0x%x\n",
		       (int)SCpnt->scsi_done);
		SCpnt->result = DID_ERROR << 16;
		add_completion_timer(SCpnt);
		SCpnt->scsi_done(SCpnt);
		IS1000_CMD_UNLOCK;
		return;
	}

        if (pTidData->status == -1)
	{
		printk(KERN_ERR "is1000: scsi target already removed\n");
		SCpnt->result = DID_ERROR << 16;
		add_completion_timer(SCpnt);
		SCpnt->scsi_done(SCpnt);
		IS1000_CMD_UNLOCK;
		return;
	}

        /* No need to delete the timer, if we are here then
         * the timer went off and was deleted already
         timer_status = scsi_delete_timer(SCpnt); */

        if (pTidData->status == ISCSI_TARGET_STATUS_FAILED)
	{
#ifdef SCSI_DEBUG
		printk(KERN_INFO 
		       "is1000: is1000_failed_func, adding again   %d   0x%x\n",
		       (int)SCpnt->serial_number, (int)SCpnt->scsi_done);
#endif
		is1000_add_timer(SCpnt, 3 * HZ,
				 (void (*)(unsigned long)) (is1000_failed_func));
		IS1000_CMD_UNLOCK;
	}
        else
	{
		int result;
#ifdef SCSI_DEBUG            
		printk(KERN_INFO 
		       "is1000: * * * * * * *queueing command again   %d\n",
		       (int)SCpnt->serial_number);
#endif
		is1000_add_timer(SCpnt, 3 * HZ,
				 (void (*)(unsigned long)) (is1000_timeout_func));
		IS1000_CMD_UNLOCK;

		result = i2o_scsi_queuecommand(SCpnt, SCpnt->scsi_done);
	}
}


/* The done func passed here is scsi layer's scsi_done() function.
   The Scsi_Cmnd *SCpnt's scsi_done() field is ours to use.
*/
int i2o_scsi_queuecommand(Scsi_Cmnd * SCpnt, void (*done) (Scsi_Cmnd *))
{
	int i;
	int tid;
	struct i2o_controller *c;
	struct Scsi_Host *host;
	struct i2o_scsi_host *hostdata;
	u32 *msg, *mptr;
	u32 m;
	u32 *lenptr;
	int direction;
	int scsidir;
	u32 len;
	u32 reqlen;
	u32 tag;
        u32 timer_status;
        PISCSI_TID_DATA pTidData;
	
	static int max_qd = 6;
	
	/*
	 *	Do the incoming paperwork
	 */
	
	host = SCpnt->host;
	hostdata = (struct i2o_scsi_host *)host->hostdata;
	c = hostdata->controller;
	prefetch(c);
	prefetchw(&queue_depth);

	if(SCpnt->target > IOP_MAXIMUM_TARGETS)
	{
		SCpnt->result = DID_NO_CONNECT << 16;
		done(SCpnt);
		printk(KERN_ERR "i2o_scsi: Wild target %d.\n", SCpnt->target);
		return 0;
	}
	
        pTidData = hostdata->aTaskData[SCpnt->target].pTidData;

	/* We don't have such a device. Pretend we did the command 
	   and that selection timed out */
        if (pTidData == NULL)
	{
#ifdef SCSI_DEBUG
		printk(KERN_ERR "is1000: pTidData is NULL\n");
#endif
		SCpnt->result = DID_NO_CONNECT << 16;
		done(SCpnt);
		return 0;
		/* see case of tid == -1 below  - but that will NEVER happen!!  */
	}

	tid = pTidData->tid;
	
	dprintk(KERN_INFO "qcmd: SCpnt->target %d   SCpnt->lun %d\n",
		SCpnt->target, SCpnt->lun);

	SCpnt->scsi_done = done;	/* set ptr to done function */

	if(tid == -1)
	{
		SCpnt->result = DID_NO_CONNECT << 16;
		done(SCpnt);
		return 0;
	}
	
	dprintk(KERN_INFO "Real scsi messages.\n");

        /* See what kind of command it is, borrowed from linux-iscsi */


        if (SCpnt->device && (SCpnt->device->type == TYPE_DISK)) {
		switch (SCpnt->cmnd[0]) {
		case INQUIRY:
		case REPORT_LUNS:
		case TEST_UNIT_READY:
		case READ_CAPACITY:
		case START_STOP:
		case MODE_SENSE:
#ifdef SCSI_DEBUG
			printk(KERN_INFO "is1000: queuecommand: not retryable 0x%02x\n",
			       (int)SCpnt->cmnd[0]);
			dprintk(KERN_INFO 
				"is1000: not touching command: SCpnt->eh_timeout.expires 0x%x\n", 
				(int)SCpnt->eh_timeout.expires);
#endif

			if (pTidData->status != ISCSI_TARGET_STATUS_CONNECTED)
			{
#ifdef SCSI_DEBUG
				printk(KERN_INFO "is1000: not sending, status -> DID_NO_CONNECT\n");
#endif
				//SCpnt->result = DID_ERROR << 16;
				SCpnt->result = DID_NO_CONNECT << 16;
				if (done != NULL)
					done(SCpnt);
				return 0;
			}

			break;
		default:
		{
#ifdef SCSI_DEBUG
			printk(KERN_INFO "\nis1000: queuecommand: RETRYABLE 0x%02x   %d\n",
			       (int)SCpnt->cmnd[0], (int)SCpnt->serial_number);
			//dprintk(KERN_INFO "is1000: ->list.next %p  ->base %p\n",
			//SCpnt->eh_timeout.list.next,
			//SCpnt->eh_timeout.base);
#endif
			if (SCpnt->eh_timeout.function == NULL)
			{
				printk(KERN_ERR
				       "\n\nis1000: SCSI command found with no timer!\n");
				break;
			}
			else
			{
				dprintk(KERN_INFO 
					"is1000: old timeout: SCpnt->eh_timeout.expires 0x%x\n", 
					(int)SCpnt->eh_timeout.expires);

				timer_status = scsi_delete_timer(SCpnt);
#ifdef SCSI_DEBUG
				dprintk(KERN_INFO "is1000: DELETING TIMER!  timer_status = %d\n",
					timer_status);
#endif

				if (pTidData->status == ISCSI_TARGET_STATUS_FAILED)
				{
#ifdef SCSI_DEBUG
					printk(KERN_INFO "is1000: adding to timer with _failed_func\n");
#endif
					is1000_add_timer(SCpnt, 3 * HZ,
							 (void (*)(unsigned long)) (is1000_failed_func));
					return 0;
				}
				else {
					is1000_add_timer(SCpnt, 30 * HZ,
							 (void (*)(unsigned long)) (is1000_timeout_func));
				}
			}
		}
		}
        }
#ifdef SCSI_DEBUG
        else {
		printk(KERN_INFO "is1000: queuecommand: cmd not for disk: 0x%02x\n",
		       (int)SCpnt->cmnd[0]);
        }
#endif

	/*
	 *	Obtain an I2O message. Right now we _have_ to obtain one
	 *	until the scsi layer stuff is cleaned up.
	 */	
	 
	do
	{
		mb();
#ifdef RH90
		m = le32_to_cpu(I2O_POST_READ32(c));
#else
		m = I2O_POST_READ32(c);
#endif
	}
	while(m==0xFFFFFFFF);
	msg = (u32 *)(c->mem_offset + m);
	
	/*
	 *	Put together a scsi execscb message
	 */
	
	len = SCpnt->request_bufflen;
	direction = 0x00000000;			// SGL IN  (osm<--iop)
	
	/*
	 *	The scsi layer should be handling this stuff
	 */
	
	scsidir = 0x00000000;			// DATA NO XFER
	if(len)
	{
		if(is_dir_out(SCpnt))
		{
			direction=0x04000000;	// SGL OUT  (osm-->iop)
			scsidir  =0x80000000;	// DATA OUT (iop-->dev)
		}
		else
		{
			scsidir  =0x40000000;	// DATA IN  (iop<--dev)
		}
	}
	
#ifdef RH90
	i2o_raw_writel(I2O_HBA_BUS_EXEC<<24|HOST_TID<<12|tid, &msg[1]);
	i2o_raw_writel(scsi_context, &msg[2]);	/* So the I2O layer passes to us */
	/* Sorry 64bit folks. FIXME */
	i2o_raw_writel((u32)SCpnt, &msg[3]);	/* We want the SCSI control block back */
#else
	__raw_writel(I2O_HBA_BUS_EXEC<<24|HOST_TID<<12|tid, &msg[1]);
	__raw_writel(scsi_context, &msg[2]);	/* So the I2O layer passes to us */
	/* Sorry 64bit folks. FIXME */
	__raw_writel((u32)SCpnt, &msg[3]);	/* We want the SCSI control block back */
#endif

	/* LSI_920_PCI_QUIRK
	 *
	 *	Intermittant observations of msg frame word data corruption
	 *	observed on msg[4] after:
	 *	  WRITE, READ-MODIFY-WRITE
	 *	operations.  19990606 -sralston
	 *
	 *	(Hence we build this word via tag. Its good practice anyway
	 *	 we don't want fetches over PCI needlessly)
	 */

	tag=0;
	
	/*
	 *	Attach tags to the devices
	 */	
	if(SCpnt->device->tagged_supported)
	{
		/*
		 *	Some drives are too stupid to handle fairness issues
		 *	with tagged queueing. We throw in the odd ordered
		 *	tag to stop them starving themselves.
		 */
		if((jiffies - hostdata->tagclock[SCpnt->target]) > (5*HZ))
		{
			tag=0x01800000;		/* ORDERED! */
			hostdata->tagclock[SCpnt->target]=jiffies;
		}
		else
		{
			/* Hmmm...  I always see value of 0 here,
			 *  of which {HEAD_OF, ORDERED, SIMPLE} are NOT!  -sralston
			 */
			if(SCpnt->tag == HEAD_OF_QUEUE_TAG)
				tag=0x01000000;
			else if(SCpnt->tag == ORDERED_QUEUE_TAG)
				tag=0x01800000;
		}
	}

	/* Direction, disconnect ok, tag, sense_data in buffer, CDBLen */

#ifdef RH90
	i2o_raw_writel(scsidir|0x20600000|(SCpnt->lun << 8)|SCpnt->cmd_len|tag, &msg[4]);
#else  
	__raw_writel(scsidir|0x20600000| (SCpnt->lun << 8) | SCpnt->cmd_len|tag, &msg[4]);
#endif
	mptr=msg+5;

	/* 
	 *	Write SCSI command into the message - always 16 byte block 
	 */
	 
	memcpy_toio(mptr, SCpnt->cmnd, 16);
	mptr+=4;
	lenptr=mptr++;		/* Remember me - fill in when we know */
	
	reqlen = 12;		// SINGLE SGE
        //        printk(KERN_ERR "is1000: :0: reqlen: %d\n", reqlen);
	
	/*
	 *	Now fill in the SGList and command 
	 *
	 *	FIXME: we need to set the sglist limits according to the 
	 *	message size of the I2O controller. We might only have room
	 *	for 6 or so worst case
	 */
	
	if(SCpnt->use_sg)
	{
		struct scatterlist *sg = (struct scatterlist *)SCpnt->request_buffer;
		int chain = 0;
		
		len = 0;

		if((sg_max_frags > SCpnt->host->sg_tablesize) 
                   && (SCpnt->use_sg > SCpnt->host->sg_tablesize))
		{
			chain = 1;
			/*
			 *	Need to chain!
			 */
#if (defined(CONFIG_HIGHIO) || defined(AS)) && defined(IS1000_HIGHMEMIO)
			printk("is1000: !!!!!!!!! BROKEN BROKEN: HIGHMEMIO NEED TO CHAIN\n");
#endif

#ifdef RH90
			i2o_raw_writel(direction|0xB0000000|(SCpnt->use_sg*2*4), mptr++);
			i2o_raw_writel(virt_to_bus(sg_chain_pool + sg_chain_tag), mptr);
#else
			__raw_writel(direction|0xB0000000|(SCpnt->use_sg*2*4), mptr++);
			__raw_writel(virt_to_bus(sg_chain_pool + sg_chain_tag), mptr);
#endif
			mptr = (u32*)(sg_chain_pool + sg_chain_tag);
			if (SCpnt->use_sg > max_sg_len)
			{
				max_sg_len = SCpnt->use_sg;
				printk("i2o_scsi: Chain SG! SCpnt=%p, SG_FragCnt=%d, SG_idx=%d\n",
                                       SCpnt, SCpnt->use_sg, sg_chain_tag);
			}
			if ( ++sg_chain_tag == SG_MAX_BUFS )
				sg_chain_tag = 0;
			for(i = 0 ; i < SCpnt->use_sg; i++)
			{
				*mptr++=direction|0x10000000|sg->length;
				len+=sg->length;
				*mptr++=virt_to_bus(sg->address);
				sg++;
			}
			mptr[-2]=direction|0x50000000|(sg-1)->length;
		}
		else
		{		
#if (defined(CONFIG_HIGHIO) || defined(AS)) && defined(IS1000_HIGHMEMIO)
#ifdef RH90
			int sgcnt = pci_map_sg(c->pdev, sg, SCpnt->use_sg, direction);
			for(i = 0 ; i < sgcnt; i++)
			{
				i2o_raw_writel(direction|0x10000000|sg_dma_len(sg), mptr++);
				len+=sg_dma_len(sg);
				i2o_raw_writel(sg_dma_address(sg), mptr++);
				sg++;
			}
			pci_unmap_sg(c->pdev, sg, SCpnt->use_sg, direction);
#else
			int sgcnt = pci_map_sg(c->bus.pci.pdev, sg, SCpnt->use_sg, direction);
			for(i = 0 ; i < sgcnt; i++)
			{
				__raw_writel(direction|0x10000000|sg_dma_len(sg), mptr++);
				len+=sg_dma_len(sg);
				__raw_writel(sg_dma_address(sg), mptr++);
				sg++;
			}
			pci_unmap_sg(c->bus.pci.pdev, sg, SCpnt->use_sg, direction);
#endif
#else
			for(i = 0 ; i < SCpnt->use_sg; i++)
			{
#ifdef RH90
				i2o_raw_writel(direction|0x10000000|sg->length, mptr++);
				len+=sg->length;
				i2o_raw_writel(virt_to_bus(sg->address), mptr++);
#else
				__raw_writel(direction|0x10000000|sg->length, mptr++);
				len+=sg->length;
				__raw_writel(virt_to_bus(sg->address), mptr++);
#endif

				sg++;
			}
#endif //IS1000_HIGHMEMIO

		}
		
		if(!chain)
			reqlen = mptr - msg;
		
                //                printk(KERN_ERR "is1000: :1: (chain) reqlen: %d\n", reqlen);
#ifdef RH90
		i2o_raw_writel(len, lenptr);
#else
		__raw_writel(len, lenptr);
#endif
		
		if(len != SCpnt->underflow)
			printk("Cmd len %08X Cmd underflow %08X\n",
                               len, SCpnt->underflow);
	}
	else
	{
		dprintk(KERN_INFO "non sg for %p, %d\n", SCpnt->request_buffer,
                        SCpnt->request_bufflen);
#ifdef RH90
		i2o_raw_writel(len = SCpnt->request_bufflen, lenptr);
#else
		__raw_writel(len = SCpnt->request_bufflen, lenptr);
#endif
		if(len == 0)
		{
			/* If there's no buffer, add a zero length dummy buffer */
#ifdef RH90
			i2o_raw_writel(0x10000000|direction| 0, mptr++);
			i2o_raw_writel(virt_to_bus((void *) SCpnt->sense_buffer), mptr++);
#else
			__raw_writel(0x10000000|direction| 0, mptr++);
			__raw_writel(virt_to_bus((void *) SCpnt->sense_buffer), mptr++);
#endif

		}
		else
		{
#ifdef RH90
			i2o_raw_writel(0x10000000|direction|SCpnt->request_bufflen, mptr++);
			i2o_raw_writel(virt_to_bus(SCpnt->request_buffer), mptr++);
#else
			__raw_writel(0x10000000|direction|SCpnt->request_bufflen, mptr++);
			__raw_writel(virt_to_bus(SCpnt->request_buffer), mptr++);
#endif
		}
	}

        /* Add an SG element to return the sense buffer - 16 bytes in size */

#ifdef RH90
        i2o_raw_writel(0xD0000000 | sizeof(SCpnt->sense_buffer), mptr++);
        i2o_raw_writel(virt_to_bus((void *) SCpnt->sense_buffer), mptr++);
#else
        __raw_writel(0xD0000000 | sizeof(SCpnt->sense_buffer), mptr++);
        __raw_writel(virt_to_bus((void *) SCpnt->sense_buffer), mptr++);
#endif
        reqlen += 2;
	
	/*
	 *	Stick the headers on 
	 */

#ifdef RH90
	i2o_raw_writel(reqlen<<16 | SGL_OFFSET_10, msg);
#else
	__raw_writel(reqlen<<16 | SGL_OFFSET_10, msg);
#endif	


	/* Queue the message */
	i2o_post_message(c,m);
	
	atomic_inc(&queue_depth);
	
	if(atomic_read(&queue_depth)> max_qd)
	{
		max_qd=atomic_read(&queue_depth);
		dprintk(KERN_INFO "is1000: Queue depth now %d.\n", max_qd);
	}
	
	mb();
#ifdef SCSI_DEBUG
	printk(KERN_INFO "Issued %ld            %X\n", SCpnt->serial_number, 
	       (int)SCpnt);
#endif	
	return 0;

}

int is1000_eh_abort( Scsi_Cmnd *SCpnt )
{
        struct i2o_controller *c;
	struct Scsi_Host *host;
	struct i2o_scsi_host *hostdata;
        PISCSI_TID_DATA pTidData;
        DECLARE_MIDLAYER_FLAGS;
#ifdef RH90
	unsigned long msg;
#else
	u32 *msg;
#endif
	u32 m;
	int tid;
	
        if ((SCpnt->target < 0) || (SCpnt->target > IOP_MAXIMUM_TARGETS))
	{
		printk(KERN_ERR 
		       "is1000: abort error, target %d out of range\n", 
		       SCpnt->target);
		return -1;
	}
          
	host = SCpnt->host;
	hostdata = (struct i2o_scsi_host *)host->hostdata;
        pTidData = hostdata->aTaskData[SCpnt->target].pTidData;

        if (pTidData == NULL)
	{
		printk(KERN_ERR "is1000: no data on target to abort.\n");
		return -1;
	}

	if ((tid = pTidData->tid) == -1)
	{
		printk(KERN_ERR "is1000: can not abort on target %d\n",
                       SCpnt->target);
		return -1;
	}

        if (pTidData->status == -1)
	{
		printk(KERN_ERR 
                       "is1000: can not abort on target %d, bad target status\n",
                       SCpnt->target);
		return -1;
	}

#ifdef RECOVERY_DEBUG
	printk(KERN_ERR "i2o_scsi_abort() tid %d   status %d\n", 
               tid, pTidData->status);
#endif

        UNLOCK_MIDLAYER_LOCK(host);

        printk("is1000: Sending abort: target %d\n", SCpnt->target);

	c = hostdata->controller;
	
	/*
	 *	Obtain an I2O message. Right now we _have_ to obtain one
	 *	until the scsi layer stuff is cleaned up.
	 */	
	 
	do
	{
		mb();
#ifdef RH90
		m = le32_to_cpu(I2O_POST_READ32(c));
#else
		m = I2O_POST_READ32(c);
#endif
	}
	while(m==0xFFFFFFFF);
#ifdef RH90
	msg = c->mem_offset + m;
#else
	msg = (u32 *)(c->mem_offset + m);
#endif	
	SCpnt->scsi_done = internal_done;
        //	SCpnt->SCp.Status = -1;
	SCpnt->SCp.Status = 
		(ISCSI_CMD_WAS_ABORTED | ISCSI_CMD_ABORT_SENT);

#ifdef RH90
	i2o_raw_writel(FOUR_WORD_MSG_SIZE, msg); /* was FIVE_ */
	i2o_raw_writel(I2O_CMD_SCSI_ABORT<<24|HOST_TID<<12|tid, msg+4);
	i2o_raw_writel(scsi_context, msg+8);
	i2o_raw_writel((u32)SCpnt, msg+12);
        /*i2o_raw_writel((u32)SCpnt, msg+16);*/	/* FIXME 32bitism */
#else
	__raw_writel(FOUR_WORD_MSG_SIZE, &msg[0]); /* was FIVE_ */
	__raw_writel(I2O_CMD_SCSI_ABORT<<24|HOST_TID<<12|tid, &msg[1]);
	__raw_writel(scsi_context, &msg[2]);
	__raw_writel((u32)SCpnt, &msg[3]);
        /*_raw_writel((u32)SCpnt, &msg[4]);*/
#endif
	
	wmb();
	i2o_post_message(c,m);
	wmb();

	/* wait for response */
        /* new foo */
        //	while (SCpnt->SCp.Status == -1)
	while (SCpnt->SCp.Status & ISCSI_CMD_ABORT_SENT)
        {
                dprintk(KERN_INFO "is1000: waiting for abort reply\n");
                barrier();
        }

        printk(KERN_INFO "is1000: done waiting for abort reply\n");
#ifdef RECOVERY_DEBUG
        printk(KERN_INFO "is1000: done waiting for abort reply\n");
#endif
        LOCK_MIDLAYER_LOCK(host);

	if (SCpnt->result == (DID_ABORT << 16))
        {
#ifdef RECOVERY_DEBUG
                dprintk(KERN_INFO 
                        "is1000: _eh_abort() returning 0\n");
#endif
                return 0;
        }
	else
        {
#ifdef RECOVERY_DEBUG
                dprintk(KERN_INFO 
                        "is1000: _eh_abort() returning -1, SCpnt->result = %d\n",
                        SCpnt->result);
#endif
                return -1;
        }
}

/* Currently the firmware would do nothing with this, we'd
 * just fail it, too, so by not providing it at all we
 * can fail it really fast
 * 
 * If we did do something io lock would have to be released first.
 */
#if 0
int is1000_eh_device_reset( Scsi_Cmnd *sc )
{
        DECLARE_MIDLAYER_FLAGS;

        dprintk(KERN_INFO "is1000: is1000_eh_device_reset()   -1\n");

        UNLOCK_MIDLAYER_LOCK(host);

        /* Do actual work here, send msg, etc */

        LOCK_MIDLAYER_LOCK(host);

        return -1;
}
#endif

int is1000_bus_reset(u32 unit, u32 targetId, u32 context)
{
	int tid;
	struct i2o_controller *c;
        PISCSI_TID_DATA pTidData;
	struct i2o_scsi_host *hostdata;
	u32 m;
#ifdef RH90
	unsigned long msg;
#else
	u32 *msg;
#endif
	/*
	 *	Find the TID for the bus
	 */

	printk("is1000: Attempting a bus reset.\n");
	
        hostdata = is1000_get_hostdata(unit);

        if (hostdata == NULL)
	{
		dprintk(KERN_ERR "is1000: reset found NULL i2o_scsi_host *\n");
		return -1;
	}

        pTidData = hostdata->aTaskData[targetId].pTidData;

        if (pTidData == NULL)
	{
		dprintk(KERN_ERR "is1000: reset pTidData == NULL\n");
		return -1;
	}
        
        tid = pTidData->tid;

        if (tid == -1)
	{
		dprintk (KERN_ERR "is1000: No tid found on adapter %d\n",
			 unit);
		return -1;
	}

	c = hostdata->controller;

#ifdef RECOVERY_DEBUG
	printk(KERN_INFO "is1000_bus_reset(): tid %d  scsi_context %d   context 0x%x (cmd)\n",
	       tid, scsi_context, (int)context);
#endif
	/*
	 *	Now send a SCSI reset request. Any remaining commands
	 *	will be aborted by the IOP. We need to catch the reply
	 *	possibly ?
	 */
	 
#ifdef RH90
	m = le32_to_cpu(I2O_POST_READ32(c));
#else
	m = I2O_POST_READ32(c);
#endif	
	/*
	 *	No free messages, try again next time - no big deal
	 */
	if(m == 0xFFFFFFFF)
		return -1;
	
#ifdef RH90
	msg = c->mem_offset + m;
	i2o_raw_writel(FOUR_WORD_MSG_SIZE|SGL_OFFSET_0, msg);
	i2o_raw_writel(I2O_HBA_BUS_RESET<<24|HOST_TID<<12|tid, msg+4);
	i2o_raw_writel(scsi_context|0x80000000, msg+8);
#else
	msg = (u32 *)(c->mem_offset + m);
	__raw_writel(FOUR_WORD_MSG_SIZE|SGL_OFFSET_0, &msg[0]);
	__raw_writel(I2O_HBA_BUS_RESET<<24|HOST_TID<<12|tid, &msg[1]);
	__raw_writel(scsi_context|0x80000000, &msg[2]);
#endif

        /* if the context is null, this is a reset sent from an
         * ioctl call, and we will take divergent paths in the reply
         */
        if (context == 0)
	{
#ifdef RH90
		i2o_raw_writel(0xFFFFFFFF, msg+12);
#else
		__raw_writel(0xFFFFFFFF, &msg[3]);
#endif
	}
        else
	{
#ifdef RH90
		i2o_raw_writel(context, msg+12);
#else
		__raw_writel(context, &msg[3]);
#endif
	}

#ifdef RECOVERY_DEBUG
        printk(KERN_INFO "bus reset      msg=   %08X\n", *(unsigned int *)msg);
        printk(KERN_INFO "               msg+4= %08X\n", *(unsigned int *)(msg+4));
        printk(KERN_INFO "               msg+8= %08X\n", *(unsigned int *)(msg+8));
        printk(KERN_INFO "               msg+12=%08X\n", *(unsigned int *)(msg+12));
        printk(KERN_INFO "               msg+16=%08X\n", *(unsigned int *)(msg+16));
#endif
	wmb();
        i2o_post_message(c,m);

        /* Currently no way to easily wait for a response... */

#ifdef RECOVERY_DEBUG
	printk(KERN_INFO "is1000_bus_reset(): sent request\n");
#endif
	return 0;
}

int is1000_eh_bus_reset( Scsi_Cmnd *sc )
{
	struct Scsi_Host *host;
        DECLARE_MIDLAYER_FLAGS;
	struct i2o_controller *c;
	struct i2o_scsi_host *hostdata;
        int ret;

	host = sc->host;
	hostdata = (struct i2o_scsi_host *)host->hostdata;
        if (hostdata == NULL)
	{
		printk(KERN_ERR 
		       "is1000: bus reset error on target %d, no host?\n", 
		       sc->target);
		return -1;
	}
	if((sc->target < 0) || (sc->target > IOP_MAXIMUM_TARGETS))
	{
		printk(KERN_ERR 
		       "is1000: bus reset error, target %d out of range\n", 
		       sc->target);
		return -1;
	}

	c = hostdata->controller;

	sc->scsi_done = internal_done;
	sc->SCp.Status |= ISCSI_CMD_BUS_RESET_SENT;

        UNLOCK_MIDLAYER_LOCK(host);

        ret = is1000_bus_reset(c->unit, sc->target, (u32) sc);

        //	while (sc->SCp.Status == -1)
	while (sc->SCp.Status & ISCSI_CMD_BUS_RESET_SENT)
	{
		dprintk(KERN_INFO "is1000: waiting for reset reply\n");
		barrier();
        }

        printk(KERN_INFO "is1000: done waiting for reset reply\n");

#ifdef RECOVERY_DEBUG
        printk(KERN_INFO "is1000: done waiting for reset reply\n");
#endif
        
        LOCK_MIDLAYER_LOCK(host);

        if (ret < 0)
	{
		dprintk(KERN_INFO 
			"is1000: is1000_eh_bus_reset returning status %d\n", 
			ret);
		return -1;
	}

	if (sc->result == (DID_RESET << 16))
        {
                dprintk(KERN_INFO 
                        "is1000: _eh_abort() returning 0\n");
                return 0;
        }
	else
        {
                dprintk(KERN_INFO 
                        "is1000: _eh_abort() returning -1, SCpnt->result = %d\n",
                        sc->result);
                return -1;
        }

        return 0;
}

/* Currently the firmware would do nothing with this, we'd
 * just fail it, too, so by not providing it at all we
 * can fail it really fast
 * 
 * If we did do something io lock would have to be released first.
 */
#if 0
int is1000_eh_host_reset( Scsi_Cmnd *sc )
{
	struct Scsi_Host *host;
        DECLARE_MIDLAYER_FLAGS;

	host = sc->host;

        UNLOCK_MIDLAYER_LOCK(host);


        LOCK_MIDLAYER_LOCK(host);

        dprintk(KERN_INFO "is1000: is1000_eh_host_reset()   -1\n");
        return -1;
}
#endif

/*
 *	This is anyones guess quite frankly.
 */
 
int i2o_scsi_bios_param(Disk * disk, kdev_t dev, int *ip)
{
	int size;

	size = disk->capacity;
	ip[0] = 64;		/* heads                        */
	ip[1] = 32;		/* sectors                      */
	if ((ip[2] = size >> 11) > 1024) {	/* cylinders, test for big disk */
		ip[0] = 255;	/* heads                        */
		ip[1] = 63;	/* sectors                      */
		ip[2] = size / (255 * 63);	/* cylinders    */
	}
	return 0;
}


/* IS1000: These functions are copied from scsi_module.c,
   instead of included, so we can have control over when to 
   call them. */
int init_this_scsi_driver(void)
{
	driver_template.module = THIS_MODULE;

	scsi_register_module(MODULE_SCSI_HA, &driver_template);
	if (driver_template.present)
		return 0;

	scsi_unregister_module(MODULE_SCSI_HA, &driver_template);
	return -ENODEV;
}

void exit_this_scsi_driver(void)
{
	scsi_unregister_module(MODULE_SCSI_HA, &driver_template);
}

/*
 * Overrides for Emacs so that we follow Linus's tabbing style.
 * Emacs will notice this stuff at the end of the file and automatically
 * adjust the settings for this buffer only.  This must remain at the end
 * of the file.
 * ---------------------------------------------------------------------------
 * Local variables:
 * c-indent-level: 4
 * c-brace-imaginary-offset: 0
 * c-brace-offset: -4
 * c-argdecl-indent: 4
 * c-label-offset: -4
 * c-continued-statement-offset: 4
 * c-continued-brace-offset: 0
 * indent-tabs-mode: nil
 * tab-width: 8
 * End:
 */
