/*******************************************************************************
 * Copyright (C) 2002-2003 Intel Corporation                                   *
 *                                                                             *
 * This program is free software; you can redistribute it and/or               *
 * modify it under the terms of the GNU General Public License                 *
 * as published by the Free Software Foundation; either version 2              *
 * of the License, or (at your option) any later version.                      *
 *                                                                             *
 * This program is distributed in the hope that it will be useful,             *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of              *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               *
 * GNU General Public License for more details.                                *
 *                                                                             *
 * You should have received a copy of the GNU General Public License           *
 * along with this program; if not, write to the Free Software                 *
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA  *
 ******************************************************************************/
/*******************************************************************************
 *                                                                             *
 * File Name:    isdaemon.cpp                                                  *
 *                                                                             *
 * Abstract:     The isdaemon.  This daemon receives asynchronous notifications*
 *               from the driver about target connection status changes.       *
 *                                                                             *
 * Environment:  User space daemon started by is1000 script.                   *
 *                                                                             *
 * Author:       Thomas C. Lea                                                 *
 *                                                                             *
 * Origination:  April 21, 2003                                                *
 ******************************************************************************/

// Include Files
#include <iostream>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <signal.h>
#include <fcntl.h>
#include <string.h>
#include <string>
#include <syslog.h>
#include <stdio.h>
#include "is1000events.h"
#include "is1000privateapi.h"

using namespace std;

// Macro definitions
#define LOCKFILE "/var/run/is1000.pid"

// Local definitions
static bool stop_requested = false;

/** ****************************************************************************
 * Procedure:   mactostring
 *
 * Description: This routine converts a IS_MAC_ADDRESS to a string.
 *
 * @param       MacAddr the MAC address to convert
 *
 * @return      the string representation
 *
 ******************************************************************************/
std::string mactostring (
    const IS_MAC_ADDRESS &MacAddr
)
{
    if (MacAddr.first == 0 && MacAddr.second == 0)
        return "";

    char buf[13];
    sprintf(buf, "%02lX%02lX%02lX%02lX%02lX%02lX",
        MacAddr.first >> 0  & 0xFF,
        MacAddr.first >> 8  & 0xFF,
        MacAddr.first >> 16 & 0xFF,
        MacAddr.first >> 24 & 0xFF,
        MacAddr.second >> 0 & 0xFF,
        MacAddr.second >> 8 & 0xFF);
    return buf;
}

/** ****************************************************************************
 * Procedure:   process_should_exit
 *
 * Description: This routine returns true of the daemon should exit.
 *
 * @return      true if it should exit.
 *
 ******************************************************************************/
bool process_should_exit(void)
{
    return stop_requested;
}

/** ****************************************************************************
 * Procedure:   lock_file
 *
 * Description: This routine locks the referenced file.
 *
 * @param       fd the opened file descriptor of the target file.
 *
 * @return      true if the lock succeeded.
 *
 ******************************************************************************/
bool lock_file(
    int fd
)
{
    struct flock lock;
    int result = 0;

    memset(&lock, 0, sizeof(lock));

    lock.l_type = F_WRLCK;

    // entire file
    lock.l_start = 0;
    lock.l_whence = SEEK_SET;
    lock.l_len = 0;

    // try to acquire the lock, returning if we can't
    result = fcntl(fd, F_SETLK, &lock);

    return (result >= 0);
}

/** ****************************************************************************
 * Procedure:   unlock_file
 *
 * Description: This routine unlocks a file that was previously locked with
 *              lock_file.
 *
 * @param       fd the opened file descriptor of the file.
 *
 * @return      true if the unlock succeeded.
 *
 ******************************************************************************/
bool unlock_file(
    int fd
)
{
    struct flock lock;
    int result = 0;

    memset(&lock, 0, sizeof(lock));

    lock.l_type = F_UNLCK;

    // entire file
    lock.l_start = 0;
    lock.l_whence = SEEK_SET;
    lock.l_len = 0;

    // try to release the lock
    result = fcntl(fd, F_SETLKW, &lock);

    return (result >= 0);
}

/** ****************************************************************************
 * Procedure:   lock_pid_file
 *
 * Description: This routine locks the lockfile and writes our pid to it in
 *              order to prevent more than one daemon from running at a time.
 *
 * @return      true if the operation succeeded.
 *
 ******************************************************************************/
bool lock_pid_file(void)
{
    int fd = open(LOCKFILE, O_CREAT | O_RDWR, S_IRUSR | S_IWUSR);

    if(fd < 0)
    {
        syslog(LOG_ERR, "is1000-daemon: open of lock_file %s failed", LOCKFILE);
        close(fd);
        return false;
    }

    if(!lock_file(fd))
    {
        syslog(LOG_ERR, "is1000-daemon: unable to lock pid file %s.  isdaemon already running?", LOCKFILE);
        close(fd);
        return false;
    }

    // truncate the file, and write our pid
    ftruncate(fd, 0);
    char buffer[64];
    sprintf(buffer, "%d\n", getpid());
    if (write(fd, buffer, strlen(buffer)) != (ssize_t)strlen(buffer))
    {
        syslog(LOG_ERR, "is1000-daemon: error writing pid file %s", LOCKFILE);
        close(fd);
        return false;
    }

    // sync the file to disk
    fsync(fd);

    return true;
}

/** ****************************************************************************
 * Procedure:   findTarget
 *
 * Description: This routine locates a target's corresponding IS_STORAGE_TARGET
 *              structure.
 *
 * @param       mac the corresponding adapter's MAC address
 *
 * @param       tid the target's SCSI target id
 *
 * @param       targetname the target's iSCSI name
 *
 * @param       st the structure to receive the target details
 *
 * @return      true if the operation succeeded.
 *
 ******************************************************************************/
bool findTarget (
    const IS_MAC_ADDRESS &mac,
    u32 tid,
    const char *targetname,
    IS_STORAGE_TARGET &st
)
{
    IS_RESULT result = IS_SUCCESS;
    u32 numTargets = 0;
    if((result = GetTargetsCount(&mac, &numTargets, 0)) != IS_SUCCESS)
    {
        return false;
    }
    else
    {
        for(u32 i = 0; i < numTargets; i++)
        {
            if((result = EnumTargets(&mac, i, &st, 0)) != IS_SUCCESS)
            {
                return false;
            }
            else
            {
                //if we are passed the portalgroup on the end of the targetname,
                // remove it
                char *pgpos = strrchr(targetname, ',');
                if(pgpos)
                    *pgpos='\0';

                if(!strncmp(targetname, st.iscsi_nodename, IS_MAX_NODENAME_STRING) &&
                    st.targetId == tid)
                    return true;
            }
        }
    }

    return false;
}

/** ****************************************************************************
 * Procedure:   sigterm_handler
 *
 * Description: This routine handles the SIGTERM signal and notifies the daemon
 *              that it should exit.
 *
 * @param       unused unused signal handler argument.
 ******************************************************************************/
void sigterm_handler(
    int unused
)
{
    syslog(LOG_INFO, "is1000-daemon: received termination signal, shutting down");
    stop_requested = true;
}

/** ****************************************************************************
 * Procedure:   main
 *
 * Description: This is the main function for the isdaemon.  It contains the
 *              processing loop that receives events from the driver.
 *
 * @param       argc the number of command-line arguments.
 *
 * @param       argv the array of command-line string arguments.
 *
 * @return      zero on success, non-zero on failure.
 ******************************************************************************/
int main(
    int argc,
    char **argv
)
{
    int rc = 0;
    int fd;
    int foreground = 0;             //default is to run in background

    if(argc > 1)
    {
        if(!strcmp("-f", argv[1]))  //-f will cause us to run in foreground
            foreground = 1;
    }

    if (!foreground)
    {
        pid_t pid;

        // Turn into a daemon

        close(STDIN_FILENO);
        close(STDOUT_FILENO);
        close(STDERR_FILENO);
        chdir("/");

        if ((pid = fork()))
        {
            // the parent exits
            return 0;
        }
        else if (pid < 0)
        {
            syslog(LOG_ERR, "is1000-daemon: fork failed");
            return -1;
        }
        else
        {
            // child starts a new session
            setsid();
        }
    } // if (!foreground)

    // make sure we are the only instance running
    if(!lock_pid_file())
    {
        return -1;
    }

    // set up our TERM signal handler
    struct sigaction action;
    memset(&action, 0, sizeof(action));
    action.sa_sigaction = NULL;
    action.sa_flags = 0;
    action.sa_handler = sigterm_handler;
    sigaction(SIGTERM, &action, 0);
    sigaction(SIGINT, &action, 0);

    if(InitializeLibrary(0) != IS_SUCCESS)
    {
        syslog(LOG_ERR, "is1000-daemon: InitilizeLibrary failed");
        rc = -1;
        goto cleanup;
    }

    fd = open("/dev/is1000", O_RDONLY);
    if(fd == -1)
    {
        syslog(LOG_ERR, "is1000-daemon: Unable to open driver.");
        rc = -1;
        goto cleanup;
    }

    while(!process_should_exit())
    {
        IS1000_EVENT event;
        memset(&event, 0, sizeof(IS1000_EVENT));
        int numbytesread;
        if((numbytesread = read(fd, (void*)&event, sizeof(IS1000_EVENT))) == -1)
        {
            syslog(LOG_ERR, "is1000-daemon: Read error.  Bailing.");
            rc = -1;
            goto cleanup;
        }

        //could have been interrupted in order to shut down
        if(process_should_exit())
        {
            break;
        }

        if(numbytesread != sizeof(IS1000_EVENT))
        {
            syslog(LOG_ERR,
                   "is1000-daemon: Received message with size %d, but I was expecting %d bytes!",
                   numbytesread, sizeof(IS1000_EVENT));
                   
            break;
        }
        else
        {
            IS_RESULT result = IS_SUCCESS;
            IS_STORAGE_TARGET st;
            bool foundTarget = findTarget(event.MACAddr,
                                          event.targetId,
                                          event.targetname,
                                          st);

            switch(event.event)
            {
            case ISEVENT_TARGET_PERMANENTLY_REDIRECTED:
            {
                syslog(LOG_INFO,
                       "is1000-daemon: TARGET PERMANENTLY REDIRECTED: %s",
                       event.targetname);
                       
                if(foundTarget)
                {
                    result = TargetPropertiesChanged(&event.MACAddr,
                                                     st.address.ip,
                                                     st.address.port,
                                                     event.targetname,
                                                     0);
                                                     
                    if(result != IS_SUCCESS)
                    {
                        syslog(LOG_INFO,
                               "is1000-daemon: TargetPropertiesChanged failed: %ld",
                               result);
                    }
                }
                else
                {
                    syslog(LOG_INFO, "is1000-daemon: Target not found!");
                }
            }
            break;

            case ISEVENT_TARGET_CONNECTION_FAILED:
            {
                syslog(LOG_WARNING,
                       "is1000-daemon: TARGET CONNECTION FAILED: %s",
                       event.targetname);
                       
                if(foundTarget)
                {
                    result = SetTargetStatus(&event.MACAddr,
                                             st.address.ip,
                                             st.address.port,
                                             event.targetname,
                                             TARGET_STATUS_RECOVERING,
                                             0);
                                             
                    if(result != IS_SUCCESS)
                    {
                        syslog(LOG_ERR,
                               "is1000-daemon: SetTargetStatus failed: %ld",
                               result);
                    }
                }
                else
                {
                    syslog(LOG_ERR, "is1000-daemon: Target not found!");
                }
            }
            break;

            case ISEVENT_TARGET_CONNECTION_NORMAL:
            {
                syslog(LOG_INFO,
                       "is1000-daemon: TARGET CONNECTION NORMAL: %s",
                       event.targetname);
                       
                if(foundTarget)
                {
                    result = SetTargetStatus(&event.MACAddr,
                                             st.address.ip,
                                             st.address.port,
                                             event.targetname,
                                             TARGET_STATUS_CONNECTED,
                                             0);
                                             
                    if(result != IS_SUCCESS)
                    {
                        syslog(LOG_ERR,
                               "is1000-daemon: SetTargetStatus failed: %ld",
                               result);
                    }
                }
                else
                {
                    syslog(LOG_ERR, "is1000-daemon: Target not found!");
                }
            }
            break;

            case ISEVENT_TARGET_CONNECTION_FAULTED:
            {
                syslog(LOG_CRIT,
                       "is1000-daemon: TARGET CONNECTION FAULTED: %s",
                       event.targetname);
                       
                if(foundTarget)
                {
                    result = SetTargetStatus(&event.MACAddr,
                                             st.address.ip,
                                             st.address.port,
                                             event.targetname,
                                             TARGET_STATUS_CONNECTION_LOST,
                                             0);
                                             
                    if(result != IS_SUCCESS)
                    {
                        syslog(LOG_ERR,
                               "is1000-daemon: SetTargetStatus failed: %ld",
                               result);
                    }
                }
                else
                {
                    syslog(LOG_ERR, "is1000-daemon: Target not found!");
                }
            }
            break;
            
            case ISEVENT_HOST_ADAPTER_LINK_DOWN:
            {
                syslog(LOG_WARNING,
                       "is1000-daemon: LINK DOWN ON ADAPTER: (MAC #%s, HBA #%ld)",
                        mactostring(event.MACAddr).c_str(),
                        event.host);
            }
            break;

            case ISEVENT_HOST_ADAPTER_LINK_UP:
            {
                syslog(LOG_INFO,
                       "is1000-daemon: LINK UP ON ADAPTER: (MAC #%s, HBA #%ld)",
                        mactostring(event.MACAddr).c_str(),
                        event.host);
            }
            break;

            case ISEVENT_ERROR:
            {
                char buf[1024];
                if(foundTarget)
                {
                    sprintf(buf,
                            "(MAC #%s, HBA #%ld): %s",
                            mactostring(event.MACAddr).c_str(),
                            event.host,
                            event.targetname);
                }
                else
                {
                    sprintf(buf,
                            "(MAC #%s, HBA #%ld): Target not found!",
                            mactostring(event.MACAddr).c_str(),
                            event.host);
                }

                switch(event.eventDetail)
                {
                case ISCSI_ISM_ERROR_NOP_FAIL:
                    syslog(LOG_WARNING,
                           "is1000-daemon: NOP Timeout %s",
                           buf);
                    break;

                case ISCSI_ISM_ERROR_TCP_FAIL:
                    syslog(LOG_WARNING,
                           "is1000-daemon: TCP Error %s",
                           buf);
                    break;

                case ISCSI_ISM_ERROR_TARGET_FAIL:
                    syslog(LOG_WARNING,
                           "is1000-daemon: Target Wants Tear Down %s",
                           buf);
                    break;

                case ISCSI_ISM_ERROR_HEAD_CRC_FAIL:
                    syslog(LOG_WARNING,
                           "is1000-daemon: Header CRC Failure %s",
                           buf);
                    break;

                case ISCSI_ISM_ERROR_DATA_CRC_FAIL:
                    syslog(LOG_WARNING,
                           "is1000-daemon: Data CRC Failure %s",
                           buf);
                    break;

                case ISCSI_ISM_ERROR_PROTOCOL_FAIL:
                    syslog(LOG_WARNING,
                           "is1000-daemon: Protocol Violation %s",
                           buf);
                    break;

                case ISCSI_ISM_ERROR_REJECT_PDU:
                    syslog(LOG_WARNING,
                           "is1000-daemon: Reject PDU Received %s",
                           buf);
                    break;

                case ISCSI_ISM_ERROR_NO_RESOURCE:
                    syslog(LOG_CRIT,
                           "is1000-daemon: Initiator Low Resource %s",
                           buf);
                    break;

                case ISCSI_ISM_ERROR_MAX_REDIRECT_EVENT:
                    syslog(LOG_WARNING,
                           "is1000-daemon: Maximum Login Redirects %s",
                           buf);
                    break;

                case ISCSI_ISM_ERROR_BAD_ITT:
                    syslog(LOG_WARNING,
                           "is1000-daemon: Bad ITT From Target %s",
                           buf);
                    break;

                case ISCSI_ISM_ERROR_DMA_FAIL:
                    syslog(LOG_CRIT,
                           "is1000-daemon: DMA Failure %s",
                           buf);
                    break;

                case ISCSI_ISM_ERROR_LOGOUT:
                    syslog(LOG_INFO,
                           "is1000-daemon: Logout Response %s",
                           buf);
                    break;

                case ISCSI_ISM_ERROR_BUS_RESET:
                    syslog(LOG_INFO,
                           "is1000-daemon: Host Request Bus Reset %s",
                           buf);
                    break;

                case ISCSI_ISM_ERROR_TIMEOUT:
                    syslog(LOG_WARNING,
                           "is1000-daemon: Target Timout %s",
                           buf);
                    break;

                default:
                    break;
                }

            }
            break;

            case ISEVENT_HOST_ADAPTER_DHCP_LEASE_NEW:
            {
                syslog(LOG_INFO,
                       "is1000-daemon: ADAPTER NEW DHCP LEASE: (MAC #%s, HBA #%ld)",
                        mactostring(event.MACAddr).c_str(), event.host);
            }
            break;

            case ISEVENT_HOST_ADAPTER_DHCP_LEASE_EXPIRE:
            {
                syslog(LOG_INFO,
                       "is1000-daemon: ADAPTER DHCP LEASE EXPIRED: (MAC #%s, HBA #%ld)",
                        mactostring(event.MACAddr).c_str(), event.host);
            }
            break;

            case ISEVENT_HOST_ADAPTER_DHCP_BIND_FAIL:
            {
                syslog(LOG_ALERT,
                       "is1000-daemon: ADAPTER DHCP BIND FAILED: (MAC #%s, HBA #%ld)",
                        mactostring(event.MACAddr).c_str(), event.host);
            }
            break;

            default:
            {
                syslog(LOG_ERR,
                       "is1000-daemon: received unsupported event = %d",
                       event.event);
            }
            break;
            }
        }  // else (numbytesread == sizeof(IS1000_EVENT))
    } // while(!process_should_exit())

    cleanup:
    close(fd);
    syslog(LOG_INFO, "is1000-daemon: shutting down.  rc = %d", rc);

    return rc;
}
