/*******************************************************************************
* Copyright (C) 2002-2003 Intel Corporation                                    *
*                                                                              *
* This program is free software; you can redistribute it and/or                *
* modify it under the terms of the GNU General Public License                  *
* as published by the Free Software Foundation; either version 2               *
* of the License, or (at your option) any later version.                       *
*                                                                              *
* This program is distributed in the hope that it will be useful,              *
* but WITHOUT ANY WARRANTY; without even the implied warranty of               *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                *
* GNU General Public License for more details.                                 *
*                                                                              *
* You should have received a copy of the GNU General Public License            *
* along with this program; if not, write to the Free Software                  *
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA   *
*******************************************************************************/
/*******************************************************************************
*                                                                              *
* File Name:    isimain.cpp                                                    *
*                                                                              *
* Abstract:     This file contains the main() function for the isiconfig       *
*               utility.  It hands off most commands to a more specialized     *
*               module after parsing the command-line options.                 *
*                                                                              *
* Environment:  This file should only be used by the isiconfig utility.        *
*                                                                              *
* Author:       Thomas C. Lea                                                  *
*                                                                              *
* Origination:  April 21, 2003                                                 *
*                                                                              *
* Modification History                                                         *
* Date     Initials   Description                                              *
* 04/21/03 TCL        Genesis                                                  *
*******************************************************************************/

// Include Files
#include <iostream>
#include "is1000api.h"
#include "utility.h"
#include <popt.h>
#include "adapter.h"
#include "target.h"

using namespace std;

void showSummary();

// Global definitions
    // top level commands
int adapter = 0;
int target = 0;
int generatename = 0;
int init = 0;
int version = 0;
int rescan = 0;
int help = 0;
    //common
int update = 0;
int adapternumber = 0;
char *ip = NULL;
int port = 0;
int timeout = IS_DEFAULT_TARGET_TIMEOUT_SECONDS;
int retries = 0;

struct poptOption topOptions[] =
{
    //Top level command options
    {
        "adapter",
        'a',
        POPT_ARG_NONE,
        &adapter,
        0,
        NULL,
        NULL
    },
    {
        "target",
        't',
        POPT_ARG_NONE,
        &target,
        0,
        NULL,
        NULL
    },
    {
        "generateuniquename",
        'g',
        POPT_ARG_NONE,
        &generatename,
        0,
        NULL,
        NULL
    },
    {
        "init",
        'i',
        POPT_ARG_NONE,
        &init,
        0,
        NULL,
        NULL
    },
    {
        NULL,
        '\0',
        POPT_ARG_INCLUDE_TABLE,
        targetOptions,
        0,
        NULL,
        NULL
    },
    {
        NULL,
        '\0',
        POPT_ARG_INCLUDE_TABLE,
        adapterOptions,
        0,
        NULL,
        NULL
    },
    {
        "version",
        'v',
        POPT_ARG_NONE,
        &version,
        0,
        NULL,
        NULL
    },
    {
        "rescan",
        '\0',
        POPT_ARG_NONE,
        &rescan,
        0,
        NULL,
        NULL
    },
    {
        "help",
        '?',
        POPT_ARG_NONE,
        &help,
        0,
        NULL,
        NULL
    },

    //Common options
    {
        "adapternumber",
        '\0',
        POPT_ARG_INT,
        &adapternumber,
        0,
        NULL,
        NULL
    },
    {
        "update",
        'u',
        POPT_ARG_NONE,
        &update,
        0,
        NULL,
        NULL
    },
    {
        "ip",
        '\0',
        POPT_ARG_STRING,
        &ip,
        0,
        NULL,
        NULL
    },
    {
        "timeout",
        '\0',
        POPT_ARG_INT,
        &timeout,
        0,
        NULL,
        NULL
    },
    {
        "retries",
        '\0',
        POPT_ARG_INT,
        &retries,
        0,
        NULL,
        NULL
    },

    POPT_TABLEEND
}; // struct poptOption topOptions[]

/** ****************************************************************************
 * Procedure:   showUsage
 *
 * Description: This routine displays the general usage/help
 *
 * @param   progname the name of the executable
 *
 ******************************************************************************/
void showUsage(const char *progname)
{
    cout << "Usage:" << endl;
    cout << progname << endl;
    cout << progname << " -a, --adapter\t\t[ADAPTER OPTIONS]" << endl;
    cout << progname << " -t, --target\t\t[TARGET OPTIONS]" << endl;
    cout << progname << " -g, --generateuniquename" << endl;
    cout << progname << " -v, --version" << endl;
    cout << progname << "     --rescan" << endl;
    cout << progname << " -?, --help" << endl;
    cout << endl;

    showAdapterUsage();
    showTargetUsage();
}

/** ****************************************************************************
 * Procedure:   main
 *
 * Description: This routine is the main entrypoint for isiconfig
 *
 * @param   argc the number of arguments in argv
 *
 * @param   argv the array of string arguments
 *
 * @return  zero on success, nonzero on failure
 *
 ******************************************************************************/
int main (
    int argc,
    const char **argv
)
{
    IS_RESULT libresult;

    poptContext pctx = poptGetContext("isiconfig",
        argc,
        argv,
        topOptions,
        0);

    //parse all options
    int rc = poptGetNextOpt(pctx);
    if(rc < -1)  //-1 is normal completion.  Other negatives are errors
    {
        cerr << "Invalid arguments. Error = " << rc << endl;
        poptFreeContext(pctx);
        return(rc);
    }

    const char *leftover;
    while((leftover = poptGetArg(pctx)) != NULL)
    {
        cerr << "invalid/ignored command-line argument: "  << leftover << endl;
    }

    poptFreeContext(pctx);

    if(help != 0)
    {
        showUsage(argv[0]);
        exit(-1);
    }

    if(retries < 0 || retries > 1000)
    {
        cerr << "Invalid retries value.  Range is 0-1000 (0 indicates infinite)." << endl;
        exit(-1);
    }

    if(timeout < 120 || timeout > 600)
    {
        cerr << "Invalid timeout value.  Range is 120-600." << endl;
        exit(-1);
    }

    if((libresult = InitializeLibrary(0)) != IS_SUCCESS)
    {
        if(libresult == IS_NO_DRIVER)
            cerr << "Unable to open the is1000 device driver, is it loaded?" << endl;
        else
            cerr << "InitilizeLibrary failed: " <<  RESULTSTRING(libresult) << endl;
        return(libresult);
    }

    if(rescan != 0)
    {
        updateDeviceHive();
        exit(0);
    }

    if(version != 0)
    {
        showVersion();
        exit(0);
    }

    if(adapter + target + generatename + init > 1)
    {
        cerr << "Invalid options (main)" << endl;
        exit(-1);
    }

    if(argc == 1)
    {
        showSummary();
    }
    else //one command specified, do it!
    {
        if(adapter)
            performAdapter();
        else if(target)
            performTarget();
        else if(generatename)
        {
            char name[IS_MAX_NODENAME_STRING];
            GenerateUniqueNodename(name, IS_MAX_NODENAME_STRING, 0);

            cout << name << endl;
        }
        else if(init)
        {
            libresult = InitializeDriver(0);
            if(libresult != IS_SUCCESS)
            {
                cerr << "Driver failed to initialize: " <<
                    RESULTSTRING(libresult) << endl;
                return libresult;
            }

            updateDeviceHive();
        }

    }

    return 0;
} //main(int argc, const char ** argv)
