/*******************************************************************************
* Copyright (C) 2002-2003 Intel Corporation                                    *
*                                                                              *
* This program is free software; you can redistribute it and/or                *
* modify it under the terms of the GNU General Public License                  *
* as published by the Free Software Foundation; either version 2               *
* of the License, or (at your option) any later version.                       *
*                                                                              *
* This program is distributed in the hope that it will be useful,              *
* but WITHOUT ANY WARRANTY; without even the implied warranty of               *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                *
* GNU General Public License for more details.                                 *
*                                                                              *
* You should have received a copy of the GNU General Public License            *
* along with this program; if not, write to the Free Software                  *
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA   *
*******************************************************************************/
/*******************************************************************************
*                                                                              *
* File Name:    summary.cpp                                                    *
*                                                                              *
* Abstract:     This module provides the summary display capability of         *
*               isiconfig.                                                     *
*                                                                              *
* Environment:  This file should only be used by isiconfig internals.          *
*                                                                              *
* Author:       Thomas C. Lea                                                  *
*                                                                              *
* Origination:  April 21, 2003                                                 *
*                                                                              *
* Modification History                                                         *
* Date     Initials   Description                                              *
* 04/21/03 TCL        Genesis                                                  *
*******************************************************************************/

// Include Files
#include <iostream>
#include <set>
#include <ftw.h>
#include <stdio.h>
#include "is1000api.h"
#include "utility.h"

using namespace std;

// Local private definitions
//  temps used for the ftw
static u8 currentHost;
static u8 currentId;
static u8 currentLun;

/** ****************************************************************************
 * Procedure:   dumpDeviceEntry
 *
 * Description: This routine displays information about an iSCSI device entry
 *
 * @param   file the full path and name of the device node
 *
 * @param   sb unused
 *
 * @return  zero on success, nonzero on failure
 *
 ******************************************************************************/
static int dumpDeviceEntry (
    const char *file,
    const struct stat *sb,
    int flag
)
{
    int h,i,l,p;
    char type;
    int result = 0;
    if((result = sscanf(file,
                        "/dev/is1000devs/s%ch%d-i%dl%dp%d",
                        &type,
                        &h,
                        &i,
                        &l,
                        &p)) > 0)
    {
        if(h == (int)currentHost &&
            i == (int)currentId &&
            l == (int)currentLun)
        {
            cout << "\t\t\t" << file;
            if(type == 'g')
            {
                cout << "\t[Generic]";
            }
            else if(type == 'd')
            {
                if(result == 5) //the 'p' value was read too
                    cout << "\t[Partition]";
                else
                    cout << "\t[Disk]";
            }
            cout << endl;
        }
    }

    return 0;
}

/** ****************************************************************************
 * Procedure:   dumpDevices
 *
 * Description: This routine walks the tree of iSCSI device nodes, calling
 *              dumpDeviceEntry for each node.
 *
 * @param   host the SCSI host number to display
 *
 * @param   id the SCSI id number to display
 *
 * @param   lun the SCSI lun number to display
 *
 ******************************************************************************/
void dumpDevices (
    u8 host,
    u8 id,
    u8 lun
)
{
    currentHost = host;
    currentId = id;
    currentLun = lun;
    ftw("/dev/is1000devs", dumpDeviceEntry, 1);
}

/** ****************************************************************************
 * Procedure:   showSummary
 *
 * Description: This routine is the main summary display entrypoint.  It builds
 *              and displays an appropriate amount of summary information to
 *              stdout.
 *
 ******************************************************************************/
void showSummary(void)
{
    cout << "Intel PRO/1000 T IP Storage Adapter Summary" << endl;
    u32 numadapters;
    IS_RESULT result = GetAdapterCount(&numadapters, 0);
    if(result != IS_SUCCESS)
    {
        cerr << "GetAdapterCount failed: " << RESULTSTRING(result) << endl;
        exit(result);
    }

    if(numadapters == 0)
    {
        cerr << "No adapters found!" << endl;
        exit(1);
    }
    else
    {
        for(u32 adapter_index = 0; adapter_index < numadapters; adapter_index++)
        {
            cout << endl << "Adapter #" << adapter_index;
            IS_MAC_ADDRESS tmp_mac;
            if((result = getAdapterMac(adapter_index, tmp_mac)) != IS_SUCCESS)
            {
                cerr << "Unable to get MAC address for adapter #" <<
                    adapter_index << endl;
                continue;
            }
            else
            {
                cout << ", MAC Address = " << mactostring(tmp_mac) << endl;
            }

            IS_ADAPTER_PROPS_WITH_DESC props;
            memset(&props, 0, sizeof(props));
            if((result = GetAdapterProperties(&tmp_mac, &props, 0)) != IS_SUCCESS)
            {
                cerr << "Unable to retrieve adapter properties.  Error = " <<
                    RESULTSTRING(result) << endl;
            }
            else
            {
                cout << "Name: " << props.props.s.iscsi_nodename << endl;
                cout << "Address: " <<
                    iptostring(props.props.s.AdapterIPAddress) << endl;
                cout << "Link: ";
                if(props.props.HasLink)
                {
                    cout << props.props.LinkSpeed << " Mbps ";
                    if(props.props.IsFullDuplex)
                        cout << "full-";
                    else
                        cout << "half-";
                    cout << "duplex" << endl;
                }
                else
                {
                    cout << "NONE" << endl;
                }
            }

            u32 tcount;
            if((result = GetTargetsWithLunsCount(&tmp_mac,
                                                 &tcount,
                                                 0)) != IS_SUCCESS)
            {
                cerr << "GetTargetsWithLunsCount failed: " <<
                    RESULTSTRING(result) << endl;
                exit(result);
            }

                // a set of target ids that we will use later to display all
                // the targets that dont have LUNs (that didnt get connected).
                set<u32> found_tids;

            if(tcount > 0)
            {
                cout << "Targets:" << endl;
                for(u32 targlun = 0; targlun < tcount; targlun++)
                {
                    IS_STORAGE_TARGET st;
                    memset(&st, 0, sizeof(st));

                    IS_SCSI_ID idlist[IS_MAX_LUNS];
                    memset(idlist, 0, sizeof(idlist));
                    u32 numids = IS_MAX_LUNS;
                    if((result = EnumTargetsWithLuns(&tmp_mac,
                                                     targlun,
                                                     &st,
                                                     idlist,
                                                     &numids,
                                                     0)) != IS_SUCCESS)
                    {
                        if(result == IS_TARGET_NOT_FOUND)
                        {
                            //this can happen if you remove a target before unmounting it
                            cerr << endl;
                            cerr << "ERROR: target appears to be mounted, but the configuration is missing!" << endl;
                            cerr << "       Was the target removed before it was unmounted?" << endl;
                            cerr << "       To recover make sure all targets are unmounted, then" << endl;
                            cerr << "       restart the is1000 service." << endl << endl;
                            continue;
                        }

                        cerr << "EnumTargetsWithLuns failed: " <<
                            RESULTSTRING(result) << endl;
                        exit(result);
                    }
                    found_tids.insert(st.targetId);
                    cout << "\t" << targlun << ": " << st.iscsi_nodename << endl;
                    cout << "\t\tstatus = ";
                    if(st.status >= 0 &&
                       st.status < LAST_AND_INVALID_TARGET_STATUS)
                    {
                        cout << TARGET_STATUS_TEXT[st.status];
                        if(st.status == TARGET_STATUS_RECOVERING)
                        {
                            //show recovery progress
                            u32 timeout;
                            u32 maxretries;
                            u32 currentretries;
                            IS_RESULT recovres = GetTargetRecoveryStatus(&tmp_mac,
                                st.address.ip,
                                st.address.port,
                                st.iscsi_nodename,
                                &timeout,
                                &maxretries,
                                &currentretries, 0);
                            if(recovres == IS_SUCCESS)
                            {
                                                  //infinite retries
                                if(maxretries == 0)
                                {
                                    cout << " infinite retries, ";
                                }
                                else
                                {
                                    cout << " retry " << currentretries << "/"
                                        << maxretries << ", ";
                                }

                                cout << timeout << " seconds between attempts.";
                            }
                        }

                        cout << endl;
                    }
                    else
                        cout << "UNKNOWN" << endl;

                    cout << "\t\tAddress = " << iptostring(st.address.ip) <<
                        ":" << st.address.port <<
                        "," << st.address.portalgroup << endl;

                    for(u32 devidx = 0; devidx < numids; devidx++)
                    {
                        cout << "\t\tSCSI (host,id,lun) = " <<
                            (int)idlist[devidx].host << ", " <<
                            (int)idlist[devidx].id << ", " <<
                            (int)idlist[devidx].lun << ":" << endl;

                        dumpDevices(idlist[devidx].host,
                                    idlist[devidx].id,
                                    idlist[devidx].lun);
                    }
                }
            }

            // now lets show all the other targets that werent returned via the
            // scan for targets that had LUNs
            u32 maxtcount;
            if((result = GetTargetsCount(&tmp_mac, &maxtcount, 0)) != IS_SUCCESS)
            {
                cerr << "GetTargetsCount failed: " << RESULTSTRING(result) << endl;
                exit(result);
            }

            // if we had more targets than we just reported, we have leftover
            // targets to report
            if(maxtcount > tcount)
            {
                u32 displayidx = tcount;
                for(u32 tidx = 0; tidx < maxtcount; tidx++)
                {
                    IS_STORAGE_TARGET st;
                    memset(&st, 0, sizeof(st));

                    if((result = EnumTargets(&tmp_mac,
                                             tidx,
                                             &st,
                                             0)) != IS_SUCCESS)
                    {
                        cerr << "EnumTargets failed: " <<
                            RESULTSTRING(result) << endl;
                        exit(result);
                    }

                    set<u32>::iterator thistid = found_tids.find(st.targetId);
                    
                    if(thistid == found_tids.end()) //wasnt in the set
                    {
                        cout << "\t" << displayidx++ << ": " <<
                            st.iscsi_nodename << endl;
                        cout << "\t\tstatus = ";
                        if(st.status >= 0 &&
                           st.status < LAST_AND_INVALID_TARGET_STATUS)
                        {
                            cout << TARGET_STATUS_TEXT[st.status] << endl;
                        }
                        else
                        {
                            cout << "UNKNOWN" << endl;
                        }

                        cout << "\t\tAddress = " << iptostring(st.address.ip) <<
                            ":" << st.address.port <<
                            "," << st.address.portalgroup << endl;
                    } //if(thistid == found_tids.end())
                } //for(u32 tidx = 0; tidx < maxtcount; tidx++)
            } //if(maxtcount > tcount)
        } //for(u32 adapter_index = 0; adapter_index < numadapters; adapter_index++)
    } //else numadapters != 0
} //void showSummary(void)
