/*******************************************************************************
* Copyright (C) 2002-2003 Intel Corporation                                    *
*                                                                              *
* This program is free software; you can redistribute it and/or                *
* modify it under the terms of the GNU General Public License                  *
* as published by the Free Software Foundation; either version 2               *
* of the License, or (at your option) any later version.                       *
*                                                                              *
* This program is distributed in the hope that it will be useful,              *
* but WITHOUT ANY WARRANTY; without even the implied warranty of               *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                *
* GNU General Public License for more details.                                 *
*                                                                              *
* You should have received a copy of the GNU General Public License            *
* along with this program; if not, write to the Free Software                  *
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA   *
*******************************************************************************/
/*******************************************************************************
*                                                                              *
* File Name:    utility.cpp                                                    *
*                                                                              *
* Abstract:     This module provides utility functions to the rest of          *
*               isiconfig.                                                     *
*                                                                              *
* Environment:  This file should be used only by the isiconfig internals.      *
*                                                                              *
* Author:       Thomas C. Lea                                                  *
*                                                                              *
* Origination:  April 21, 2003                                                 *
*                                                                              *
* Modification History                                                         *
* Date     Initials   Description                                              *
* 04/21/03 TCL        Genesis                                                  *
*******************************************************************************/

// Include Files
#include <iostream>
#include <fstream>
#include <stdio.h>
#include "utility.h"
#include <linux/limits.h>

using namespace std;

char errorStrings[IS_RESULT_CODE_COUNT][MAX_RESULT_STRING_LENGTH] =
{
    "SUCCESS",
    "NO DRIVER",
    "BAD CONFIGURATION",
    "INDEX OUT OF RANGE",
    "ADAPTER NOT FOUND",
    "BAD PARAMETER",
    "NO LINK",
    "BUFFER TOO SMALL",
    "DHCP IN PROGRESS",
    "DHCP FAILURE",
    "LOGIN IN PROGRESS",
    "LOGIN FAILURE",
    "TCP FAILURE",
    "CHAP AUTHENTICATION FAILURE",
    "INBOUND CHAP AUTHENTICATION FAILURE",
    "HARDWARE FAILURE",
    "OUT OF MEMORY",
    "TOO MANY TARGETS",
    "TARGET NOT FOUND",
    "OPERATION PENDING",
    "LIBRARY NOT INITIALIZED",
    "USER NOT AUTHORIZED",
    "NOT IMPLEMENTED",
    "CONFIGURATION ERROR",
    "SYNCHRONIZATION FAILURE",
    "IO ERROR",
    "GENERAL FAILURE",
    "OPERATION TIMEOUT"
};

/** ****************************************************************************
 * Procedure:   mactostring
 *
 * Description: This routine converts a IS_MAC_ADDRESS to a string.
 *
 * @param       MacAddr the MAC address to convert
 *
 * @return      the string representation
 *
 ******************************************************************************/
string mactostring(const IS_MAC_ADDRESS &MacAddr)
{
    if (MacAddr.first == 0 && MacAddr.second == 0)
        return "";

    char buf[13];
    sprintf(buf, "%02lX%02lX%02lX%02lX%02lX%02lX",
        MacAddr.first >> 0  & 0xFF,
        MacAddr.first >> 8  & 0xFF,
        MacAddr.first >> 16 & 0xFF,
        MacAddr.first >> 24 & 0xFF,
        MacAddr.second >> 0 & 0xFF,
        MacAddr.second >> 8 & 0xFF);
    return buf;
}

/** ****************************************************************************
 * Procedure:   iptostring
 *
 * Description: This routine converts a IP_TYPE to a string.
 *
 * @param       ip the IP address to convert
 *
 * @return      the string representation
 *
 ******************************************************************************/
string iptostring(const IP_TYPE &ip)
{
    char buf[16];
    sprintf(buf, "%ld.%ld.%ld.%ld",
        ip >> 24 & 0xFF,
        ip >> 16 & 0xFF,
        ip >> 8  & 0xFF,
        ip >> 0  & 0xFF);
    return buf;
}

/** ****************************************************************************
 * Procedure:   dumpLoginStats
 *
 * Description: This routine dumps a login statistics structure to stdout
 *
 * @param       stats the login stats to display
 ******************************************************************************/
void dumpLoginStats(IS_LOGIN_STATS &stats)
{
    cout << "Login Statistics:" << endl;
    cout << "\taccepts = " << stats.accepts << endl;
    cout << "\tfailResp = " << stats.failResp << endl;
    cout << "\tredirect = " << stats.redirect << endl;
    cout << "\tauthFailResp = " << stats.authFailResp << endl;
    cout << "\tauthFailTgt = " << stats.authFailTgt << endl;
    cout << "\tnegotiationFail = " << stats.negotiationFail << endl;
    cout << "\tlogoutGood = " << stats.logoutGood << endl;
    cout << "\tlogoutBad = " << stats.logoutBad << endl;
    cout << "\tloginFails = " << stats.loginFails << endl;
}

/** ****************************************************************************
 * Procedure:   dumpInitiatorStats
 *
 * Description: This routine dumps an initiator statistics structure to stdout
 *
 * @param       stats the initiator stats to display
 ******************************************************************************/
void dumpInitiatorStats(IS_INITIATOR_STATS &stats)
{
    cout << "Initiator Statistics:" << endl;
    cout << "\tdigestErrors = " << stats.digestErrors << endl;
    cout << "\tconnTimeout = " << stats.connTimeout << endl;
    cout << "\tformatErrors = " << stats.formatErrors << endl;
    cout << "\tsessFail = " << stats.sessFail << endl;
}

/** ****************************************************************************
 * Procedure:   dumpNicStats
 *
 * Description: This routine dumps a NIC statistics structure to stdout
 *
 * @param       stats the NIC stats to display
 ******************************************************************************/
void dumpNicStats(IS_NIC_STATS &stats)
{
    cout << "NIC Statistics:" << endl;
    cout << "\tbytesTx = " << stats.bytesTx << endl;
    cout << "\tbytesRx = " << stats.bytesRx << endl;
    cout << "\tpduTx = " << stats.pduTx << endl;
    cout << "\tpduRx = " << stats.pduRx << endl;
}

/** ****************************************************************************
 * Procedure:   dumpConnectionStats
 *
 * Description: This routine dumps a connection statistics structure to stdout
 *
 * @param       stats the connection stats to display
 ******************************************************************************/
void dumpConnectionStats(IS_CONNECTION_STATS &stats)
{
    cout << "Connection Statistics:" << endl;
    cout << "\tbytesSent = " << stats.bytesSent << endl;
    cout << "\tbytesReceived = " << stats.bytesReceived << endl;
    cout << "\tpduSent = " << stats.pduSent << endl;
    cout << "\tpduReceived = " << stats.pduReceived << endl;
}

/** ****************************************************************************
 * Procedure:   dumpExtSessionStats
 *
 * Description: This routine dumps an extended session statistics structure
 *              to stdout
 *
 * @param       stats the extended session stats to display
 ******************************************************************************/
void dumpExtSessionStats(IS_EXT_SESSION_STATS &stats)
{
    cout << "Extended Session Statistics:" << endl;
    cout << "\tbytesSent = " << stats.bytesSent << endl;
    cout << "\tbytesReceived = " << stats.bytesReceived << endl;
    cout << "\tpduSent = " << stats.pduSent << endl;
    cout << "\tpduReceived = " << stats.pduReceived << endl;
    cout << "\tdigestErrors = " << stats.digestErrors << endl;
    cout << "\tconnTimeouts = " << stats.connTimeouts << endl;
    cout << "\tformatErrors = " << stats.formatErrors << endl;
}

/** ****************************************************************************
 * Procedure:   getAdapterMac
 *
 * Description: This routine retrieves an adapter's MAC address
 *
 * @param       index the index of the adapter
 *
 * @param       mac the returned MAC address
 *
 * @return      the status code of the operation
 ******************************************************************************/
IS_RESULT getAdapterMac (
    u32 index,
    IS_MAC_ADDRESS &mac
)
{
    IS_RESULT libresult;
    u32 count;

    if((libresult = GetAdapterCount(&count, 0)) != IS_SUCCESS)
    {
        cerr << "GetAdapterCount failed: " << libresult << endl;
        return(libresult);
    }

    //is the specified number out of range?
    if(index < 0 || index >= count)
    {
        cerr << "The specified adapter (" << index << ") is out of range."
            << endl;
        return(IS_INDEX_OUT_OF_RANGE);
    }

    //we have a valid adapter index.  Get its MAC
    char name[256];
    if((libresult = EnumAdapters(index, &mac, name, 256, 0)) != IS_SUCCESS)
    {
        cerr << "EnumAdapters failed: " << (int)libresult << endl;
        return(libresult);
    }

    return libresult;
}

/** ****************************************************************************
 * Procedure:   stringtoip
 *
 * Description: This routine converts a string into an IP_TYPE
 *
 * @param       ipstr the string to convert
 *
 * @param       ip the result of the conversion if successful
 *
 * @return      true if successful, false otherwise
 ******************************************************************************/
bool stringtoip(const std::string ipstr, IP_TYPE &ip)
{
    u32 b1, b2, b3, b4;
    char dummy[2];
    if (sscanf(ipstr.c_str(),
               "%3lu.%3lu.%3lu.%3lu%1s",
               &b1,
               &b2,
               &b3,
               &b4,
               dummy) != 4)
    {
        return false;
    }
    if (b1 > 255 || b2 > 255 || b3 > 255 || b4 > 255)
    {
        return false;
    }

    ip = (b1 << 24) | (b2 << 16) | (b3 << 8) | b4;

    return true;
}

/** ****************************************************************************
 * Procedure:   stringtomac
 *
 * Description: This routine converts a string into an IS_MAC_ADDRESS
 *
 * @param       macstr the string to convert
 *
 * @param       mac the result of the conversion if successful
 *
 * @return      true if successful, false otherwise
 ******************************************************************************/
bool stringtomac(const std::string macstr, IS_MAC_ADDRESS &mac)
{
    if (macstr.size() == 0)
    {
        mac.first = mac.second = 0;
        return false;
    }

    u32 m1, m2, m3, m4, m5, m6;
    char dummy[2];

    if (sscanf(macstr.c_str(),
               "%02lx%02lx%02lx%02lx%02lx%02lx%1s",
               &m1,
               &m2,
               &m3,
               &m4,
               &m5,
               &m6,
               dummy) != 6
        &&
        sscanf(macstr.c_str(),
               "%02lx-%02lx-%02lx-%02lx-%02lx-%02lx%1s",
               &m1,
               &m2,
               &m3,
               &m4,
               &m5,
               &m6,
               dummy) != 6)
        return false;

    mac.first =  (m1 << 0  & 0x000000FF) |
        (m2 << 8  & 0x0000FF00) |
        (m3 << 16 & 0x00FF0000) |
        (m4 << 24 & 0xFF000000);
    mac.second = (m5 << 0  & 0x000000FF) |
        (m6 << 8  & 0x0000FF00);
    return true;
}

/** ****************************************************************************
 * Procedure:   showVersion
 *
 * Description: This routine displays the version of isiconfig, the api,
 *              the driver, the firmware, and the overall product.
 ******************************************************************************/
void showVersion(void)
{
    cout << "Intel PRO/1000 T IP Storage Adapter Configuration Utility Version "
        << VERSION << endl;

    IS_VERSION_INFO version_info;
    memset(&version_info, 0, sizeof(version_info));
    if(GetVersionInfo(&version_info, 0) != IS_SUCCESS)
    {
        cout << "Error retrieving full version status.  Some information may not be accurate." << endl;
    }

    cout << "Product Version: " << version_info.product_version << endl;
    cout << "API Version ";
    cout << ((version_info.api_version >> 24)&0xff) << "." <<
        ((version_info.api_version >> 16)&0xff) << "." <<
        ((version_info.api_version >> 8)&0xff) << "." <<
        ((version_info.api_version)&0xff) << endl;
    cout << "Driver Version ";
    cout << ((version_info.driver_version >> 24)&0xff) << "." <<
        ((version_info.driver_version >> 16)&0xff) << "." <<
        ((version_info.driver_version >> 8)&0xff) << "." <<
        ((version_info.driver_version)&0xff) << endl;

    for(u32 i = 0; i < version_info.adapter_count; i++)
    {
        cout << "Adapter #" << i << " using firmware version " <<
            ((version_info.adapter_firmware[i] >> 24)&0xff) << "." <<
            ((version_info.adapter_firmware[i] >> 16)&0xff) << "." <<
            ((version_info.adapter_firmware[i] >> 8)&0xff) << "." <<
            ((version_info.adapter_firmware[i])&0xff) << endl;
    }
}

/** ****************************************************************************
 * Procedure:   showDetails
 *
 * Description: This routine displays the details for a particular adapter
 *
 * @param       adapternumber the index of the adapter
 *
 * @param       mac the MAC address of the adapter
 ******************************************************************************/
void showDetails(int adapternumber, IS_MAC_ADDRESS &mac)
{
    IS_RESULT result = IS_SUCCESS;

    cout << "Details for adapter #" << adapternumber << ", MAC " <<
        mactostring(mac) << ":" << endl << endl;

    IS_ADAPTER_PROPS_WITH_DESC props;
    memset(&props, 0, sizeof(IS_ADAPTER_PROPS_WITH_DESC));
    if((result = GetAdapterProperties(&mac, &props, 0)) != IS_SUCCESS)
    {
        cerr << "Unable to retrieve adapter properties.  Error = "
            << result << endl;
    }
    else
    {
        cout << "Properties:" << endl;
        cout << "\tmanufacturer: " << props.manufacturer << endl;
        cout << "\tmodel: " << props.model << endl;
        cout << "\thw_version: " << props.hw_version << endl;
        cout << "\tASIC_version: " << props.ASIC_version << endl;
        cout << "\tfirmware_version: " << props.firmware_version << endl;
        cout << "\tos_name: " << props.os_name << endl;
        cout << "\tProcessorID: " << props.props.ProcessorID << endl;
        cout << "\tEthernetDeviceID: " << props.props.EthernetDeviceID << endl;
        cout << "\tEthernetRevision: " << props.props.EthernetRevision << endl;
        cout << "\tBoardRevisionNumber: " << props.props.BoardRevisionNumber << endl;
        cout << "\tCPLDRevisionNumber: " << props.props.CPLDRevisionNumber << endl;
        cout << "\tProcessorRevision: " << props.props.ProcessorRevision << endl;
        cout << "\tAdapterRTOSVersion: " << props.props.AdapterRTOSVersion << endl;
        cout << "\tHasLink: " << props.props.HasLink << endl;
        cout << "\tLinkSpeed: " << props.props.LinkSpeed << endl;
        cout << "\tIsFullDuplex: " << props.props.IsFullDuplex << endl;
        cout << "\tTransmissionPaused: " << props.props.TransmissionPaused << endl;
        cout << "\tEthernetInitialized: " << props.props.EthernetInitialized << endl;
        cout << "\tMaxTargets: " << props.props.MaxTargets << endl;
        cout << "\tStatus: " << props.props.Status << endl;
        cout << "\tIPAddress: " << iptostring(props.props.IPAddress) << endl;
        cout << "Configuration:" << endl;
        cout << "\tiscsi_nodename: " << props.props.s.iscsi_nodename << endl;
        cout << "\tAdapterIPAddress: " << iptostring(props.props.s.AdapterIPAddress) << endl;
        cout << "\tAdapterIPMask: " << iptostring(props.props.s.AdapterIPMask) << endl;
        cout << "\tAdapterGateway: " << iptostring(props.props.s.AdapterGateway) << endl;
        cout << "\tUseDhcp: " << props.props.s.UseDhcp << endl;
        cout << "\tFrameSize: " << props.props.s.FrameSize << endl;
        cout << "\tDefaultRetries: " << props.props.s.DefaultRetries << endl;
        cout << "\tDefaultTimeout: " << props.props.s.DefaultTimeout << endl;
        cout << "\tAutoNegEnabled: " << props.props.s.AutoNegEnabled << endl;
        cout << "\tDisallow10Half: " << props.props.s.Disallow10Half << endl;
        cout << "\tDisallow10Full: " << props.props.s.Disallow10Full << endl;
        cout << "\tDisallow100Half: " << props.props.s.Disallow100Half << endl;
        cout << "\tDisallow100Full: " << props.props.s.Disallow100Full << endl;
        cout << "\tDisallow1000Full: " << props.props.s.Disallow1000Full << endl;
        cout << "\tForcedLinkSpeed: " << props.props.s.ForcedLinkSpeed << endl;
        cout << "\tForcedFullDuplex: " << props.props.s.ForcedFullDuplex << endl;
        cout << "\tFlowControlEnabled: " << props.props.s.FlowControlEnabled << endl;
        cout << "\tReservedBits: " << props.props.s.ReservedBits << endl;
        cout << "\tLAA_Address: " << mactostring(props.props.s.LAA_Address) << endl;
    }

    IS_SESSION_HBA_STATS statistics;
    memset(&statistics, 0, sizeof(statistics));
    if((result = GetAdapterStatistics(&mac, &statistics, 0)) != IS_SUCCESS)
    {
        cerr << "Unable to retrieve statistics.  Error = " << result << endl;
    }
    else
    {
        dumpLoginStats(statistics.loginStats);
        dumpInitiatorStats(statistics.initStats);
        dumpNicStats(statistics.nicStats);
    }
} //showDetails()

extern "C"
{
    int builddevs(int *hbas, int numhbas);
}

/** ****************************************************************************
 * Procedure:   updateDeviceHive
 *
 * Description: This routine builds the /dev/is1000devs device hive to match
 *              the currently connected iSCSI devices.
 ******************************************************************************/
void updateDeviceHive()
{
    int hbas[IS_MAX_BOARDS];
    u32 numhbas;

    if((GetAdapterCount(&numhbas, 0)) != IS_SUCCESS)
        return;

    for(u32 i = 0; i < numhbas; i++)
    {
        IS_MAC_ADDRESS mac;
        char name[256];
        if((EnumAdapters(i, &mac, name, 256, 0)) != IS_SUCCESS)
            return;

        IS_ADAPTER_PROPS_WITH_DESC props;
        if((GetAdapterProperties(&mac, &props, 0)) != IS_SUCCESS)
            return;

        hbas[i] = props.scsi_host_number;
    }

    builddevs(hbas, numhbas);
}

// ok we have to enumerate all the targets until we find the one we are looking for
// then we can check to see if any matches for this target exist in /proc/mounts
/** ****************************************************************************
 * Procedure:   isTargetMounted
 *
 * Description: This routine enumerates all targets until we find the one we
 *              are looking for.  Then we check to see if this target exists
 *              in /proc/mounts.
 *
 * @param       mac the MAC address of the adapter to search
 *
 * @param       ip the IP address of the target we are checking on
 *
 * @param       port the port of the connection to the target
 *
 * @param       targetname the iSCSI nodename of the target
 *
 * @return      true if the target is mounted, false otherwise
 ******************************************************************************/
bool isTargetMounted (
    IS_MAC_ADDRESS *mac,
    IP_TYPE &ip,
    int port,
    char *targetname
)
{
    bool result = false;
    u32 targetcount = 0;
    IS_RESULT is;
    if((is = GetTargetsWithLunsCount(mac, &targetcount, 0)) != IS_SUCCESS)
    {
        cerr << "GetTargetsCount failed: " << RESULTSTRING(is) << endl;
        cerr << "Unable to determine if mounts exist for target" << endl;
    }
    else
    {
        for(u32 idx = 0; idx < targetcount; idx++)
        {
            IS_STORAGE_TARGET st;
            IS_SCSI_ID idlist[IS_MAX_LUNS];
            u32 numids = IS_MAX_LUNS;
            memset(&st, 0, sizeof(st));
            memset(idlist, 0, sizeof(idlist));
            if((is = EnumTargetsWithLuns(mac,
                                         idx,
                                         &st,
                                         idlist,
                                         &numids,
                                         0)) != IS_SUCCESS)
            {
                cerr << "EnumTargets failed: " << RESULTSTRING(is) << endl;
                cerr << "Unable to determine if mounts exist for target" << endl;
                break;
            }
            else if(numids == 0)
            {
                //this target has no luns -- it shouldnt have been returned
                break;
            }
            else
            {
                if(st.address.ip == ip &&
                    st.address.port == port &&
                    (strcmp(targetname, st.iscsi_nodename) == 0))
                {
                    //we found our man.  Are there any matching mounts?
                    ifstream in("/proc/mounts");
                    if(!in.is_open())
                    {
                        cerr << "Unable to open /proc/mounts." << endl;
                        cerr << "Unable to determine if mounts exist for target" << endl;
                    }
                    else
                    {
                        char targetdev[PATH_MAX];
                        sprintf(targetdev,
                                "/dev/is1000devs/sdh%d-i%d",
                                idlist[0].host,
                                idlist[0].id);

                        while(!in.eof())
                        {
                            //mount device + mount point + options
                            char buf[PATH_MAX * 2 + 255];
                            in.getline(buf, sizeof(buf));
                            if(!in.good() && !in.eof())
                            {
                                cerr << "Error reading /proc/mounts" << endl;
                                cerr << "Unable to determine if mounts exist for target" << endl;
                                break;
                            }

                            if(strncmp(targetdev, buf, strlen(targetdev)) == 0)
                            {
                                result = true;
                                break;
                            }
                        }

                        in.close();
                    }

                    break;
                }
            } //else
        } //for(u32 idx = 0; idx < targetcount; idx++)
    }//else

    return result;
} //isTargetMounted
