/*******************************************************************************
* Copyright (C) 2002-2003 Intel Corporation                                    *
*                                                                              *
* This program is free software; you can redistribute it and/or                *
* modify it under the terms of the GNU General Public License                  *
* as published by the Free Software Foundation; either version 2               *
* of the License, or (at your option) any later version.                       *
*                                                                              *
* This program is distributed in the hope that it will be useful,              *
* but WITHOUT ANY WARRANTY; without even the implied warranty of               *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                *
* GNU General Public License for more details.                                 *
*                                                                              *
* You should have received a copy of the GNU General Public License            *
* along with this program; if not, write to the Free Software                  *
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA   *
*******************************************************************************/
/*******************************************************************************
*                                                                              *
* File Name:    adapter.cpp                                                    *
*                                                                              *
* Abstract:     This module provides a class that represents an Intel PRO/1000 *
*               T IP Storage Adapter.  It is the root node in a tree of        *
*               device concepts.  It manages a list of Routers that, in turn,  *
*               manage a list of Targets.                                      *
*                                                                              *
* Environment:  This file should only be used by the library internals.        *
*                                                                              *
* Author:       Thomas C. Lea                                                  *
*                                                                              *
* Origination:  April 21, 2003                                                 *
*******************************************************************************/

// Include Files
#include <iostream>
#include <unistd.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include "is1000api.h"
#include "is1000ioctl.h"
#include "adapter.h"
#include "router.h"
#include "target.h"
#include "library.h"

using namespace std;

#define MAX_ENUMERATION_RETRIES 10

Adapter::Adapter()
{
    found_in_driver = false;
    login_complete = false;
    hostNumber = -1;
    memset(&props, 0, sizeof(props));
    mac.first = mac.second = 0;
    props.AdapterIPAddress = 0;
    props.AdapterIPMask = 0;
    props.AdapterGateway = 0;
    props.UseDhcp = false;
    props.FrameSize = ISCSI_FRAMESIZE_1500;
    props.DefaultRetries = 0;
    props.DefaultTimeout = IS_DEFAULT_TARGET_TIMEOUT_SECONDS;
    props.AutoNegEnabled = true;
    props.FlowControlEnabled = true;
    props.ForcedFullDuplex = true;
    props.ForcedLinkSpeed = 1000;
}

bool Adapter::is_valid()
{
    if (!(mac.first != 0 || mac.second != 0) &&
        props.AdapterIPAddress != 0 &&
        props.AdapterIPAddress != 0 &&
        strlen(props.iscsi_nodename) != 0)
        return false;

    for (ROUTER_LIST::iterator p = routers.begin(); p != routers.end(); ++p)
        if (!(*p)->is_valid())
            return false;

    return true;
}

IS_RESULT Adapter::getProperties(PIS_ADAPTER_PROPS_WITH_DESC pprops)
{
    IS_RESULT result = IS_SUCCESS;

    if(pprops == NULL)
        return IS_BAD_PARAMETER;

    if(is_valid())
    {
        ISCSI_GET_ADAPTER_PROPERTIES properties_request;
        memset(&properties_request,0,sizeof(properties_request));
        memcpy(&properties_request.MACAddr, &mac, sizeof(IS_MAC_ADDRESS));

        if((result = CallDriver(ISCSI_IOCTL_GET_ADAPTER_PROPERTIES,
                                &properties_request)) == IS_SUCCESS)
        {
            // is there another place to get this?
            sprintf(pprops->manufacturer,"Intel");
            sprintf(pprops->model,
                    "PRO/1000 T IP Storage Adapter HBA #%d",
                    hostNumber);

            sprintf(pprops->hw_version,
                    "%lx",
                    properties_request.props.BoardRevisionNumber);
            sprintf(pprops->ASIC_version,
                    "%lx",
                    properties_request.props.ProcessorRevision);
            sprintf(pprops->firmware_version,
                    "%ld.%ld.%ld.%ld",
                    (properties_request.props.AdapterFirmwareVersion>>24)&0xff,
                    (properties_request.props.AdapterFirmwareVersion>>16)&0xff,
                    (properties_request.props.AdapterFirmwareVersion>> 8)&0xff,
                    (properties_request.props.AdapterFirmwareVersion    )&0xff);

            sprintf(pprops->os_name, "Linux");

            memcpy(&(pprops->props),
                   &properties_request.props,
                   sizeof(IS_ADAPTER_PROPS));

            //fill in stuff that we store, but the driver does not.
            pprops->props.s.DefaultRetries = props.DefaultRetries;
            pprops->props.s.DefaultTimeout = props.DefaultTimeout;

            //fill in our host number
            pprops->scsi_host_number = hostNumber;
        }
    } // if(is_valid())
    else
    {
        result = IS_CONFIG_ERROR;
    }

    return result;
} //getProperties()

IS_RESULT Adapter::setProperties(PIS_ADAPTER_CONF_PROPS pprops)
{
    IS_RESULT result = IS_SUCCESS;

    if(pprops->FrameSize != 1500 && pprops->FrameSize != 9000)
        return IS_BAD_PARAMETER;

    memcpy(&props, pprops, sizeof(IS_ADAPTER_CONF_PROPS));

    result = syncPropsToDriver();

    return result;
}

IS_RESULT Adapter::getStatistics(PIS_SESSION_HBA_STATS pstats)
{
    IS_RESULT result = IS_SUCCESS;

    if(pstats == NULL)
        return IS_BAD_PARAMETER;

    if(is_valid())
    {
        ISCSI_ADAPTER_STATISTICS statistics_request;
        memset(&statistics_request,0,sizeof(statistics_request));
        memcpy(&statistics_request.MACAddr, &mac, sizeof(IS_MAC_ADDRESS));

        if((result = CallDriver(ISCSI_IOCTL_GET_STATISTICS,
                                &statistics_request)) == IS_SUCCESS)
        {
            memcpy(pstats,
                   &statistics_request.stats,
                   sizeof(IS_SESSION_HBA_STATS));
        }
    }
    else
    {
        result = IS_CONFIG_ERROR;
    }

    return result;
}

IS_RESULT Adapter::getDiagnostics(u32 *diag_value)
{
    IS_RESULT result = IS_SUCCESS;

    if(is_valid())
    {
        ISCSI_DIAGNOSTICS diag_request;
        memset(&diag_request, 0, sizeof(diag_request));
        memcpy(&diag_request.MACAddr, &mac, sizeof(IS_MAC_ADDRESS));

        if((result = CallDriver(ISCSI_IOCTL_GET_DIAGNOSTICS,
                                &diag_request)) == IS_SUCCESS)
        {
            *diag_value = diag_request.diagValue;
        }
    }
    else
    {
        result = IS_CONFIG_ERROR;
    }

    return result;
}

IS_RESULT Adapter::getTargetCount(u32 *target_count)
{
    IS_RESULT result = IS_SUCCESS;

    *target_count = 0;

    ROUTER_LIST::iterator i;
    for(i = routers.begin(); i != routers.end(); i++)
    {
        Router *pTmpRouter = *i;
        *target_count += pTmpRouter->targets.size();
    }

    return result;
}

IS_RESULT Adapter::getTarget(u32 index, PIS_STORAGE_TARGET st)
{
    u32 tcount = 0;

    //find which router has our target index
    ROUTER_LIST::iterator i;
    for(i = routers.begin(); i != routers.end(); i++)
    {
        Router *pTmpRouter = *i;
        u32 rcount = pTmpRouter->targets.size();

        if(tcount + rcount > index) // this router is our man!
        {
            //have to reset our index to work inside this router
            if(pTmpRouter->getTarget(index - tcount, st) == IS_SUCCESS)
            {
                //update the status field
                return IS_SUCCESS;
            }
        }
        else
        {
            tcount += rcount;
        }
    }

    return IS_INDEX_OUT_OF_RANGE;
}

IS_RESULT Adapter::getQueryTargetCount (
    u32 ip,
    u32 port,
    PIS_CHAP_SETTINGS pchap_settings,
    u32 *target_count
)
{
    IS_RESULT result = IS_SUCCESS;

    if(pchap_settings == NULL)
    {
        return IS_BAD_PARAMETER;
    }

    if(is_valid())
    {
        ISCSI_GET_STORAGE_TARGET_COUNT count_request;
        memset(&count_request, 0, sizeof(count_request));
        memcpy(&count_request.MACAddr, &mac, sizeof(IS_MAC_ADDRESS));
        count_request.IPAddress = ip;
        count_request.port = port;
        memcpy(&count_request.chap, pchap_settings, sizeof(IS_CHAP_SETTINGS));

        //first ping the address to see if its valid
        u32 ping_result = 0;
        if((result = ping(ip, &ping_result)) == IS_SUCCESS)
        {
            if(ping_result != 0)    //failed
                return IS_TCP_FAILURE;
        }
        else
        {
            return result;
        }

        if((result = CallDriver(ISCSI_IOCTL_GET_STORAGE_TARGET_COUNT,
                                &count_request)) == IS_SUCCESS)
        {
            IS_RESULT tmpstat = IS_SUCCESS;
            u32 retries = MAX_ENUMERATION_RETRIES;
            while(tmpstat == IS_SUCCESS &&
                count_request.status == ENUMERATION_STATUS_PENDING &&
                retries > 0)
            {
                //are we done with this one? if so, let's abort it.
                if (retries == 1)
                {
                    count_request.status = ENUMERATION_STATUS_ABORT;
                }
                else
                {
                    count_request.status = ENUMERATION_STATUS_POLL;
                }

                tmpstat = CallDriver(ISCSI_IOCTL_GET_STORAGE_TARGET_COUNT,
                                     &count_request);
                retries--;
                sleep(SECONDS_PER_POLL);
            }

            if(tmpstat != IS_SUCCESS)
            {
                // the call to the driver failed.  Return that code
                result = tmpstat;
            }
            else if (count_request.status == ENUMERATION_STATUS_SUCCESS)
            {
                *target_count = count_request.numTargets;
            }
            else  //handle all other returned errors
            {
                // strip off any iSCSI detail from upper word
                switch (count_request.status & 0xffff)
                {
                    //we never got a valid response
                    case ENUMERATION_STATUS_PENDING:
                        result = IS_OPERATION_TIMEOUT;
                        break;

                    case ENUMERATION_STATUS_ABORT:
                    case ENUMERATION_STATUS_FAILED:
                    case ENUMERATION_STATUS_MORE_MEMORY:
                        result = IS_GENERAL_FAILURE;
                        break;

                    // handle the special codes from the firmware
                    case ISCSI_ISM_ERROR_LOGIN_FAIL:
                        result = IS_LOGIN_FAILURE;
                        break;

                    case ISCSI_ISM_ERROR_AUTH_FAIL:
                        result = IS_AUTH_FAILURE;
                        break;

                    case ISCSI_ISM_ERROR_CONNECT_FAIL:
                        result = IS_TCP_FAILURE;
                        break;

                    case ISCSI_ISM_ERROR_MAX_REDIRECT:
                        result = IS_LOGIN_FAILURE;
                        break;

                    case ISCSI_ISM_ERROR_LOGIN_TCP_FAIL:
                        result = IS_TCP_FAILURE;
                        break;

                    case ISCSI_ISM_ERROR_LOGIN_BAD_STATE:
                        result = IS_LOGIN_FAILURE;
                        break;

                    case ISCSI_ISM_ERROR_LOGIN_BAD_PDU:
                        result = IS_LOGIN_FAILURE;
                        break;

                    case ISCSI_ISM_ERROR_LOGIN_BAD_USER:
                        result = IS_LOGIN_FAILURE;
                        break;

                    case ISCSI_ISM_ERROR_LOGIN_CANT_AGREE:
                        result = IS_LOGIN_FAILURE;
                        break;

                    default:
                        result = IS_GENERAL_FAILURE;
                        break;
                }
            }

            if (count_request.status == ENUMERATION_STATUS_PENDING)
            {
                result = IS_OPERATION_TIMEOUT;
            }
            else if (tmpstat != IS_SUCCESS ||
                count_request.status == ENUMERATION_STATUS_ABORT ||
                count_request.status == ENUMERATION_STATUS_FAILED ||
                count_request.status == ENUMERATION_STATUS_MORE_MEMORY)
            {
                //if we had a valid error, return it. Otherwise return a
                //generic error
                if(tmpstat == IS_SUCCESS)
                    result = IS_GENERAL_FAILURE;
                else
                    result = tmpstat;
            }
            else
            {
                *target_count = count_request.numTargets;
            }
        }
        else
        {
            //CallDriver failed....
            *target_count = 0;
        }
    } //if(is_valid())
    else
    {
        result = IS_CONFIG_ERROR;
    }

    return result;
} //getQueryTargetCount()

IS_RESULT Adapter::enumQueryTargets (
    u32 ip,
    u32 port,
    PIS_CHAP_SETTINGS pchap_settings,
    u32 index,
    char *targetname,
    u32 targetnameBufferLength,
    IS_TARGET_ADDRESS *addresses,
    u32 *numaddresses
)
{
    IS_RESULT result = IS_SUCCESS;

    if( pchap_settings == NULL ||
        targetname == NULL ||
        addresses == NULL ||
        numaddresses == NULL ||
        targetnameBufferLength == 0 ||
        *numaddresses == 0)
    {
        return IS_BAD_PARAMETER;
    }

    if(is_valid())
    {
        ISCSI_GET_STORAGE_TARGET target_request;
        memset(&target_request, 0, sizeof(target_request));
        memcpy(&target_request.MACAddr, &mac, sizeof(IS_MAC_ADDRESS));
        target_request.IPAddress = ip;
        target_request.port = port;
        target_request.targetIndex = index;

        if((result = CallDriver(ISCSI_IOCTL_GET_STORAGE_TARGET,
                                &target_request)) == IS_SUCCESS)
        {
            if(target_request.status == ENUMERATION_STATUS_SUCCESS)
            {
                //the number of addresses returned cant be greater
                //than *numaddresses
                u32 numreturned = (*numaddresses < target_request.numAddresses)?
                                    *numaddresses : target_request.numAddresses;
                *numaddresses = numreturned;
                for(u32 i = 0; i < numreturned; i++)
                {
                    memcpy(&addresses[i],
                           &target_request.target_addresses[i],
                           sizeof(IS_TARGET_ADDRESS));
                }

                u32 namelength = strlen(target_request.iscsi_nodename);
                if(targetnameBufferLength <= namelength)
                {
                    result = IS_BUFFER_TOO_SMALL;
                    //we are guaranteed that it is nonzero by above check
                    namelength = targetnameBufferLength - 1;
                }

                memcpy(targetname, target_request.iscsi_nodename, namelength);
                targetname[namelength] = '\0';
            }
            else //status failed
            {
                result = IS_GENERAL_FAILURE;
                *numaddresses = 0;
            }
        }
    } //if(is_valid())
    else
    {
        result = IS_CONFIG_ERROR;
    }

    return result;
}  //enumQueryTargets()

IS_RESULT Adapter::setTargetParameters (
    u32 ip,
    u32 port,
    const char *nodename,
    u32 timeoutSeconds,
    u32 retries
)
{
    IS_RESULT result = IS_SUCCESS;

    if(is_valid())
    {
        Router *pRouter = getRouter(ip, port);
        if(pRouter == NULL)
            return IS_TARGET_NOT_FOUND;

        Target *pTarget = NULL;
        if((result = pRouter->getTarget(nodename, &pTarget)) == IS_SUCCESS)
        {
            result = pRouter->setTargetParameters(nodename,
                                                  timeoutSeconds,
                                                  retries);
        }
        else
        {
            //fall through with result
        }
    } //if(is_valid())
    else
    {
        result = IS_CONFIG_ERROR;
    }

    return result;
} //setTargetParameters()

IS_RESULT Adapter::setTargetCRC (
    u32 ip,
    u32 port,
    const char *nodename,
    IS_CRC_MODE crcmode
)
{
    IS_RESULT result = IS_SUCCESS;

    if(is_valid())
    {
        Router *pRouter = getRouter(ip, port);
        if(pRouter == NULL)
            return IS_TARGET_NOT_FOUND;

        Target *pTarget = NULL;
        if((result = pRouter->getTarget(nodename, &pTarget)) == IS_SUCCESS)
        {
            pTarget->st.crcMode = crcmode;
        }
        else
        {
            //fall through with result
        }
    }  //if(is_valid())
    else
    {
        result = IS_CONFIG_ERROR;
    }

    return result;
} //setTargetCRC()

IS_RESULT Adapter::setTargetSecurity (
    u32 ip,
    u32 port,
    const char *nodename,
    PIS_CHAP_SETTINGS pchap_settings
)
{
    IS_RESULT result = IS_SUCCESS;

    if(is_valid())
    {
        Router *pRouter = getRouter(ip,port);
        if(pRouter == NULL)
        {
            //if we didnt find the router, we arent gonna find the target
            return IS_TARGET_NOT_FOUND;           
        }

        Target *pTarget = NULL;
        if((result = pRouter->getTarget(nodename, &pTarget)) == IS_SUCCESS)
        {
            result = pRouter->setTargetSecurity(nodename, pchap_settings);
        }
        else
        {
            //fall through with result
        }
    } //if(is_valid())
    else
    {
        result = IS_CONFIG_ERROR;
    }

    return result;
} //setTargetSecurity()

IS_RESULT Adapter::addTarget (
    u32 ip,
    u32 port,
    PIS_CHAP_SETTINGS pchap_settings,
    const char *nodename,
    IS_CRC_MODE crcmode
)
{
    IS_RESULT result = IS_SUCCESS;

    if(is_valid())
    {
        Router *pRouter = getRouter(ip,port);
        //do we already have this router object?
        if(pRouter == NULL) //we didnt find the router.  create it.
        {
            pRouter = new Router;
            if(pRouter == NULL)
                return IS_OUT_OF_MEMORY;
            pRouter->ip = ip;
            pRouter->port = port;
            routers.push_back(pRouter);
        }

        Target *pTarget = NULL;
        result = pRouter->getTarget(nodename, &pTarget);
        if(result == IS_SUCCESS)
        {
            cerr << "Target " << nodename << " is already configured (" <<
                ((ip >> 24) & 0xFF) << "." <<
                ((ip >> 16) & 0xFF) << "." <<
                ((ip >>  8) & 0xFF) << "." <<
                ((ip >>  0) & 0xFF) <<
                ":" << port << "," << pTarget->st.address.portalgroup <<
                ")." << endl;
        }
        else if(result == IS_TARGET_NOT_FOUND)    //create it.
        {
            result = pRouter->addTarget(nodename, pchap_settings,
                                        props.DefaultTimeout,
                                        props.DefaultRetries,
                                        mac,
                                        hostNumber,
                                        crcmode);
            if(result == IS_SUCCESS)
            {
                if((result = pRouter->getTarget(nodename, &pTarget)) != IS_SUCCESS)
                {
                    cerr << "Unable to get newly added target for configuration." << endl;
                }
                else
                {
                    pTarget->mapLUNS();
                }
            }
        }
        else
        {
            //fall through with result
        }
    } //if(is_valid())
    else
    {
        result = IS_CONFIG_ERROR;
    }

    return result;
} //addTarget()

IS_RESULT Adapter::getTargetsWithLunsCount(u32 *target_count)
{
    IS_RESULT result = IS_SUCCESS;
    int previousTID = -1;

    *target_count = 0;

    if(is_valid())
    {
        ifstream in("/proc/scsi/scsi");
        if (!in.is_open())
            return IS_IO_ERROR;

        while (!in.eof())
        {
            char buf[1024];
            in.getline(buf, sizeof(buf));

            if (!in.good() && !in.eof())
                return IS_IO_ERROR;

            int hostnum = -1;
            int tid = -1;
            int lun = -1;
            sscanf(buf,
                   "Host: scsi%d Channel: 00 Id: %d Lun: %d",
                   &hostnum, &tid, &lun);
            if(hostnum == hostNumber && tid != previousTID)
            {
                (*target_count)++;
                previousTID = tid;
            }
        }

        in.close();
    } //if(is_valid())
    else
    {
        result = IS_CONFIG_ERROR;
    }

    return result;
} //getTargetsWithLunsCount()

Target *Adapter::getTargetByTID(int tid)
{
    Target *pTarget = NULL;

    ROUTER_LIST::iterator i;
    for(i = routers.begin(); i != routers.end(); i++)
    {
        Router *pRouter = *i;
        pTarget = pRouter->getTargetByTID(tid);
        if(pTarget != NULL)                       //we found it
            break;
    }

    return pTarget;
} //getTargetByTID()

IS_RESULT Adapter::enumTargetsWithLuns (
    u32 index,
    PIS_STORAGE_TARGET st,
    IS_SCSI_ID *devices,
    u32 *numdevices
)
{
    IS_RESULT result = IS_SUCCESS;
    u32 maxdevices = *numdevices;
    Target *pTarget = NULL;
    u32 target_count = 0;
    int previousTID = -1;
    *numdevices = 0;

    if(is_valid())
    {
        ifstream in("/proc/scsi/scsi");
        if (!in.is_open())
            return IS_IO_ERROR;

        while (!in.eof())
        {
            char buf[1024];
            in.getline(buf, sizeof(buf));

            if (!in.good() && !in.eof())
                return IS_IO_ERROR;

            int hostnum = -1;
            int tid = -1;
            int lun = -1;
            sscanf(buf,
                   "Host: scsi%d Channel: 00 Id: %d Lun: %d",
                   &hostnum,
                   &tid,
                   &lun);
                   
            if(hostnum == hostNumber)
            {
                if(tid != previousTID)
                    target_count++;

                //does this entry belong to the target we are looking for?
                if(index == (target_count-1))
                {
                    if(pTarget == NULL)//this is our first entry for this target
                    {
                        pTarget = getTargetByTID(tid);

                        if(pTarget == NULL)
                        {
                            result = IS_TARGET_NOT_FOUND;
                            break;
                        }
                        else
                        {
                            memcpy(st, &pTarget->st, sizeof(IS_STORAGE_TARGET));
                        }
                    }

                    // if we are already full and we dont have enough room for
                    // more let the caller know their buffer was too small
                    if(*numdevices == maxdevices)
                    {
                        result = IS_BUFFER_TOO_SMALL;
                        break;
                    }

                    devices[*numdevices].host = hostNumber;
                    devices[*numdevices].id = tid;
                    devices[*numdevices].lun = lun;
                    (*numdevices)++;
                }

                previousTID = tid;
            }
        }

        if(index >= target_count) // the index was out of range
            result = IS_INDEX_OUT_OF_RANGE;

        in.close();

    } //if(is_valid())
    else
    {
        result = IS_CONFIG_ERROR;
    }

    return result;
} //enumTargetsWithLuns()

IS_RESULT Adapter::blink(int seconds)
{
    IS_RESULT result = IS_SUCCESS;

    if(is_valid())
    {
        ISCSI_BLINK_LED blink_request;
        memcpy(&blink_request.MACAddr, &mac, sizeof(IS_MAC_ADDRESS));
        blink_request.iSeconds = seconds;
        blink_request.dwStatus = 0;

        result = CallDriver(ISCSI_IOCTL_BLINK_LED, &blink_request);
    } //if(is_valid())
    else
    {
        result = IS_CONFIG_ERROR;
    }

    return result;
} //blink()

IS_RESULT Adapter::syncPropsToDriver()
{
    IS_RESULT result = IS_SUCCESS;

    if(is_valid() && found_in_driver)
    {
        ISCSI_SET_ADAPTER_PROPERTIES set_prop_request;
        memset(&set_prop_request, 0, sizeof(set_prop_request));
        memcpy(&set_prop_request.MACAddr, &mac, sizeof(IS_MAC_ADDRESS));
        memcpy(&set_prop_request.props, &props, sizeof(IS_ADAPTER_CONF_PROPS));

        if((result = CallDriver(ISCSI_IOCTL_SET_ADAPTER_PROPERTIES,
                                &set_prop_request)) == IS_SUCCESS)
        {
            if(set_prop_request.dwStatus == ISCSI_ADAPTER_CONFIG_FAILED)
                result = IS_TCP_FAILURE;
            else if(set_prop_request.dwStatus == ISCSI_ADAPTER_CONFIG_DHCP_FAILED)
                result = IS_DHCP_FAILURE;
        }
    }
    else
    {
        return IS_CONFIG_ERROR;
    }

    return result;
} //syncPropsToDriver()

IS_RESULT Adapter::syncToDriver()
{
    IS_RESULT result = IS_SUCCESS;

    if(is_valid() && found_in_driver)
    {
        result = syncPropsToDriver();

        if(result != IS_SUCCESS)
            return result;

        //let each router add itself
        ROUTER_LIST::iterator i;
        for(i = routers.begin(); i != routers.end(); i++)
        {
            Router *pRouter = *i;
            IS_RESULT tmpresult = pRouter->syncToDriver();
            if(tmpresult != IS_SUCCESS)
                result = tmpresult;
        }
    }
    else
    {
        return IS_CONFIG_ERROR;
    }

    return result;
} //syncToDriver()

IS_RESULT Adapter::initiateLogins()
{
    IS_RESULT result = IS_SUCCESS;

    if(found_in_driver)
    {
        //let each router initiate its logins
        ROUTER_LIST::iterator i;
        for(i = routers.begin(); i != routers.end(); i++)
        {
            Router *pRouter = *i;
            result = pRouter->initiateLogins();
        }
    }
    else
    {
        // no need to continue with logins on this adapter.
        result = IS_CONFIG_ERROR;
        login_complete = true; 
    }

    return result;
} //initiateLogins()

IS_RESULT Adapter::loginPoll()
{
    IS_RESULT result = IS_SUCCESS;
    IS_RESULT tmpresult = IS_SUCCESS;
    int pendingRouters = 0;

    if(found_in_driver)
    {
        ROUTER_LIST::iterator i;
        for(i = routers.begin(); i != routers.end(); i++)
        {
            Router *pRouter = *i;
            if(pRouter->login_complete)
                continue;

            tmpresult = pRouter->loginPoll();
            if(tmpresult == IS_LOGIN_IN_PROGRESS)
                pendingRouters++;
            else if(tmpresult != IS_SUCCESS && tmpresult != IS_LOGIN_FAILURE)
            {
                cerr << "Router failed to complete logins: " <<
                    tmpresult << endl;
                login_complete = true;            // dont continue;
                result = tmpresult;
                break;
            }
        }
    }
    else
    {
        // no need to continue with logins on this adapter, actually we shouldnt be here.
        result = IS_CONFIG_ERROR;
        login_complete = true;
    }

    if(pendingRouters == 0) // all done
        login_complete = true;

    return result;
} //loginPoll()

IS_RESULT Adapter::mapLUNS()
{
    IS_RESULT result = IS_SUCCESS;
    if(found_in_driver)
    {
        ROUTER_LIST::iterator i;
        for(i = routers.begin(); i != routers.end(); i++)
        {
            Router *pRouter = *i;
            pRouter->mapLUNS();
        }
    }

    return result;
} //mapLUNS()

IS_RESULT Adapter::removeTarget (
    u32 ip,
    u32 port,
    const char *nodename
)
{
    IS_RESULT result = IS_SUCCESS;

    if(is_valid())
    {
        Router *pRouter = NULL;

        //cant use getRouter() because we need the iterator for the
        //remove later...
        ROUTER_LIST::iterator i;
        for(i = routers.begin(); i != routers.end(); i++)
        {
            Router *pTmpRouter = *i;
            if(pTmpRouter->ip == ip &&
                pTmpRouter->port == port)
            {
                pRouter = pTmpRouter;
                break;
            }
        }

        if(pRouter == NULL)
        {
            // didnt find the router, so we couldnt find the target either.
            return IS_TARGET_NOT_FOUND;
        }

        result = pRouter->removeTarget(nodename);
        if(result == IS_SUCCESS)
        {
            //was that the last target in the router? if so, delete the
            //router too.
            if(pRouter->targets.empty())
            {
                //iterator position from the above search
                routers.erase(i);
            }
        }
        else
        {
            //fall through with result
        }
    }
    else
    {
        result = IS_CONFIG_ERROR;
    }

    return result;
}

IS_RESULT Adapter::ping(u32 ip, u32 *ping_result)
{
    IS_RESULT result = IS_SUCCESS;

    if(is_valid())
    {
        ISCSI_PING ping_request;
        memset(&ping_request, 0, sizeof(ping_request));
        memcpy(&ping_request.MACAddr, &mac, sizeof(IS_MAC_ADDRESS));
        ping_request.IPAddress = ip;

        if((result = CallDriver(ISCSI_IOCTL_PING, &ping_request)) == IS_SUCCESS)
        {
            *ping_result = ping_request.dwStatus;
        }
    }
    else
    {
        result = IS_CONFIG_ERROR;
    }

    return result;
} //ping()

IS_RESULT Adapter::loopback(u8 *pinbuffer, u8 *poutbuffer, u32 buffsize)
{
    IS_RESULT result = IS_SUCCESS;

    if(is_valid())
    {
        ISCSI_LOOPBACK loopback_request;
        memset(&loopback_request, 0, sizeof(loopback_request));
        memcpy(&loopback_request.MACAddr, &mac, sizeof(IS_MAC_ADDRESS));
        memcpy(loopback_request.buffer, pinbuffer, buffsize);
        loopback_request.buffsize = buffsize;

        if((result = CallDriver(ISCSI_IOCTL_LOOPBACK,
                                &loopback_request)) == IS_SUCCESS)
        {
            memcpy(poutbuffer, loopback_request.buffer, buffsize);
        }
    } //if(is_valid())
    else
    {
        result = IS_CONFIG_ERROR;
    }

    return result;
} //loopback()

IS_RESULT Adapter::resetTargetConnection (
    u32 ip,
    u32 port,
    const char *nodename
)
{
    IS_RESULT result = IS_SUCCESS;

    if(is_valid())
    {
        Router *pRouter = getRouter(ip,port);
        if(pRouter == NULL)
        {
            //if we didnt find the router, we arent gonna find the target
            return IS_TARGET_NOT_FOUND;
        }

        Target *pTarget = NULL;
        if((result = pRouter->getTarget(nodename, &pTarget)) == IS_SUCCESS)
        {
            if(pTarget == NULL)
                return IS_TARGET_NOT_FOUND;

            result = pTarget->resetConnection();

            //we only need to map the luns if we are connected....
            if(result == IS_SUCCESS &&
               pTarget->st.status == TARGET_STATUS_CONNECTED)
                pTarget->mapLUNS();
        }
        else
        {
            //fall through with result
        }
    } //if(is_valid())
    else
    {
        result = IS_CONFIG_ERROR;
    }

    return result;
} //resetTargetConnection

IS_RESULT Adapter::getTargetRecoveryStatus (
    u32 ip,
    u32 port,
    const char *nodename,
    u32 *timeout,
    u32 *maxretries,
    u32 *currentretries
)
{
    IS_RESULT result = IS_SUCCESS;

    if(is_valid())
    {
        Router *pRouter = getRouter(ip,port);
        if(pRouter == NULL)
        {
            //if we didnt find the router, we arent gonna find the target
            return IS_TARGET_NOT_FOUND;
        }

        Target *pTarget = NULL;
        if((result = pRouter->getTarget(nodename, &pTarget)) == IS_SUCCESS)
        {
            if(pTarget == NULL)
                return IS_TARGET_NOT_FOUND;

            result = pTarget->getRecoveryStatus(timeout,
                                                maxretries,
                                                currentretries);
        }
        else
        {
            //fall through with result
        }
    } //if(is_valid())
    else
    {
        result = IS_CONFIG_ERROR;
    }

    return result;
} //getTargetRecoveryStatus

IS_RESULT Adapter::getTargetStatistics (
    u32 ip,
    u32 port,
    const char *nodename,
    PIS_SESSION_STATS pstats)
{
    IS_RESULT result = IS_SUCCESS;

    if(is_valid())
    {
        Router *pRouter = getRouter(ip,port);
        if(pRouter == NULL)
        {
            //if we didnt find the router, we arent gonna find the target
            return IS_TARGET_NOT_FOUND;
        }

        Target *pTarget = NULL;
        if((result = pRouter->getTarget(nodename, &pTarget)) == IS_SUCCESS)
        {
            if(pTarget == NULL)
                return IS_TARGET_NOT_FOUND;

            result = pTarget->getStatistics(pstats);
        }
        else
        {
            //fall through with result
        }
    } //if(is_valid())
    else
    {
        result = IS_CONFIG_ERROR;
    }

    return result;
} //getTargetStatistics()

IS_RESULT Adapter::updateFlash(PIS_FLASH_UPDATE_PACKET pfup)
{
    IS_RESULT result = IS_SUCCESS;

    ISCSI_FLASH_UPDATE_PACKET flash_request;
    memset(&flash_request, 0, sizeof(flash_request));
    memcpy(&flash_request.MACAddr, &mac, sizeof(IS_MAC_ADDRESS));
    memcpy(&flash_request.packet, pfup, sizeof(IS_FLASH_UPDATE_PACKET));

    result = CallDriver(ISCSI_IOCTL_UPDATE_FLASH, &flash_request);

    return result;
} //updateFlash()

IS_RESULT Adapter::getEEPROMData(u8 *pbuffer, u32 *pbuffer_size)
{
    IS_RESULT result = IS_SUCCESS;

    if(pbuffer == NULL || pbuffer_size == NULL || *pbuffer_size < IS_EEPROM_DATA_SIZE)
    {
        return IS_BAD_PARAMETER;
    }

    ISCSI_GET_EEPROM_DATA eeprom_request;
    memset(&eeprom_request, 0, sizeof(eeprom_request));
    memcpy(&eeprom_request.MACAddr, &mac, sizeof(IS_MAC_ADDRESS));
    eeprom_request.timeOut = 10*1000;             //10 seconds
    eeprom_request.bufferSize = IS_EEPROM_DATA_SIZE;
    if((result = CallDriver(ISCSI_IOCTL_GET_EEPROM_INFO, &eeprom_request)) == IS_SUCCESS)
    {
        memcpy(pbuffer, eeprom_request.dataBuffer, IS_EEPROM_DATA_SIZE);
        *pbuffer_size = IS_EEPROM_DATA_SIZE;
    }
    else
    {
        *pbuffer_size = 0;
    }

    return result;
} //getEEPROMData()

IS_RESULT Adapter::clearTargetStatus()
{
    IS_RESULT result = IS_SUCCESS;

    ROUTER_LIST::iterator i;
    for(i = routers.begin(); i != routers.end(); i++)
    {
        Router *pRouter = *i;
        pRouter->clearTargetStatus();
    }

    return result;
} //clearTargetStatus()

IS_RESULT Adapter::setTargetStatus (
    u32 ip,
    u32 port,
    const char *nodename,
    TARGET_STATUS newStatus
)
{
    Router *pRouter = getRouter(ip, port);
    if(pRouter != NULL)
        return pRouter->setTargetStatus(nodename, newStatus);
    else
        return IS_TARGET_NOT_FOUND;
} //setTargetStatus()

IS_RESULT Adapter::getTargetStatus (
    u32 ip,
    u32 port,
    const char *nodename,
    TARGET_STATUS *targetStatus)
{
    Router *pRouter = getRouter(ip,port);
    if(pRouter != NULL)
        return pRouter->getTargetStatus(nodename, targetStatus);
    else
        return IS_TARGET_NOT_FOUND;
}

void Adapter::setHostNumber(int hn)
{
    hostNumber = hn;

    //let the children cache it
    ROUTER_LIST::iterator i;
    for(i = routers.begin(); i != routers.end(); i++)
    {
        Router *pRouter = *i;
        pRouter->setHostNumber(hn);
    }
} //setHostNumber()

Router *Adapter::getRouter(u32 ip, u32 port)
{
    Router *pRouter = NULL;

    ROUTER_LIST::iterator i;
    for(i = routers.begin(); i != routers.end(); i++)
    {
        Router *pTmpRouter = *i;
        if(pTmpRouter->ip == ip &&
            pTmpRouter->port == port)
        {
            pRouter = pTmpRouter;
            break;
        }
    }

    return pRouter;
}

IS_RESULT Adapter::targetPropertiesChanged(u32 ip, u32 port, const char *nodename)
{
    Router *pRouter = getRouter(ip,port);
    if(pRouter == NULL)
        return IS_TARGET_NOT_FOUND;
    else
        return pRouter->targetPropertiesChanged(nodename);
}
