/*******************************************************************************
* Copyright (C) 2002-2003 Intel Corporation                                    *
*                                                                              *
* This program is free software; you can redistribute it and/or                *
* modify it under the terms of the GNU General Public License                  *
* as published by the Free Software Foundation; either version 2               *
* of the License, or (at your option) any later version.                       *
*                                                                              *
* This program is distributed in the hope that it will be useful,              *
* but WITHOUT ANY WARRANTY; without even the implied warranty of               *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                *
* GNU General Public License for more details.                                 *
*                                                                              *
* You should have received a copy of the GNU General Public License            *
* along with this program; if not, write to the Free Software                  *
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA   *
*******************************************************************************/
/*******************************************************************************
*                                                                              *
* File Name:    adapter.h                                                      *
*                                                                              *
* Abstract:     This file describes the interface to the Adapter class that    *
*               represents an Intel PRO/1000 T IP Storage Adapter.             *
*                                                                              *
* Environment:  This file should only be used by the library internals.        *
*                                                                              *
* Author:       Thomas C. Lea                                                  *
*                                                                              *
* Origination:  April 21, 2003                                                 *
*******************************************************************************/

#ifndef __ADAPTER_H__
#define __ADAPTER_H__

//Include Files
#include "is1000api.h"
#include "library.h"

class Adapter
{
    public:
/** ****************************************************************************
 * Procedure:   Constructor
 *
 * Description: This is the constructor for the Adapter class
 ******************************************************************************/
        Adapter();

/** ****************************************************************************
 * Procedure:   is_valid
 *
 * Description: This routine checks the validity of the adapter instance
 *
 * @return  true if this instance is valid, false otherwise
 ******************************************************************************/
        bool is_valid();

/** ****************************************************************************
 * Procedure:   blink
 *
 * Description: This routine causes this adapter to blink a LED for
 *              identification and diagnostics purposes.
 *
 * @param   seconds the number of seconds to blink the LED
 *
 * @return  IS_SUCCESS if successful or other codes on error
 * @return  IS_SUCCESS if successful or other codes on error
 * @return  IS_SUCCESS if successful or other codes on error
 ******************************************************************************/
        IS_RESULT blink(int seconds);

/** ****************************************************************************
 * Procedure:   getProperties
 *
 * Description: This routine retrieves the properties of an adapter
 *
 * @param   pprops a pointer to a structure that will receive the data
 *
 * @return  IS_SUCCESS if successful or other codes on error
 ******************************************************************************/
        IS_RESULT getProperties(PIS_ADAPTER_PROPS_WITH_DESC pprops);

/** ****************************************************************************
 * Procedure:   setProperties
 *
 * Description: This routine sets the properties of an adapter
 *
 * @param   pprops a pointer to a structure that contans the new data
 *
 * @return  IS_SUCCESS if successful or other codes on error
 ******************************************************************************/
        IS_RESULT setProperties(PIS_ADAPTER_CONF_PROPS pprops);

/** ****************************************************************************
 * Procedure:   targetPropertiesChanged
 *
 * Description: This routine is used by the isdaemon to notify the target that
 *              its properties have changed.
 *
 * @param   ip the address of the target
 *
 * @param   port the port of the target
 *
 * @param   nodename the name of the target
 *
 * @return  IS_SUCCESS if successful or other codes on error
 ******************************************************************************/
        IS_RESULT targetPropertiesChanged(u32 ip,
                                          u32 port,
                                          const char *nodename);
        
/** ****************************************************************************
 * Procedure:   syncToDriver
 *
 * Description: This routine causes a heirarchical synchronization of properties
 *              and configuration from this adapter down to any targets it may
 *              have with the driver and firmware.
 *
 * @return  IS_SUCCESS if successful or other codes on error
 ******************************************************************************/
        IS_RESULT syncToDriver();

/** ****************************************************************************
 * Procedure:   initiateLogins
 *
 * Description: This routine causes a heirarchical target login of all targets
 *              managed by this adapter.  After this call, loginPoll is used
 *              to check on the progress of the logins.
 *
 * @return  IS_SUCCESS if successful or other codes on error
 ******************************************************************************/
        IS_RESULT initiateLogins();

/** ****************************************************************************
 * Procedure:   loginPoll
 *
 * Description: This routine perfoms one complete login poll on all routers
 *
 * @return  IS_SUCCESS if successful or other codes on error
 ******************************************************************************/
        IS_RESULT loginPoll();

/** ****************************************************************************
 * Procedure:   mapLUNS
 *
 * Description: This routine instructs each managed router to map its LUNs
 *
 * @return  IS_SUCCESS if successful or other codes on error
 ******************************************************************************/
        IS_RESULT mapLUNS();

/** ****************************************************************************
 * Procedure:   getStatistics
 *
 * Description: This routine retrieves the statistics of an adapter
 *
 * @param   pstats a pointer to a structure that will receive the data
 *
 * @return  IS_SUCCESS if successful or other codes on error
 ******************************************************************************/
        IS_RESULT getStatistics(PIS_SESSION_HBA_STATS pstats);

/** ****************************************************************************
 * Procedure:   getDiagnostics
 *
 * Description: This routine retrieves an adapter's current diagnostics value
 *
 * @param   diag_value a pointer to a u32 that will receive the value
 *
 * @return  IS_SUCCESS if successful or other codes on error
 ******************************************************************************/
        IS_RESULT getDiagnostics(u32 *diag_value);

/** ****************************************************************************
 * Procedure:   getTargetCount
 *
 * Description: This routine retrieves the number ot targets managed by an
 *              adapter
 *
 * @param   target_count a pointer to a u32 that will receive the value
 *
 * @return  IS_SUCCESS if successful or other codes on error
 ******************************************************************************/
        IS_RESULT getTargetCount(u32 *target_count);

/** ****************************************************************************
 * Procedure:   getTarget
 *
 * Description: This routine retrieves the specified target structure
 *
 * @param   index the index of the desired target
 *
 * @param   st a pointer to the structure that is to receive the data
 *
 * @return  IS_SUCCESS if successful or other codes on error
 ******************************************************************************/
        IS_RESULT getTarget(u32 index,
                            PIS_STORAGE_TARGET st);

/** ****************************************************************************
 * Procedure:   getQueryTargetCount
 *
 * Description: This routine performs an iSCSI SEND TARGETS operation and
 *              returns the number of targets that were reported.  As a side-
 *              effect, the driver caches the list of returned targets for later
 *              enumeration.Each time this method is called, a new SEND TARGETS
 *              request gets sent which clears the driver's cached list of
 *              targets.
 *
 * @param   ip the address of the portal to query
 *
 * @param   port the port of the portal to query
 *
 * @param   pchap_settings the CHAP configuration to use for the query
 *
 * @param   target_count on success, will contain the number of discovered
 *          targets
 *
 * @return  IS_SUCCESS if successful or other codes on error
 ******************************************************************************/
        IS_RESULT getQueryTargetCount(u32 ip,
                                      u32 port,
                                      PIS_CHAP_SETTINGS pchap_settings,
                                      u32 *target_count);
        
/** ****************************************************************************
 * Procedure:   enumQueryTargets
 *
 * Description: This routine is used to enumerate the targets discovered via
 *              a call to getQueryTargetsCount().
 *
 * @param   ip the address of the portal used in the query
 *
 * @param   port the port of the portal used in the query
 *
 * @param   pchap_settings the CHAP configuration used for the query
 *
 * @param   index the index of the desired target
 *
 * @param   targetname a pointer to a buffer that will receive the name of the
 *          target
 *
 * @param   targetnameBufferLength the maximum number of bytes that can be
 *          stored in targetname
 *
 * @param   addresses a pointer to an allocated array of IS_TARGET_ADDRESS
 *          structures that will receive the addresses that are associated
 *          with the selected target
 *
 * @param   numaddresses on input identifies how many IS_TARGET_ADDRESS
 *          structures were allocated in the addresses paramter.  On return,
 *          it provides the caller the number of addresses returned
 *
 * @return  IS_SUCCESS if successful or other codes on error
 ******************************************************************************/
        IS_RESULT enumQueryTargets(u32 ip,
                                   u32 port,
                                   PIS_CHAP_SETTINGS pchap_settings,
                                   u32 index,
                                   char *targetname, u32 targetnameBufferLength,
                                   IS_TARGET_ADDRESS *addresses, u32 *numaddresses);

/** ****************************************************************************
 * Procedure:   setTargetParameters
 *
 * Description: This routine updates a target's timeout and retry values.
 *
 * @param   ip the address of the target
 *
 * @param   port the port of the target
 *
 * @param   nodename the name of the target
 *
 * @param   timeoutSeconds the number of seconds to wait between retries
 *
 * @param   retries the number of times to retry a connection recovery
 *
 * @return  IS_SUCCESS if successful or other codes on error
 ******************************************************************************/
        IS_RESULT setTargetParameters(u32 ip,
                                      u32 port,
                                      const char *nodename,
                                      u32 timeoutSeconds, u32 retries);

/** ****************************************************************************
 * Procedure:   setTargetCRC
 *
 * Description: This routine updates a target's CRC mode
 *
 * @param   ip the address of the target
 *
 * @param   port the port of the target
 *
 * @param   nodename the name of the target
 *
 * @param   crcmode the new CRC mode to use for the target
 *
 * @return  IS_SUCCESS if successful or other codes on error
 ******************************************************************************/
        IS_RESULT setTargetCRC(u32 ip,
                               u32 port,
                               const char *nodename,
                               IS_CRC_MODE crcmode);

/** ****************************************************************************
 * Procedure:   setTargetSecurity
 *
 * Description: This routine updates a target's CHAP configuration
 *
 * @param   ip the address of the target
 *
 * @param   port the port of the target
 *
 * @param   nodename the name of the target
 *
 * @param   pchap_settings the new CHAP configuration to use
 *
 * @return  IS_SUCCESS if successful or other codes on error
 ******************************************************************************/
        IS_RESULT setTargetSecurity(u32 ip,
                                    u32 port,
                                    const char *nodename,
                                    PIS_CHAP_SETTINGS pchap_settings);

/** ****************************************************************************
 * Procedure:   addTarget
 *
 * Description: This routine adds a new target and establishes a connection
 *
 * @param   ip the address of the target
 *
 * @param   port the port of the target
 *
 * @param   pchap_settings the new CHAP configuration to use
 *
 * @param   nodename the name of the target
 *
 * @param   crcmode the CRC mode to use for the target
 *
 * @return  IS_SUCCESS if successful or other codes on error
 ******************************************************************************/
        IS_RESULT addTarget(u32 ip,
                            u32 port,
                            PIS_CHAP_SETTINGS pchap_settings,
                            const char *nodename,
                            IS_CRC_MODE crcmode);

/** ****************************************************************************
 * Procedure:   getTargetsWithLunsCount
 *
 * Description: This routine returns the number of targets that are currently
 *              providing SCSI LUNs that are associated with this adapter
 *
 * @param   target_count on success will contain the result
 *
 * @return  IS_SUCCESS if successful or other codes on error
 ******************************************************************************/
        IS_RESULT getTargetsWithLunsCount(u32 *target_count);

/** ****************************************************************************
 * Procedure:   enumTargetsWithLuns
 *
 * Description: This routine returns the selected target that is providing
 *              SCSI LUNs.
 *
 * @param   index the index of the desired target
 *
 * @param   st a pointer to a IS_STORAGE_TARGET structure to receive the details
 *
 * @param   devices a pointer to an allocated array of IS_SCSI_ID structures
 *          which will be populated with SCSI device details about the selected
 *          target
 *
 * @param   numdevices on input it contains the number of allocated IS_SCSI_ID
 *          structures pointed to by devices.  On output it containes the number
 *          of returned IS_SCSI_ID structures.
 *
 * @return  IS_SUCCESS if successful or other codes on error
 ******************************************************************************/
        IS_RESULT enumTargetsWithLuns(u32 index,
                                      PIS_STORAGE_TARGET st,
                                      IS_SCSI_ID *devices,
                                      u32 *numdevices);
        
/** ****************************************************************************
 * Procedure:   removeTarget
 *
 * Description: This routine removes a target and tears down its connection
 *
 * @param   ip the address of the target
 *
 * @param   port the port of the target
 *
 * @param   nodename the name of the target
 *
 * @return  IS_SUCCESS if successful or other codes on error
 ******************************************************************************/
        IS_RESULT removeTarget(u32 ip,
                               u32 port,
                               const char *nodename);

/** ****************************************************************************
 * Procedure:   resetTargetConnection
 *
 * Description: This routine causes a target's connection to be reset
 *
 * @param   ip the address of the target
 *
 * @param   port the port of the target
 *
 * @param   nodename the name of the target
 *
 * @return  IS_SUCCESS if successful or other codes on error
 ******************************************************************************/
        IS_RESULT resetTargetConnection(u32 ip,
                                        u32 port,
                                        const char *nodename);

/** ****************************************************************************
 * Procedure:   getTargetRecoveryStatus
 *
 * Description: This routine retrieves the current recovery status for a target
 *
 * @param   ip the address of the target
 *
 * @param   port the port of the target
 *
 * @param   nodename the name of the target
 *
 * @param   timeout on success contains the timeout setting for the target
 *
 * @param   maxretries on success contains the retries setting for the target
 *
 * @param   currentretries on success contains the number of times this target's
 *          connection has been retried
 *
 * @return  IS_SUCCESS if successful or other codes on error
 ******************************************************************************/
        IS_RESULT getTargetRecoveryStatus(u32 ip,
                                          u32 port,
                                          const char *nodename,
                                          u32 *timeout,
                                          u32 *maxretries,
                                          u32 *currentretries);

/** ****************************************************************************
 * Procedure:   getTargetStatistics
 *
 * Description: This routine retrieves the statistics for the specified target
 *
 * @param   ip the address of the target
 *
 * @param   port the port of the target
 *
 * @param   nodename the name of the target
 *
 * @param   pstats on success contains the statistics for the target
 *
 * @return  IS_SUCCESS if successful or other codes on error
 ******************************************************************************/
        IS_RESULT getTargetStatistics(u32 ip,
                                      u32 port,
                                      const char *nodename,
                                      PIS_SESSION_STATS pstats);
        
/** ****************************************************************************
 * Procedure:   ping
 *
 * Description: This routine performs an ICMP echo operation from this adapter
 *              to the specified IP address.
 *
 * @param   ip the address of the target
 *
 * @param   ping_result on success will contain the result of the ping
 *          operation.  Zero indicates success, nonzero for failure.
 *
 * @return  IS_SUCCESS if successful or other codes on error
 ******************************************************************************/
        IS_RESULT ping(u32 ip,
                       u32 *ping_result);

/** ****************************************************************************
 * Procedure:   loopback
 *
 * Description: This routine performs a diagnostic loopback operation where
 *              a buffer of data is sent to the firmware and then echoed back.
 *              the caller would typically compare the input and output buffers
 *              to check that the datapath is functional and error free.
 *
 * @param   pinbuffer the array of bytes to send via loopback
 *
 * @param   poutbuffer the echoed bytes
 *
 * @param   buffsize the number of bytes in the input and output buffers
 *
 * @return  IS_SUCCESS if successful or other codes on error
 ******************************************************************************/
        IS_RESULT loopback(u8 *pinbuffer,
                           u8 *poutbuffer,
                           u32 buffsize);

/** ****************************************************************************
 * Procedure:   updateFlash
 *
 * Description: This routine uploads one packet of flash data to the firmware
 *
 * @param   pfup the packet of flash data
 *
 * @return  IS_SUCCESS if successful or other codes on error
 ******************************************************************************/
        IS_RESULT updateFlash(PIS_FLASH_UPDATE_PACKET pfup);

/** ****************************************************************************
 * Procedure:   getEEPROMData
 *
 * Description: This routine retrieves the contents of the adapter's EEPROM
 *
 * @param   pbuffer on success contains the EEPROM data
 *
 * @param   pbuffer_size on input contains the allocated size of pbuffer, and
 *          on output contains the number of bytes written to pbuffer.
 *
 * @return  IS_SUCCESS if successful or other codes on error
 ******************************************************************************/
        IS_RESULT getEEPROMData(u8 *pbuffer,
                                u32 *pbuffer_size);

/** ****************************************************************************
 * Procedure:   setTargetStatus
 *
 * Description: This routine sets the status of the specified target
 *
 * @param   ip the address of the target
 *
 * @param   port the port of the target
 *
 * @param   nodename the name of the target
 *
 * @param   newStatus the status for the target
 *
 * @return  IS_SUCCESS if successful or other codes on error
 ******************************************************************************/
        IS_RESULT setTargetStatus(u32 ip,
                                  u32 port,
                                  const char *nodename,
                                  TARGET_STATUS newStatus);

/** ****************************************************************************
 * Procedure:   getTargetStatus
 *
 * Description: This routine retrieves the status of the specified target
 *
 * @param   ip the address of the target
 *
 * @param   port the port of the target
 *
 * @param   nodename the name of the target
 *
 * @param   targetStatus the status of the target
 *
 * @return  IS_SUCCESS if successful or other codes on error
 ******************************************************************************/
        IS_RESULT getTargetStatus(u32 ip,
                                  u32 port,
                                  const char *nodename,
                                  TARGET_STATUS *targetStatus);

/** ****************************************************************************
 * Procedure:   clearTargetStatus
 *
 * Description: This routine marks all managed targets with an uninitialized
 *              status.  Primarily used during initialization.
 *
 * @return  IS_SUCCESS if successful or other codes on error
 ******************************************************************************/
        IS_RESULT clearTargetStatus();

/** ****************************************************************************
 * Procedure:   setHostNumber
 *
 * Description: This routine sets the SCSI HBA number for this adapter.  The
 *              HBA number is selected by the kernel.
 *
 * @param   hn the SCSI host number
 *
 * @return  IS_SUCCESS if successful or other codes on error
 ******************************************************************************/
        void         setHostNumber(int hn);

        // Public data members
        IS_MAC_ADDRESS mac;
        IS_ADAPTER_CONF_PROPS props;
        ROUTER_LIST routers;
        bool found_in_driver;
        bool login_complete;
        int hostNumber;

    private:
/** ****************************************************************************
 * Procedure:   syncPropsToDriver
 *
 * Description: This routine updates the driver and firmware with the current
 *              view of adapter properties
 *
 * @return  IS_SUCCESS if successful or other codes on error
 ******************************************************************************/
        IS_RESULT syncPropsToDriver();

/** ****************************************************************************
 * Procedure:   getRouter
 *
 * Description: This routine retrieves a specific router object
 *
 * @param   ip the IP address of the router
 *
 * @param   port the port number of the router
 *
 * @return  IS_SUCCESS if successful or other codes on error
 ******************************************************************************/
        Router *getRouter(u32 ip,
                          u32 port);

/** ****************************************************************************
 * Procedure:   getTargetByTID
 *
 * Description: This routine returns the target associated with the provided
 *              target id.
 *
 * @param   tid the target id of the desired target
 *
 * @return  IS_SUCCESS if successful or other codes on error
 ******************************************************************************/
        Target *getTargetByTID(int tid);
};

#endif // __ADAPTER_H__
