/*******************************************************************************
* Copyright (C) 2002-2003 Intel Corporation                                    *
*                                                                              *
* This program is free software; you can redistribute it and/or                *
* modify it under the terms of the GNU General Public License                  *
* as published by the Free Software Foundation; either version 2               *
* of the License, or (at your option) any later version.                       *
*                                                                              *
* This program is distributed in the hope that it will be useful,              *
* but WITHOUT ANY WARRANTY; without even the implied warranty of               *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                *
* GNU General Public License for more details.                                 *
*                                                                              *
* You should have received a copy of the GNU General Public License            *
* along with this program; if not, write to the Free Software                  *
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA   *
*******************************************************************************/
/*******************************************************************************
*                                                                              *
* File Name:    inteliscsi.cpp                                                 *
*                                                                              *
* Abstract:     This module provides the entry points to the shared library.   *
*               See is1000api.h for function documentation.                    *
*                                                                              *
* Environment:  This file should only be used by the library internals.        *
*                                                                              *
* Author:       Thomas C. Lea                                                  *
*                                                                              *
* Origination:  April 21, 2003                                                 *
*******************************************************************************/

//Include Files
#include <iostream>
#include <unistd.h>
#include <sys/types.h>
#include <sys/utsname.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/time.h>
#include "is1000api.h"
#include "library.h"
#include "adapter.h"
#include "md5.h"

// Local Private Definitions
static bool is_initialized = false;

// Global Definitions
LibraryHelper libhelper;

// Local Prototypes
bool OpenDriver();
bool CloseDriver();

char TARGET_STATUS_TEXT[LAST_AND_INVALID_TARGET_STATUS][IS_MAX_TARGET_STATUS_STRING] =
{
    "INVALID",
    "SYNC REQUIRED",
    "LOGIN REQUIRED",
    "LOGIN IN PROGRESS",
    "LOGIN FAILURE",
    "CONNECTED",
    "RECOVERING",
    "CONNECTION LOST"
};

/** ****************************************************************************
 * Procedure:   isAuthorized
 *
 * Description: This routine checks to make sure that the calling user has
 *              authorization to use the library.
 *
 * @return  true if the calling user is authorized, false otherwise
 ******************************************************************************/
bool isAuthorized()
{
#ifndef ALLOW_ANY_USER
    return (getuid() == 0);
#else
    return true;
#endif
}

IS_RESULT InitializeLibrary(u64 reserved)
{
    if(!isAuthorized())
        return IS_USER_NOT_AUTH;

    //check to see if the driver is loaded
    if(OpenDriver())
        CloseDriver();
    else
        return IS_NO_DRIVER;

    //TODO: there is a race condition here
    if(libhelper.lib == NULL)
    {
        libhelper.lib = new Library();
        if(libhelper.lib == NULL)
            return IS_OUT_OF_MEMORY;
    }

    IS_RESULT is = libhelper.lib->Lock();

    if (is == IS_SUCCESS)
    {
        is_initialized = true;
        is = libhelper.lib->Unlock();
    }

    return is;
} //InitializeLibrary()

IS_RESULT InitializeDriver(u64 reserved)
{
    if(!isAuthorized())
        return IS_USER_NOT_AUTH;

    if (!is_initialized)
        return IS_LIB_NOT_INIT;

    IS_RESULT is = libhelper.lib->Lock();

    if (is == IS_SUCCESS)
    {
        is = libhelper.lib->InitializeDriver();

        //unlock failures take precendence
        IS_RESULT unlockresult = libhelper.lib->Unlock();
        if(unlockresult != IS_SUCCESS)
            is = unlockresult;
    }

    return is;
} //InitializeDriver()

IS_RESULT GetVersionInfo (
    PIS_VERSION_INFO version_info,
    u64 reserved
)
{
    if(!isAuthorized())
        return IS_USER_NOT_AUTH;

    if (!is_initialized)
        return IS_LIB_NOT_INIT;

    if(version_info == NULL)
        return IS_BAD_PARAMETER;

    IS_RESULT is = libhelper.lib->Lock();

    if (is == IS_SUCCESS)
    {
        memset(version_info, 0, sizeof(IS_VERSION_INFO));
        version_info->api_version = IS_VERSION;

        is = libhelper.lib->GetDriverVersion(&version_info->driver_version);
        version_info->adapter_count = libhelper.lib->GetAdapterCount();
        for(u32 i = 0; i < version_info->adapter_count; i++)
        {
            IS_ADAPTER_PROPS_WITH_DESC props;

            Adapter *pAdapter = libhelper.lib->GetAdapter(i);
            if(pAdapter != NULL)
            {
                is = pAdapter->getProperties(&props);
                if(is == IS_SUCCESS)
                    version_info->adapter_firmware[i] = props.props.AdapterFirmwareVersion;
            }
        }

        char tic[256];
        libhelper.lib->GetTicNumber(tic);

        sprintf(version_info->product_version,
                "%d.%d.%s",
                (int)(IS_VERSION >> 24),
                (int)(IS_VERSION&0xff0000 >> 16),
                tic);

        //unlock failures take precendence
        IS_RESULT unlockresult = libhelper.lib->Unlock();
        if(unlockresult != IS_SUCCESS)
            is = unlockresult;
    }

    return is;
}  //GetVersionInfo

IS_RESULT GetAdapterCount (
    u32 *adapter_count,
    u64 reserved
)
{
    if(!isAuthorized())
        return IS_USER_NOT_AUTH;

    if (!is_initialized)
        return IS_LIB_NOT_INIT;

    if(adapter_count == NULL)
        return IS_BAD_PARAMETER;

    IS_RESULT is = libhelper.lib->Lock();

    if (is == IS_SUCCESS)
    {
        *adapter_count = libhelper.lib->GetAdapterCount();
        is = libhelper.lib->Unlock();
    }

    return is;
} //GetAdapterCount

IS_RESULT EnumAdapters (
    u32 index,
    IS_MAC_ADDRESS *mac,
    char *friendly_name,
    u32 name_buf_size,
    u64 reserved
)
{
    if(!isAuthorized())
        return IS_USER_NOT_AUTH;

    if (!is_initialized)
        return IS_LIB_NOT_INIT;

    if(mac == NULL || friendly_name == NULL)
        return IS_BAD_PARAMETER;

    if(index >= IS_MAX_BOARDS)
        return IS_INDEX_OUT_OF_RANGE;

    IS_RESULT is = libhelper.lib->Lock();

    if(is == IS_SUCCESS)
    {
        Adapter *a = libhelper.lib->GetAdapter(index);

        if(a == NULL)
        {
            is = IS_INDEX_OUT_OF_RANGE;
        }
        else
        {
            mac->first = a->mac.first;
            mac->second = a->mac.second;
            char namebuf[64];
            sprintf(namebuf,
                    "Intel(R) PRO/1000 T IP Storage Adapter #%d",
                    (int)index);
            if(name_buf_size > 1 && friendly_name != NULL)
            {
                strncpy(friendly_name,namebuf, name_buf_size-1);
                friendly_name[name_buf_size-1] = '\0';
            }

            //report an error if the input buffer size is too small
            // the returned buffer will be truncated
            if(name_buf_size <= strlen(namebuf))
                is = IS_BUFFER_TOO_SMALL;
        }

        //unlock failures take precendence
        IS_RESULT unlockresult = libhelper.lib->Unlock();
        if(unlockresult != IS_SUCCESS)
            is = unlockresult;
    }

    return is;
} //EnumAdapters

IS_RESULT GetAdapterProperties (
    const IS_MAC_ADDRESS *mac,
    PIS_ADAPTER_PROPS_WITH_DESC pprops,
    u64 reserved)
{
    if(!isAuthorized())
        return IS_USER_NOT_AUTH;

    if (!is_initialized)
        return IS_LIB_NOT_INIT;

    if(mac == NULL || pprops == NULL)
        return IS_BAD_PARAMETER;

    IS_RESULT is = libhelper.lib->Lock();

    if(is == IS_SUCCESS)
    {
        Adapter *a = libhelper.lib->GetAdapter(mac);
        if(a == NULL)
        {
            is = IS_ADAPTER_NOT_FOUND;
        }
        else
        {
            is = a->getProperties(pprops);
        }

        //unlock failures take precendence
        IS_RESULT unlockresult = libhelper.lib->Unlock();
        if(unlockresult != IS_SUCCESS)
            is = unlockresult;
    }

    return is;
} //GetAdapterProperties()

IS_RESULT SetAdapterProperties (
    const IS_MAC_ADDRESS *mac,
    PIS_ADAPTER_CONF_PROPS pprops,
    u64 reserved
)
{
    if(!isAuthorized())
        return IS_USER_NOT_AUTH;

    if (!is_initialized)
        return IS_LIB_NOT_INIT;

    if(mac == NULL || pprops == NULL)
        return IS_BAD_PARAMETER;

    IS_RESULT is = libhelper.lib->Lock();

    if(is == IS_SUCCESS)
    {
        Adapter *a = libhelper.lib->GetAdapter(mac);
        if(a == NULL)
        {
            is = IS_ADAPTER_NOT_FOUND;
        }
        else
        {
            is = a->setProperties(pprops);
            libhelper.lib->is_modified = true;
        }

        //unlock failures take precendence
        IS_RESULT unlockresult = libhelper.lib->Unlock();
        if(unlockresult != IS_SUCCESS)
            is = unlockresult;
    }

    return is;
} //SetAdapterProperties()

IS_RESULT GetAdapterStatistics (
    const IS_MAC_ADDRESS *mac,
    PIS_SESSION_HBA_STATS pstats,
    u64 reserved)
{
    if(!isAuthorized())
        return IS_USER_NOT_AUTH;

    if (!is_initialized)
        return IS_LIB_NOT_INIT;

    if(mac == NULL || pstats == NULL)
        return IS_BAD_PARAMETER;

    IS_RESULT is = libhelper.lib->Lock();

    if(is == IS_SUCCESS)
    {
        Adapter *a = libhelper.lib->GetAdapter(mac);
        if(a == NULL)
        {
            is = IS_ADAPTER_NOT_FOUND;
        }
        else
        {
            is = a->getStatistics(pstats);
        }

        //unlock failures take precendence
        IS_RESULT unlockresult = libhelper.lib->Unlock();
        if(unlockresult != IS_SUCCESS)
            is = unlockresult;
    }

    return is;
} //GetAdapterStatistics()

IS_RESULT GetAdapterDiagnostics (
    const IS_MAC_ADDRESS *mac,
    u32 *diag_value,
    u64 reserved)
{
    if(!isAuthorized())
        return IS_USER_NOT_AUTH;

    if (!is_initialized)
        return IS_LIB_NOT_INIT;

    if(mac == NULL || diag_value == NULL)
        return IS_BAD_PARAMETER;

    IS_RESULT is = libhelper.lib->Lock();

    if(is == IS_SUCCESS)
    {
        Adapter *a = libhelper.lib->GetAdapter(mac);
        if(a == NULL)
        {
            is = IS_ADAPTER_NOT_FOUND;
        }
        else
        {
            is = a->getDiagnostics(diag_value);
        }

        //unlock failures take precendence
        IS_RESULT unlockresult = libhelper.lib->Unlock();
        if(unlockresult != IS_SUCCESS)
            is = unlockresult;
    }

    return is;
} //GetAdapterDiagnostics()

IS_RESULT GetQueryTargetCount (
    const IS_MAC_ADDRESS *mac,
    u32 ip,
    u16 port,
    const PIS_CHAP_SETTINGS pchap_settings,
    u32 *target_count,
    u64 reserved)
{
    if(!isAuthorized())
        return IS_USER_NOT_AUTH;

    if (!is_initialized)
        return IS_LIB_NOT_INIT;

    if(mac == NULL || target_count == NULL || pchap_settings == NULL)
        return IS_BAD_PARAMETER;

    IS_RESULT is = libhelper.lib->Lock();

    if(is == IS_SUCCESS)
    {
        Adapter *a = libhelper.lib->GetAdapter(mac);
        if(a == NULL)
        {
            is = IS_ADAPTER_NOT_FOUND;
        }
        else
        {
            is = a->getQueryTargetCount(ip,port,pchap_settings, target_count);
        }

        //unlock failures take precendence
        IS_RESULT unlockresult = libhelper.lib->Unlock();
        if(unlockresult != IS_SUCCESS)
            is = unlockresult;
    }

    return is;
} //GetQueryTargetCount()

IS_RESULT EnumQueryTargets (
    const IS_MAC_ADDRESS *mac,
    u32 ip,
    u16 port,
    const PIS_CHAP_SETTINGS pchap_settings,
    u32 index,
    char *targetname,
    u32 targetnameBufferLength,
    IS_TARGET_ADDRESS *addresses,
    u32 *numaddresses, u64 reserved
)
{
    if(!isAuthorized())
        return IS_USER_NOT_AUTH;

    if (!is_initialized)
        return IS_LIB_NOT_INIT;

    if(mac == NULL ||
       pchap_settings == NULL ||
       targetname == NULL ||
       addresses == NULL ||
       numaddresses == NULL)
        return IS_BAD_PARAMETER;

    IS_RESULT is = libhelper.lib->Lock();

    if(is == IS_SUCCESS)
    {
        Adapter *a = libhelper.lib->GetAdapter(mac);
        if(a == NULL)
        {
            is = IS_ADAPTER_NOT_FOUND;
        }
        else
        {
            is = a->enumQueryTargets(ip,
                port,
                pchap_settings,
                index,
                targetname,
                targetnameBufferLength,
                addresses,
                numaddresses);
        }

        //unlock failures take precendence
        IS_RESULT unlockresult = libhelper.lib->Unlock();
        if(unlockresult != IS_SUCCESS)
            is = unlockresult;
    }

    return is;
} //EnumQueryTargets()

IS_RESULT AddTarget (
    const IS_MAC_ADDRESS *mac,
    u32 ip,
    u16 port,
    const PIS_CHAP_SETTINGS pchap_settings,
    const char *nodename,
    IS_CRC_MODE crcmode,
    u64 reserved
)
{
    if(!isAuthorized())
        return IS_USER_NOT_AUTH;

    if (!is_initialized)
        return IS_LIB_NOT_INIT;

    if(mac == NULL || pchap_settings == NULL || nodename == NULL)
        return IS_BAD_PARAMETER;

    if(strlen(nodename) == 0 || strlen(nodename) >= 256)
        return IS_BAD_PARAMETER;

    if(crcmode != CRC_MODE_NONE &&
        crcmode != CRC_MODE_HEADER &&
        crcmode != CRC_MODE_DATA &&
        crcmode != CRC_MODE_BOTH)
        return IS_BAD_PARAMETER;

    if(pchap_settings->useCHAP)
    {
        if(strlen(pchap_settings->CHAPusername) == 0 ||
            strlen(pchap_settings->CHAPpassword) == 0)
        {
            return IS_BAD_PARAMETER;
        }

        if(pchap_settings->useInboundCHAP)
        {
            if(strlen(pchap_settings->inboundCHAPusername) == 0 ||
                strlen(pchap_settings->inboundCHAPpassword) == 0)
            {
                return IS_BAD_PARAMETER;
            }

            if(!strcmp(pchap_settings->inboundCHAPpassword,
                       pchap_settings->CHAPpassword))
            {
                return IS_BAD_PARAMETER;
            }
        }
    }
    else if(pchap_settings->useInboundCHAP)
    {
        // cant have inbound set without outbound
        return IS_BAD_PARAMETER;
    }

    IS_RESULT is = libhelper.lib->Lock();

    if(is == IS_SUCCESS)
    {
        Adapter *a = libhelper.lib->GetAdapter(mac);
        if(a == NULL)
        {
            is = IS_ADAPTER_NOT_FOUND;
        }
        else
        {
            is = a->addTarget(ip, port, pchap_settings, nodename, crcmode);
            libhelper.lib->is_modified = true;
        }

        //unlock failures take precendence
        IS_RESULT unlockresult = libhelper.lib->Unlock();
        if(unlockresult != IS_SUCCESS)
            is = unlockresult;
    }

    return is;
} //AddTarget()

IS_RESULT RemoveTarget (
    const IS_MAC_ADDRESS *mac,
    u32 ip,
    u16 port,
    const char *nodename,
    u64 reserved
)
{
    if(!isAuthorized())
        return IS_USER_NOT_AUTH;

    if (!is_initialized)
        return IS_LIB_NOT_INIT;

    if(mac == NULL || nodename == NULL)
        return IS_BAD_PARAMETER;

    IS_RESULT is = libhelper.lib->Lock();

    if(is == IS_SUCCESS)
    {
        Adapter *a = libhelper.lib->GetAdapter(mac);
        if(a == NULL)
        {
            is = IS_ADAPTER_NOT_FOUND;
        }
        else
        {
            is = a->removeTarget(ip,port,nodename);
            libhelper.lib->is_modified = true;
        }

        //unlock failures take precendence
        IS_RESULT unlockresult = libhelper.lib->Unlock();
        if(unlockresult != IS_SUCCESS)
            is = unlockresult;
    }

    return is;
} //RemoveTarget()

IS_RESULT GetTargetsCount (
    const IS_MAC_ADDRESS *mac,
    u32 *target_count,
    u64 reserved
)
{
    if(!isAuthorized())
        return IS_USER_NOT_AUTH;

    if (!is_initialized)
        return IS_LIB_NOT_INIT;

    if(mac == NULL || target_count == NULL)
        return IS_BAD_PARAMETER;

    IS_RESULT is = libhelper.lib->Lock();

    if(is == IS_SUCCESS)
    {
        Adapter *a = libhelper.lib->GetAdapter(mac);
        if(a == NULL)
        {
            is = IS_ADAPTER_NOT_FOUND;
        }
        else
        {
            is = a->getTargetCount(target_count);
        }

        //unlock failures take precendence
        IS_RESULT unlockresult = libhelper.lib->Unlock();
        if(unlockresult != IS_SUCCESS)
            is = unlockresult;
    }

    return is;
} //GetTargetsCount()

IS_RESULT EnumTargets (
    const IS_MAC_ADDRESS *mac,
    u32 index,
    PIS_STORAGE_TARGET st,
    u64 reserved
)
{
    if(!isAuthorized())
        return IS_USER_NOT_AUTH;

    if (!is_initialized)
        return IS_LIB_NOT_INIT;

    if(mac == NULL || st == NULL)
        return IS_BAD_PARAMETER;

    IS_RESULT is = libhelper.lib->Lock();

    if(is == IS_SUCCESS)
    {
        Adapter *a = libhelper.lib->GetAdapter(mac);
        if(a == NULL)
        {
            is = IS_ADAPTER_NOT_FOUND;
        }
        else
        {
            is = a->getTarget(index, st);
        }

        //unlock failures take precendence
        IS_RESULT unlockresult = libhelper.lib->Unlock();
        if(unlockresult != IS_SUCCESS)
            is = unlockresult;
    }

    return is;
} //EnumTargets()

IS_RESULT GetTargetsWithLunsCount (
    const IS_MAC_ADDRESS *mac,
    u32 *target_count,
    u64 reserved
)
{
    if(!isAuthorized())
        return IS_USER_NOT_AUTH;

    if (!is_initialized)
        return IS_LIB_NOT_INIT;

    if(mac == NULL || target_count == NULL)
        return IS_BAD_PARAMETER;

    IS_RESULT is = libhelper.lib->Lock();

    if(is == IS_SUCCESS)
    {
        Adapter *a = libhelper.lib->GetAdapter(mac);
        if(a == NULL)
        {
            is = IS_ADAPTER_NOT_FOUND;
        }
        else
        {
            is = a->getTargetsWithLunsCount(target_count);
        }

        //unlock failures take precendence
        IS_RESULT unlockresult = libhelper.lib->Unlock();
        if(unlockresult != IS_SUCCESS)
            is = unlockresult;
    }

    return is;
} //GetTargetsWithLunsCount()

IS_RESULT EnumTargetsWithLuns (
    const IS_MAC_ADDRESS *mac,
    u32 index,
    PIS_STORAGE_TARGET st,
    IS_SCSI_ID *devices,
    u32 *numdevices,
    u64 reserved
)
{
    if(!isAuthorized())
        return IS_USER_NOT_AUTH;

    if (!is_initialized)
        return IS_LIB_NOT_INIT;

    if(mac == NULL ||
       st == NULL ||
       devices == NULL ||
       numdevices == NULL ||
       *numdevices == 0)
        return IS_BAD_PARAMETER;

    IS_RESULT is = libhelper.lib->Lock();

    if(is == IS_SUCCESS)
    {
        Adapter *a = libhelper.lib->GetAdapter(mac);
        if(a == NULL)
        {
            is = IS_ADAPTER_NOT_FOUND;
        }
        else
        {
            is = a->enumTargetsWithLuns(index, st, devices, numdevices);
        }

        //unlock failures take precendence
        IS_RESULT unlockresult = libhelper.lib->Unlock();
        if(unlockresult != IS_SUCCESS)
            is = unlockresult;
    }

    return is;
} //EnumTargetsWithLuns()

IS_RESULT SetTargetParameters (
    const IS_MAC_ADDRESS *mac,
    u32 ip,
    u16 port,
    const char *nodename,
    u32 timeoutSeconds,
    u32 retries,
    u64 reserved
)
{
    if(!isAuthorized())
        return IS_USER_NOT_AUTH;

    if (!is_initialized)
        return IS_LIB_NOT_INIT;

    if(mac == NULL || nodename == NULL)
        return IS_BAD_PARAMETER;

    if(timeoutSeconds > IS_MAX_TARGET_TIMEOUT_SECONDS || timeoutSeconds == 0)
        return IS_BAD_PARAMETER;

    IS_RESULT is = libhelper.lib->Lock();

    if(is == IS_SUCCESS)
    {
        Adapter *a = libhelper.lib->GetAdapter(mac);
        if(a == NULL)
        {
            is = IS_ADAPTER_NOT_FOUND;
        }
        else
        {
            is = a->setTargetParameters(ip,
                                        port,
                                        nodename,
                                        timeoutSeconds,
                                        retries);
            libhelper.lib->is_modified = true;
        }

        //unlock failures take precendence
        IS_RESULT unlockresult = libhelper.lib->Unlock();
        if(unlockresult != IS_SUCCESS)
            is = unlockresult;
    }

    return is;
} //SetTargetParameters()

IS_RESULT SetTargetCRC (
    const IS_MAC_ADDRESS *mac,
    u32 ip,
    u16 port,
    const char *nodename,
    IS_CRC_MODE crcmode,
    u64 reserved
)
{
    if(!isAuthorized())
        return IS_USER_NOT_AUTH;

    if (!is_initialized)
        return IS_LIB_NOT_INIT;

    if(mac == NULL || nodename == NULL)
        return IS_BAD_PARAMETER;

    if(crcmode != CRC_MODE_NONE &&
        crcmode != CRC_MODE_HEADER &&
        crcmode != CRC_MODE_DATA &&
        crcmode != CRC_MODE_BOTH)
    {
        return IS_BAD_PARAMETER;
    }

    IS_RESULT is = libhelper.lib->Lock();

    if(is == IS_SUCCESS)
    {
        Adapter *a = libhelper.lib->GetAdapter(mac);
        if(a == NULL)
        {
            is = IS_ADAPTER_NOT_FOUND;
        }
        else
        {
            is = a->setTargetCRC(ip,port,nodename,crcmode);
            libhelper.lib->is_modified = true;
        }

        //unlock failures take precendence
        IS_RESULT unlockresult = libhelper.lib->Unlock();
        if(unlockresult != IS_SUCCESS)
            is = unlockresult;
    }

    return is;
} //SetTargetCRC()

IS_RESULT SetTargetSecurity (
    const IS_MAC_ADDRESS *mac,
    u32 ip,
    u16 port,
    const char *nodename,
    const PIS_CHAP_SETTINGS pchap_settings,
    u64 reserved
)
{
    if(!isAuthorized())
        return IS_USER_NOT_AUTH;

    if (!is_initialized)
        return IS_LIB_NOT_INIT;

    if(mac == NULL || nodename == NULL || pchap_settings == NULL)
        return IS_BAD_PARAMETER;

    IS_RESULT is = libhelper.lib->Lock();

    if(is == IS_SUCCESS)
    {
        Adapter *a = libhelper.lib->GetAdapter(mac);
        if(a == NULL)
        {
            is = IS_ADAPTER_NOT_FOUND;
        }
        else
        {
            is = a->setTargetSecurity(ip,port,nodename,pchap_settings);
            libhelper.lib->is_modified = true;
        }

        //unlock failures take precendence
        IS_RESULT unlockresult = libhelper.lib->Unlock();
        if(unlockresult != IS_SUCCESS)
            is = unlockresult;
    }

    return is;
} //SetTargetSecurity()

IS_RESULT ResetTargetConnection (
    const IS_MAC_ADDRESS *mac,
    u32 ip,
    u16 port,
    const char *nodename,
    u64 reserved
)
{
    if(!isAuthorized())
        return IS_USER_NOT_AUTH;

    if (!is_initialized)
        return IS_LIB_NOT_INIT;

    if(mac == NULL || nodename == NULL)
        return IS_BAD_PARAMETER;

    IS_RESULT is = libhelper.lib->Lock();

    if(is == IS_SUCCESS)
    {
        Adapter *a = libhelper.lib->GetAdapter(mac);
        if(a == NULL)
        {
            is = IS_ADAPTER_NOT_FOUND;
        }
        else
        {
            is = a->resetTargetConnection(ip,port,nodename);
        }

        //unlock failures take precendence
        IS_RESULT unlockresult = libhelper.lib->Unlock();
        if(unlockresult != IS_SUCCESS)
            is = unlockresult;
    }

    return is;
} //ResetTargetConnection()

IS_RESULT GetTargetRecoveryStatus (
    const IS_MAC_ADDRESS *mac,
    u32 ip,
    u16 port,
    const char *nodename,
    u32 *timeout,
    u32 *maxretries,
    u32 *currentretries,
    u64 reserved
)
{
    if(!isAuthorized())
        return IS_USER_NOT_AUTH;

    if (!is_initialized)
        return IS_LIB_NOT_INIT;

    if(mac == NULL ||
       nodename == NULL ||
       timeout == NULL ||
       maxretries == NULL ||
       currentretries == NULL)
        return IS_BAD_PARAMETER;

    IS_RESULT is = libhelper.lib->Lock();

    if(is == IS_SUCCESS)
    {
        Adapter *a = libhelper.lib->GetAdapter(mac);
        if(a == NULL)
        {
            is = IS_ADAPTER_NOT_FOUND;
        }
        else
        {
            is = a->getTargetRecoveryStatus(ip,
                                            port,
                                            nodename,
                                            timeout,
                                            maxretries,
                                            currentretries);
        }

        //unlock failures take precendence
        IS_RESULT unlockresult = libhelper.lib->Unlock();
        if(unlockresult != IS_SUCCESS)
            is = unlockresult;
    }

    return is;
} //GetTargetRecoveryStatus()

IS_RESULT GetTargetStatistics (
    const IS_MAC_ADDRESS *mac,
    u32 ip,
    u16 port,
    const char *nodename,
    PIS_SESSION_STATS pstats,
    u64 reserved
)
{
    if(!isAuthorized())
        return IS_USER_NOT_AUTH;

    if (!is_initialized)
        return IS_LIB_NOT_INIT;

    if(mac == NULL || pstats == NULL || nodename == NULL)
        return IS_BAD_PARAMETER;

    IS_RESULT is = libhelper.lib->Lock();

    if(is == IS_SUCCESS)
    {
        Adapter *a = libhelper.lib->GetAdapter(mac);
        if(a == NULL)
        {
            is = IS_ADAPTER_NOT_FOUND;
        }
        else
        {
            is = a->getTargetStatistics(ip,port,nodename,pstats);
        }

        //unlock failures take precendence
        IS_RESULT unlockresult = libhelper.lib->Unlock();
        if(unlockresult != IS_SUCCESS)
            is = unlockresult;
    }

    return is;
} //GetTargetStatistics()

IS_RESULT GenerateUniqueNodename (
    char *name,
    u32 name_size,
    u64 reserved
)
{
    if(!isAuthorized())
        return IS_USER_NOT_AUTH;

    if (!is_initialized)
        return IS_LIB_NOT_INIT;

    if(name == NULL)
        return IS_BAD_PARAMETER;

    //From the linux-iscsi project version 3.2.0.1
    char iname[IS_MAX_NODENAME_STRING];
    struct timeval time;
    struct utsname system_info;
    long hostid = 0;
    struct MD5Context context;
    unsigned char digest[16];
    unsigned char *bytes = digest;
    unsigned char entropy[16];
    int e;
    int fd;
    char *prefix = NULL;

    // initialize
    memset(iname, 0, sizeof(iname));
    memset(digest, 0, sizeof(digest));
    memset(entropy, 0, sizeof(entropy));
    memset(&context, 0, sizeof(context));
    memset(&system_info, 0, sizeof(system_info));
    MD5Init(&context);

    prefix = "iqn.1986-03.com.intel:01";

    // try to feed some entropy from the pool to MD5 in order to get
    // uniqueness properties
    if ((fd = open("/dev/random", O_RDONLY)))
    {
        e = read(fd, &entropy, 16);
        if (e >= 1)
        {
            MD5Update(&context, (md5byte *)entropy, e);
        }

        close(fd);
    }

    // time the name is created is a factor in order to get uniqueness
    // properties
    if (gettimeofday(&time, NULL) < 0)
    {
        perror("error: gettimeofday failed");
        return IS_GENERAL_FAILURE;
    }
    MD5Update(&context, (md5byte *)&time.tv_sec, sizeof(time.tv_sec));
    MD5Update(&context, (md5byte *)&time.tv_usec, sizeof(time.tv_usec));

    // hostid
    hostid = gethostid();
    MD5Update(&context, (md5byte *)&hostid, sizeof(hostid));

    // get the hostname and system name
    if (uname(&system_info) < 0)
    {
        perror("error: uname failed");
        return IS_GENERAL_FAILURE;
    }
    
    MD5Update(&context,
              (md5byte *)system_info.sysname,
              sizeof(system_info.sysname));
              
    MD5Update(&context,
              (md5byte *)system_info.nodename,
              sizeof(system_info.nodename));
              
    MD5Update(&context,
              (md5byte *)system_info.release,
              sizeof(system_info.release));
              
    MD5Update(&context,
              (md5byte *)system_info.version,
              sizeof(system_info.version));
              
    MD5Update(&context,
              (md5byte *)system_info.machine,
              sizeof(system_info.machine));

    // compute the md5 hash of all the bits we just collected
    MD5Final(digest, &context);

    // vary which md5 bytes we pick (though we probably don't need to do this,
    // since hopefully MD5 produces results such that each byte is as good
    // as any other).
    if ((fd = open("/dev/random", O_RDONLY)))
    {
        if (read(fd, entropy, 1) == 1)
        {
            bytes = &digest[(entropy[0] % (sizeof(digest) - 6))];
        }

        close(fd);
    }

    // print the prefix followed by 6 bytes of the MD5 hash
    sprintf(iname, "%s.%x%x%x%x%x%x", prefix,
        bytes[0],  bytes[1],  bytes[2],
        bytes[3],  bytes[4],  bytes[5]);

    iname[sizeof(iname)-1] = '\0';

    if(strlen(iname) >= name_size)
        return IS_BUFFER_TOO_SMALL;

    strcpy(name, iname);

    return IS_SUCCESS;
} //GenerateUniqueNodename()

IS_RESULT SetAdapterBlink (
    const IS_MAC_ADDRESS *mac,
    u32 seconds,
    u64 reserved
)
{
    if(!isAuthorized())
        return IS_USER_NOT_AUTH;

    if (!is_initialized)
        return IS_LIB_NOT_INIT;

    if(mac == NULL)
        return IS_BAD_PARAMETER;

    IS_RESULT is = libhelper.lib->Lock();

    if (is == IS_SUCCESS)
    {
        Adapter *a = libhelper.lib->GetAdapter(mac);
        if(a == NULL)
        {
            is = IS_ADAPTER_NOT_FOUND;
        }
        else
        {

            is = a->blink(seconds);
        }

        //unlock failures take precendence
        IS_RESULT unlockresult = libhelper.lib->Unlock();
        if(unlockresult != IS_SUCCESS)
            is = unlockresult;
    }

    return is;
}

IS_RESULT UpdateFlash (
    const char *filename,
    u64 reserved
)
{
    struct stat statbuf;
    int status;

    if(!isAuthorized())
        return IS_USER_NOT_AUTH;

    if (!is_initialized)
        return IS_LIB_NOT_INIT;

    if(filename == NULL)
        return IS_BAD_PARAMETER;

    if((status = stat(filename, &statbuf)) == -1)
        return IS_BAD_PARAMETER;

    IS_RESULT is = libhelper.lib->Lock();

    if (is == IS_SUCCESS)
    {
        FILE *fp = fopen(filename, "rb");
        if(fp == NULL)
        {
            is = IS_IO_ERROR;
        }
        else
        {
            u32 acount = libhelper.lib->GetAdapterCount();
            for(u32 aidx = 0; aidx < acount; aidx++)
            {
                Adapter *a = libhelper.lib->GetAdapter(aidx);
                if(a == NULL)
                {
                    is = IS_ADAPTER_NOT_FOUND;
                    break;
                }
                else
                {
                    IS_FLASH_UPDATE_PACKET fup;
                    int  flen;

                    cout << "Flashing Adapter #" << aidx;

                    clearerr(fp);
                    fseek(fp, 0L, SEEK_END);
                    flen = ftell(fp);
                    fseek(fp, 0L, SEEK_SET);

                    memset(&fup, 0, sizeof(fup));
                    fup.max_frag = (flen + sizeof(fup.buffer) - 1) / sizeof(fup.buffer);
                    fup.frag_number = 1;
                    fup.total_size = flen;
                    while (!feof(fp) && is == IS_SUCCESS)
                    {
                        size_t siz = fread(fup.buffer, 1, sizeof(fup.buffer), fp);
                        int offset = ftell(fp);
                        fup.frag_size = siz;
                        if (fup.frag_number == 1)
                        {
                            unsigned long *pid = (unsigned long *)&fup.buffer[4];
                            fup.software_id =
                                ((pid[0] & 0xffff) << 16) | ((pid[0] & 0xffff0000) >> 16);
                        }

                        if (offset > flen) offset = flen;
                        cout << ".";
                        cout.flush();
                        is = a->updateFlash(&fup);
                        ++fup.frag_number;
                    }

                    cout << endl;
                } //else
            } // for(u32 aidx = 0; aidx < acount; aidx++)

            fclose(fp);
        }

        //unlock failures take precendence
        IS_RESULT unlockresult = libhelper.lib->Unlock();
        if(unlockresult != IS_SUCCESS)
            is = unlockresult;
    } //if (is == IS_SUCCESS)

    return is;
} //UpdateFlash()

IS_RESULT GetEEPromData (
    const IS_MAC_ADDRESS *mac,
    u8 *pbuffer,
    u32 *pbuffer_size,
    u64 reserved
)
{
    if(!isAuthorized())
        return IS_USER_NOT_AUTH;

    if (!is_initialized)
        return IS_LIB_NOT_INIT;

    if(mac == NULL || pbuffer == NULL || pbuffer_size == NULL)
        return IS_BAD_PARAMETER;

    IS_RESULT is = libhelper.lib->Lock();

    if (is == IS_SUCCESS)
    {
        Adapter *a = libhelper.lib->GetAdapter(mac);
        if(a == NULL)
        {
            is = IS_ADAPTER_NOT_FOUND;
        }
        else
        {
            is = a->getEEPROMData(pbuffer, pbuffer_size);
        }

        //unlock failures take precendence
        IS_RESULT unlockresult = libhelper.lib->Unlock();
        if(unlockresult != IS_SUCCESS)
            is = unlockresult;
    }

    return is;
} //GetEEPromData()

IS_RESULT Ping (
    const IS_MAC_ADDRESS *mac,
    u32 ip,
    u32 *ping_result,
    u64 reserved
)
{
    if(!isAuthorized())
        return IS_USER_NOT_AUTH;

    if (!is_initialized)
        return IS_LIB_NOT_INIT;

    if(mac == NULL || ping_result == NULL)
        return IS_BAD_PARAMETER;

    IS_RESULT is = libhelper.lib->Lock();

    if (is == IS_SUCCESS)
    {
        Adapter *a = libhelper.lib->GetAdapter(mac);
        if(a == NULL)
        {
            is = IS_ADAPTER_NOT_FOUND;
        }
        else
        {

            is = a->ping(ip, ping_result);
        }

        //unlock failures take precendence
        IS_RESULT unlockresult = libhelper.lib->Unlock();
        if(unlockresult != IS_SUCCESS)
            is = unlockresult;
    }

    return is;
} // Ping()

IS_RESULT LoopbackTest (
    const IS_MAC_ADDRESS *mac,
    u8 *pinbuffer,
    u8 *poutbuffer,
    u32 buffsize,
    u64 reserved
)
{
    if(!isAuthorized())
        return IS_USER_NOT_AUTH;

    if (!is_initialized)
        return IS_LIB_NOT_INIT;

    if(mac == NULL || pinbuffer == NULL || poutbuffer == NULL ||
        buffsize > IS_MAX_LOOPBACK_BUFFER_SIZE)
    {
        return IS_BAD_PARAMETER;
    }

    IS_RESULT is = libhelper.lib->Lock();

    if (is == IS_SUCCESS)
    {
        Adapter *a = libhelper.lib->GetAdapter(mac);
        if(a == NULL)
        {
            is = IS_ADAPTER_NOT_FOUND;
        }
        else
        {

            is = a->loopback(pinbuffer, poutbuffer, buffsize);
        }

        //unlock failures take precendence
        IS_RESULT unlockresult = libhelper.lib->Unlock();
        if(unlockresult != IS_SUCCESS)
            is = unlockresult;
    }

    return is;
} //LoopbackTest()

IS_RESULT SetTargetStatus (
    const IS_MAC_ADDRESS *mac,
    u32 ip,
    u16 port,
    const char *nodename,
    TARGET_STATUS newStatus,
    u64 reserved
)
{
    if(!isAuthorized())
        return IS_USER_NOT_AUTH;

    if (!is_initialized)
        return IS_LIB_NOT_INIT;

    if(mac == NULL || nodename == NULL)
        return IS_BAD_PARAMETER;

    IS_RESULT is = libhelper.lib->Lock();

    if (is == IS_SUCCESS)
    {
        Adapter *a = libhelper.lib->GetAdapter(mac);
        if(a == NULL)
        {
            is = IS_ADAPTER_NOT_FOUND;
        }
        else
        {

            is = a->setTargetStatus(ip, port, nodename, newStatus);
            libhelper.lib->is_modified = true;
        }

        //unlock failures take precendence
        IS_RESULT unlockresult = libhelper.lib->Unlock();
        if(unlockresult != IS_SUCCESS)
            is = unlockresult;
    }

    return is;
} //SetTargetStatus()

IS_RESULT TargetPropertiesChanged (
    const IS_MAC_ADDRESS *mac,
    u32 ip,
    u16 port,
    const char *nodename,
    u64 reserved
)
{
    if(!isAuthorized())
        return IS_USER_NOT_AUTH;

    if (!is_initialized)
        return IS_LIB_NOT_INIT;

    if(mac == NULL || nodename == NULL)
        return IS_BAD_PARAMETER;

    IS_RESULT is = libhelper.lib->Lock();

    if (is == IS_SUCCESS)
    {
        Adapter *a = libhelper.lib->GetAdapter(mac);
        if(a == NULL)
        {
            is = IS_ADAPTER_NOT_FOUND;
        }
        else
        {
            is = a->targetPropertiesChanged(ip,port,nodename);
            libhelper.lib->is_modified = true;
        }

        //unlock failures take precendence
        IS_RESULT unlockresult = libhelper.lib->Unlock();
        if(unlockresult != IS_SUCCESS)
            is = unlockresult;
    }

    return is;
}  //TargetPropertiesChanged()

/** ****************************************************************************
 * Procedure:   UninitializeLibrary
 *
 * Description: This function is used only for testing.
 *
 * @return  IS_SUCCESS on success, or other IS_* result codes.
 ******************************************************************************/
IS_RESULT UninitializeLibrary(u64 reserved)
{
    if(!isAuthorized())
        return IS_USER_NOT_AUTH;

    if (!is_initialized)
        return IS_LIB_NOT_INIT;

    delete libhelper.lib;
    libhelper.lib = NULL;

    return IS_SUCCESS;
} // UninitializeLibrary()
