/*******************************************************************************
* Copyright (C) 2002-2003 Intel Corporation                                    *
*                                                                              *
* This program is free software; you can redistribute it and/or                *
* modify it under the terms of the GNU General Public License                  *
* as published by the Free Software Foundation; either version 2               *
* of the License, or (at your option) any later version.                       *
*                                                                              *
* This program is distributed in the hope that it will be useful,              *
* but WITHOUT ANY WARRANTY; without even the implied warranty of               *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                *
* GNU General Public License for more details.                                 *
*                                                                              *
* You should have received a copy of the GNU General Public License            *
* along with this program; if not, write to the Free Software                  *
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA   *
*******************************************************************************/
/*******************************************************************************
*                                                                              *
* File Name:    library.cpp                                                    *
*                                                                              *
* Abstract:     This file manages the configuration file and synchronization   *
*               to it via shared memory and semaphores.  It detects when       *
*               another process (via this library) modifies the configuration  *
*               and reloads from disk to maintain a fresh view.                *
*                                                                              *
* Environment:  This file should only be used by the library internals.        *
*                                                                              *
* Author:       Mikal N. Hart                                                  *
*               Thomas C. Lea                                                  *
*                                                                              *
* Origination:  April 21, 2003                                                 *
*******************************************************************************/

// Include Files
#include <iostream>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <errno.h>
#include <vector>
#include <algorithm>
#include "is1000ioctl.h"
#include "is1000api.h"
#include "library.h"
#include "adapter.h"
#include "router.h"
#include "target.h"

using namespace std;

// Local Defines
#define SHMID 0x142857
#define SEMID 0x714285

// Externs
extern int errno;

// Local Definitions
static int driverfd = -1;

// Local Prototypes
void CloseDriver();

inline bool operator < (const IS_MAC_ADDRESS &mac1, const IS_MAC_ADDRESS &mac2)
{
    return mac1.first < mac2.first || (mac1.first == mac2.first && mac1.second < mac2.second);
}

Library::Library() :
    shared_mod_count(NULL),
    private_mod_count(0),
    shmid(-1),
    semid(-1),
    is_modified(false),
    is_valid(true),
    first_invocation(false),
    pAdapter(NULL),
    pRouter(NULL),
    pTarget(NULL),
    hCtrl(-1),
    targetTimeoutFound(false),
    targetRetriesFound(false),
    pDrvAdapters(NULL)
{
    // populate our shmid data member
    if(!get_shmid())
    {
        syslog(LOG_ERR, "is1000-library: Cannot get shmid");
        is_valid = false;
        return;
    }

    if (!allocate_shared_memory())
    {
        syslog(LOG_ERR, "is1000-library: Cannot allocate shared memory");
        is_valid = false;
        return;
    }

    if (!allocate_semaphore())
    {
        syslog(LOG_ERR, "is1000-library: Cannot allocate semaphore");
        is_valid = false;
        return;
    }

    // Lock the library when reading
    if (!acquire_semaphore())
    {
        syslog(LOG_ERR, "is1000-library: Cannot acquire semaphore");
        is_valid = false;
        return;
    }

    //if the configuration file isnt there, create a basic one.
    struct stat configStat;
    int rc = stat(IS_CONFIGURATION, &configStat);
    if(rc == -1 && errno == ENOENT)
    {
        cout << "Configuration file " << IS_CONFIGURATION <<
            " not found.  Creating base file." << endl;
        GetAdaptersFromDriver();
        Save();

        // set permissions on the new file to 0600
        chmod(IS_CONFIGURATION, S_IRUSR | S_IWUSR);
    }

    // Load the configuration data
    is_valid = Load();

    //we just started, so our mod_count should match the shared one.
    private_mod_count = *shared_mod_count;

    if (!release_semaphore())
    {
        syslog(LOG_ERR, "is1000-library: Cannot release semaphore");
        is_valid = false;
        return;
    }

    ADAPTER_LIST::iterator i;
    for(i = adapters.begin(); i != adapters.end(); i++)
    {
        Adapter *pAdapter = i->second;
        if (pAdapter && !pAdapter->is_valid())
        {
            is_valid = false;
            syslog(LOG_ERR, "is1000-library: Invalid adapter configuration.");
        }
    }
}

Library::~Library()
{
    delete pAdapter;
    delete pRouter;
    delete pTarget;

    // Detach the shared memory
    if (shared_mod_count != NULL)
    {
        // Detach memory
        shmdt(shared_mod_count);

#if 0
        if (delete_shared_stuff)
        {
            // Delete it
            if (shmid != -1)
                shmctl(shmid, IPC_RMID, 0);

            // Delete the semaphore
            if (semid != -1)
            {
                union semun sem_union;
                semctl(semid, 0, IPC_RMID, sem_union);
            }
        }
#endif
    }

    Unload();
}

bool Library::Load()
{
    ifstream in(IS_CONFIGURATION);
    if (!in.is_open())
    {
        syslog(LOG_ERR,
               "is1000-library: Cannot open file %s",
               IS_CONFIGURATION);
               
        return false;
    }

    // Begin parsing loop using these scratch variables
    pAdapter = NULL;
    pRouter = NULL;
    pTarget = NULL;

    while (!in.eof())
    {
        char buf[1024];
        in.getline(buf, sizeof(buf));

        if (!in.good() && !in.eof())
        {
            syslog(LOG_ERR, "is1000-library: Error reading file");
            return false;
        }

        // skip any comment lines
        if (buf[0] == '#')
            continue;

        string key, val;
        parse_line(buf, key, val);

        if (key.empty())
            ;

        // Top level items

        else if (key == GLOBALSETTINGS_STR)
        {
            // validate
            if (!is_global_context())
                return syntax_error(key, val);
        }

        else if (key == ADAPTER_STR)
        {
            if (!end_target_def(pRouter, pTarget) ||
                !end_router_def(pAdapter, pRouter) ||
                !end_adapter_def(pAdapter) ||
                !(pAdapter = new Adapter))
                return syntax_error(key, val);
        }

        else if (key == ROUTER_STR)
        {
            if (!end_target_def(pRouter, pTarget) ||
                !end_router_def(pAdapter, pRouter) ||
                !(pRouter = new Router))
                return syntax_error(key, val);
        }

        else if (key == TARGET_STR)
        {
            if (!end_target_def(pRouter, pTarget) ||
                !(pTarget = new Target))
                return syntax_error(key, val);

            pTarget->st.address.ip = pRouter->ip;
            pTarget->st.address.port = pRouter->port;
            memcpy(&pTarget->parentMAC, &pAdapter->mac, sizeof(pTarget->parentMAC));
        }

        // Global items

        else if (key == VERSION_STR)
        {
            if (!is_global_context() ||
                !parse_hex(val.c_str(), file_version))
                return syntax_error(key, val);
        }

        else if (key == TICFILE_STR)
        {
            if (!is_global_context())
                return syntax_error(key, val);
            else
                ticfile = val;
        }

        // Adapter items

        else if (key == MACADDRESS_STR)
        {
            if (!is_adapter_context() ||
                !parse_mac_address(val.c_str(), pAdapter->mac))
                return syntax_error(key, val);
        }

        else if (key == LAA_ADDRESS_STR)
        {
            if (!is_adapter_context() ||
                !parse_mac_address(val.c_str(), pAdapter->props.LAA_Address))
                return syntax_error(key, val);
        }

        else if (key == IPADDRESS_STR)
        {
            if (!is_adapter_context() ||
                !parse_ip_address(val.c_str(), pAdapter->props.AdapterIPAddress))
                return syntax_error(key, val);
        }

        else if (key == SUBNETMASK_STR)
        {
            if (!is_adapter_context() ||
                !parse_ip_address(val.c_str(), pAdapter->props.AdapterIPMask))
                return syntax_error(key, val);
        }

        else if (key == GATEWAY_STR)
        {
            if (!is_adapter_context() ||
                !parse_ip_address(val.c_str(), pAdapter->props.AdapterGateway))
                return syntax_error(key, val);
        }

        else if (key == DHCP_STR)
        {
            if (!is_adapter_context() ||
                !parse_bool(val.c_str(), pAdapter->props.UseDhcp))
                return syntax_error(key, val);
        }

        else if (key == DEFAULTRETRIES_STR)
        {
            if (!is_adapter_context() ||
                !parse_int(val.c_str(), pAdapter->props.DefaultRetries))
                return syntax_error(key, val);
        }

        else if (key == DEFAULTTIMEOUT_STR)
        {
            if (!is_adapter_context() ||
                !parse_int(val.c_str(), pAdapter->props.DefaultTimeout))
                return syntax_error(key, val);
        }

        else if (key == AUTONEG_STR)
        {
            if (!is_adapter_context() ||
                !parse_bool(val.c_str(), pAdapter->props.AutoNegEnabled))
                return syntax_error(key, val);
        }

        else if (key == FLOWCTRL_STR)
        {
            if (!is_adapter_context() ||
                !parse_bool(val.c_str(), pAdapter->props.FlowControlEnabled))
                return syntax_error(key, val);
        }

        else if (key == DIS10HALF_STR)
        {
            if (!is_adapter_context() ||
                !parse_bool(val.c_str(), pAdapter->props.Disallow10Half))
                return syntax_error(key, val);
        }

        else if (key == DIS10FULL_STR)
        {
            if (!is_adapter_context() ||
                !parse_bool(val.c_str(), pAdapter->props.Disallow10Full))
                return syntax_error(key, val);
        }

        else if (key == DIS100HALF_STR)
        {
            if (!is_adapter_context() ||
                !parse_bool(val.c_str(), pAdapter->props.Disallow100Half))
                return syntax_error(key, val);
        }

        else if (key == DIS100FULL_STR)
        {
            if (!is_adapter_context() ||
                !parse_bool(val.c_str(), pAdapter->props.Disallow100Full))
                return syntax_error(key, val);
        }

        else if (key == DIS1000FULL_STR)
        {
            if (!is_adapter_context() ||
                !parse_bool(val.c_str(), pAdapter->props.Disallow1000Full))
                return syntax_error(key, val);
        }

        else if (key == FORCEDSPEED_STR)
        {
            if (!is_adapter_context() ||
                !parse_int(val.c_str(), pAdapter->props.ForcedLinkSpeed))
                return syntax_error(key, val);
        }

        else if (key == FORCEFULLDUPLEX_STR)
        {
            if (!is_adapter_context() ||
                !parse_bool(val.c_str(), pAdapter->props.ForcedFullDuplex))
                return syntax_error(key, val);
        }

        else if (key == INITIATOR_NAME_STR)
        {
            if (!is_adapter_context() ||
                !parse_string(val.c_str(), pAdapter->props.iscsi_nodename,
                sizeof(pAdapter->props.iscsi_nodename) - 1))
                return syntax_error(key, val);

            strcpy(pAdapter->props.iscsi_nodename, val.c_str());
        }

        else if (key == FRAMESIZE_STR)
        {
            if (!is_adapter_context() ||
                !parse_int(val.c_str(), pAdapter->props.FrameSize) ||
                (pAdapter->props.FrameSize != ISCSI_FRAMESIZE_1500 &&
                pAdapter->props.FrameSize != ISCSI_FRAMESIZE_9000))
                return syntax_error(key, val);
        }

        // Router items

        else if (key == ROUTERIP_STR)
        {
            if (!is_router_context() ||
                !parse_ip_address(val.c_str(), pRouter->ip))
                return syntax_error(key, val);
        }

        else if (key == ROUTERPORT_STR)
        {
            if (!is_router_context() ||
                !parse_int(val.c_str(), pRouter->port))
                return syntax_error(key, val);
        }

        // Target items

        else if (key == TARGETNAME_STR)
        {
            if (!is_target_context() ||
                !parse_string(val.c_str(),
                              pTarget->st.iscsi_nodename,
                              sizeof(pTarget->st.iscsi_nodename) - 1))
                return syntax_error(key, val);
        }

        else if (key == PORTALGROUP_STR)
        {
            if (!is_target_context() ||
                !parse_int(val.c_str(), pTarget->st.address.portalgroup))
                return syntax_error(key, val);
        }

        else if (key == TARGETID_STR)
        {
            if (!is_target_context() ||
                !parse_int(val.c_str(), pTarget->st.targetId))
                return syntax_error(key, val);
        }

        else if (key == TARGETCRC_STR)
        {
            u32 c;
            if (!is_target_context() ||
                !parse_int(val.c_str(), c))
                return syntax_error(key, val);
            else
                pTarget->st.crcMode = (IS_CRC_MODE)c;
        }

        else if (key == TARGET_STATUS_STR)
        {
            u32 s;
            if (!is_target_context() ||
                !parse_int(val.c_str(), s))
                return syntax_error(key, val);
            else
                pTarget->st.status = (TARGET_STATUS)s;
        }

        else if (key == RETRIES_STR)
        {
            if (!is_target_context() ||
                !parse_int(val.c_str(), pTarget->st.retries))
                return syntax_error(key, val);

            //no need to use the adapter default
            targetRetriesFound = true;
        }

        else if (key == TIMEOUT_STR)
        {
            if (!is_target_context() ||
                !parse_int(val.c_str(), pTarget->st.timeout))
                return syntax_error(key, val);

            //no need to use the adapter default
            targetTimeoutFound = true;
        }

        else if (key == KEYVALUE_STR)
        {
            if (!is_target_context())
                return syntax_error(key, val);

            string newkey, newval;
            if (!parse_pair(val.c_str(), newkey, newval))
                return syntax_error(key, val);

            pTarget->kv_pairs.push_back(KEY_VALUE_PAIR(newkey, newval));
        }

        // CHAP items (both Router and Target)

        else if (key == USE_CHAP_STR)
        {
            if (!is_target_context() && !is_router_context())
                return syntax_error(key, val);

            u16 &where = is_target_context() ?
                    pTarget->st.chap.useCHAP : pRouter->discovery_chap.useCHAP;
            if (!parse_bool(val.c_str(), where))
                return syntax_error(key, val);
        }

        else if (key == USE_INBOUND_CHAP_STR)
        {
            if (!is_target_context() && !is_router_context())
                return syntax_error(key, val);

            u16 &where = is_target_context() ?
                    pTarget->st.chap.useInboundCHAP : pRouter->discovery_chap.useInboundCHAP;
            if (!parse_bool(val.c_str(), where))
                return syntax_error(key, val);

        }

        else if (key == USE_CHAP_STR)
        {
            if (!is_target_context() && !is_router_context())
                return syntax_error(key, val);

            u16 &where = is_target_context() ?
                    pTarget->st.chap.useCHAP : pRouter->discovery_chap.useCHAP;
            if (!parse_bool(val.c_str(), where))
                return syntax_error(key, val);
        }

        else if (key == CHAP_USERNAME_STR)
        {
            if (!is_target_context() && !is_router_context())
                return syntax_error(key, val);

            char *where = is_target_context() ?
                    pTarget->st.chap.CHAPusername : pRouter->discovery_chap.CHAPusername;

            if (!parse_string(val.c_str(), where, IS_MAX_USERNAME_STRING - 1))
                return syntax_error(key, val);
        }

        else if (key == INBOUND_CHAP_USER_STR)
        {
            if (!is_target_context() && !is_router_context())
                return syntax_error(key, val);

            char *where = is_target_context() ?
                    pTarget->st.chap.inboundCHAPusername : pRouter->discovery_chap.inboundCHAPusername;

            if (!parse_string(val.c_str(), where, IS_MAX_USERNAME_STRING - 1))
                return syntax_error(key, val);
        }

        else if (key == CHAP_PASSWORD_STR)
        {
            if (!is_target_context() && !is_router_context())
                return syntax_error(key, val);

            char *where = is_target_context() ?
                    pTarget->st.chap.CHAPpassword : pRouter->discovery_chap.CHAPpassword;

            if (!parse_string(val.c_str(), where, IS_MAX_PASSWORD_STRING - 1))
                return syntax_error(key, val);
        }

        else if (key == INBOUND_CHAP_PASS_STR)
        {
            if (!is_target_context() && !is_router_context())
                return syntax_error(key, val);

            char *where = is_target_context() ?
                    pTarget->st.chap.inboundCHAPpassword : pRouter->discovery_chap.inboundCHAPpassword;

            if (!parse_string(val.c_str(), where, IS_MAX_PASSWORD_STRING - 1))
                return syntax_error(key, val);
        }

        else
        {
            return syntax_error(key, val);
        }

    } // while (!in.eof())

    bool validresult = end_target_def(pRouter, pTarget) &&
        end_router_def(pAdapter, pRouter) &&
        end_adapter_def(pAdapter);

    GetAdaptersFromDriver();
    return validresult;
} // Load()

bool Library::Save()
{
    ofstream of(IS_CONFIGURATION);

    if (!of.is_open())
        return false;

    of << "# is1000.conf" << endl;
    of << "# Configuration file for Intel(R) PRO/1000 T IP adapter" << endl;
    of << "# Copyright (C) 2003 Intel Corporation" << endl;
    of << endl;
    of << "# CAUTION: DO NOT edit this file.  This file is generated automatically." << endl;
    of << endl;

    of << "# ---------- Global settings " << endl;
    of << endl;
    of << GLOBALSETTINGS_STR << endl;
    of << VERSION_STR << "=" << hex << IS_VERSION << endl;
    of << TICFILE_STR << "=" << ticfile << endl;
    of << endl;

    of << "# ---------- Adapters: " << dec << static_cast<int>(adapters.size()) << endl;
    ADAPTER_LIST::iterator i;
    for(i = adapters.begin(); i != adapters.end(); i++)
    {
        Adapter *pAdapt = i->second;
        if(pAdapt == NULL) // this shouldnt happen!
            continue;

        of << endl;
        of << ADAPTER_STR     << endl;
        of << INITIATOR_NAME_STR << "=" << pAdapt->props.iscsi_nodename                   << endl;
        of << MACADDRESS_STR  << "=" << render_mac_address(pAdapt->mac)                   << endl;
        of << LAA_ADDRESS_STR << "=" << render_mac_address(pAdapt->props.LAA_Address)     << endl;
        of << IPADDRESS_STR   << "=" << render_ip_address(pAdapt->props.AdapterIPAddress) << endl;
        of << SUBNETMASK_STR  << "=" << render_ip_address(pAdapt->props.AdapterIPMask)    << endl;
        of << GATEWAY_STR     << "=" << render_ip_address(pAdapt->props.AdapterGateway)   << endl;
        of << DHCP_STR        << "=" << render_bool(pAdapt->props.UseDhcp)                << endl;
        of << FRAMESIZE_STR   << "=" << dec << pAdapt->props.FrameSize                    << endl;
        of << DEFAULTRETRIES_STR     << "=" << dec << pAdapt->props.DefaultRetries        << endl;
        of << DEFAULTTIMEOUT_STR     << "=" << dec << pAdapt->props.DefaultTimeout        << endl;
        of << AUTONEG_STR     << "=" << render_bool(pAdapt->props.AutoNegEnabled)         << endl;
        of << FORCEDSPEED_STR << "=" << dec << pAdapt->props.ForcedLinkSpeed              << endl;
        of << FORCEFULLDUPLEX_STR << "=" << render_bool(pAdapt->props.ForcedFullDuplex)   << endl;
        of << FLOWCTRL_STR    << "=" << render_bool(pAdapt->props.FlowControlEnabled)     << endl;
        of << DIS10HALF_STR   << "=" << render_bool(pAdapt->props.Disallow10Half)         << endl;
        of << DIS10FULL_STR   << "=" << render_bool(pAdapt->props.Disallow10Full)         << endl;
        of << DIS100HALF_STR  << "=" << render_bool(pAdapt->props.Disallow100Half)        << endl;
        of << DIS100FULL_STR  << "=" << render_bool(pAdapt->props.Disallow100Full)        << endl;
        of << DIS1000FULL_STR << "=" << render_bool(pAdapt->props.Disallow1000Full)       << endl;
        of << endl;

        of << "# ---------- Routers: " << dec << static_cast<int>(pAdapt->routers.size()) << endl;

        for (ROUTER_LIST::iterator pr = pAdapt->routers.begin(); pr != pAdapt->routers.end(); ++pr)
        {
            Router *pRout = *pr;

            of << endl;
            of << ROUTER_STR << endl;
            of << ROUTERIP_STR    << "=" << render_ip_address(pRout->ip)                          << endl;
            of << ROUTERPORT_STR  << "=" << dec << pRout->port                                    << endl;
            of << USE_CHAP_STR    << "=" << render_bool(pRout->discovery_chap.useCHAP)            << endl;
            of << CHAP_USERNAME_STR << "=" << pRout->discovery_chap.CHAPusername                  << endl;
            //TODO: encrypt
            of << CHAP_PASSWORD_STR << "=" << pRout->discovery_chap.CHAPpassword                  << endl;
            of << USE_INBOUND_CHAP_STR << "=" << render_bool(pRout->discovery_chap.useInboundCHAP)<< endl;
            of << INBOUND_CHAP_USER_STR << "=" << pRout->discovery_chap.inboundCHAPusername       << endl;
            //TODO: encrypt
            of << INBOUND_CHAP_PASS_STR << "=" << pRout->discovery_chap.inboundCHAPpassword       << endl;
            of << endl;

            of << "# ---------- Targets: " << dec << static_cast<int>(pRout->targets.size()) << endl;

            for (TARGET_LIST::iterator pt = pRout->targets.begin(); pt != pRout->targets.end(); ++pt)
            {
                Target *pTarg = *pt;

                of << endl;
                of << TARGET_STR << endl;
                of << TARGETNAME_STR << "=" << pTarg->st.iscsi_nodename                         << endl;
                of << PORTALGROUP_STR << "=" << (unsigned int)pTarg->st.address.portalgroup     << endl;
                if(pTarg->st.targetId != IS_TARGETID_INVALID)
                    of << TARGETID_STR << "=" << (unsigned int)pTarg->st.targetId               << endl;
                of << TARGETCRC_STR << "=" << pTarg->st.crcMode                                 << endl;
                of << TARGET_STATUS_STR << "=" << pTarg->st.status                              << endl;
                of << TIMEOUT_STR << "=" << pTarg->st.timeout                                   << endl;
                of << RETRIES_STR << "=" << pTarg->st.retries                                   << endl;
                of << USE_CHAP_STR    << "=" << render_bool(pTarg->st.chap.useCHAP)             << endl;
                of << CHAP_USERNAME_STR << "=" << pTarg->st.chap.CHAPusername                   << endl;
                //TODO: encrypt
                of << CHAP_PASSWORD_STR << "=" << pTarg->st.chap.CHAPpassword                   << endl;
                of << USE_INBOUND_CHAP_STR << "=" << render_bool(pTarg->st.chap.useInboundCHAP) << endl;
                of << INBOUND_CHAP_USER_STR << "=" << pTarg->st.chap.inboundCHAPusername        << endl;
                //TODO: encrypt
                of << INBOUND_CHAP_PASS_STR << "=" << pTarg->st.chap.inboundCHAPpassword        << endl;

                for (PAIR_LIST::iterator pp = pTarg->kv_pairs.begin(); pp != pTarg->kv_pairs.end(); ++pp)
                {
                    const KEY_VALUE_PAIR &pair = *pp;
                    of << KEYVALUE_STR << "=" << pair.first << "=" << pair.second << endl;
                }
            } //for (TARGET_LIST::iterator pt = pRout->targets.begin(); pt != pRout->targets.end(); ++pt)
        } //for (ROUTER_LIST::iterator pr = pAdapt->routers.begin(); pr != pAdapt->routers.end(); ++pr)
    } //for(i = adapters.begin(); i != adapters.end(); i++)
    return true;
} //Save()

bool Library::GetAdaptersFromDriver()
{
    bool result = true;

    if(pDrvAdapters != NULL)
        delete pDrvAdapters;

    pDrvAdapters = new ISCSI_ENUMERATE_BOARDS;
    if(pDrvAdapters == NULL)
        return IS_OUT_OF_MEMORY;

    memset(pDrvAdapters,0,sizeof(ISCSI_ENUMERATE_BOARDS));

    if(CallDriver(ISCSI_IOCTL_ENUMERATE_ADAPTERS, pDrvAdapters) != IS_SUCCESS)
    {
        delete pDrvAdapters;
        pDrvAdapters = NULL;
        result = false;
    }
    else
    {
        // ok, we have the list of adapters from the driver.  Go through
        //  it and mark the adapters in our list as found, if possible.
        for(unsigned int i = 0; i < pDrvAdapters->nNumberAdapters; i++)
        {
            Adapter *a = GetAdapter(&pDrvAdapters->iSCSIAdapters[i].AdapterMACAddress);
            if(a != NULL)
            {
                a->found_in_driver = true;
                a->setHostNumber(pDrvAdapters->iSCSIAdapters[i].adapterNumber);
            }
            else
            {
                cout << "Unconfigured adapter found.  MAC Address = " <<
                    render_mac_address(pDrvAdapters->iSCSIAdapters[i].AdapterMACAddress) << endl;
                a = new Adapter;
                memcpy(&(a->mac),
                       &(pDrvAdapters->iSCSIAdapters[i].AdapterMACAddress),
                       sizeof(IS_MAC_ADDRESS));
                a->found_in_driver = true;
                adapters[pDrvAdapters->iSCSIAdapters[i].AdapterMACAddress] = a;
                is_modified = true;
            }
        }
    }

    return result;
}  //GetAdaptersFromDriver()

void Library::parse_line (
    char *buf,
    string &key,
    string &val
)
{
    val = "";

    // Remove trailing newline if any
    int len = strlen(buf);
    if (len > 0 && buf[len - 1] == '\n')
        buf[len - 1] = 0;

    // Find '=' sign if any
    char *p = strchr(buf, '=');
    if (p != NULL)
    {
        *p = 0;
        while (*++p == ' ') ;
        val = p;
    }

    for (p = buf; *p; ++p)
        if (islower(*p)) *p = toupper(*p);
    for (p = buf; *p == ' '; ++p);
    key = p;
}  //parse_line()

bool Library::end_target_def (
    Router *&pRouter,
    Target *&pTarget
)
{
    if (pTarget)
    {
        if (!pRouter) return false;

        //if we didnt find a timeout or retries entry, use the default that was
        //assigned to the adapter
        if(!targetTimeoutFound)
            pTarget->st.timeout = pAdapter->props.DefaultTimeout;
        if(!targetRetriesFound)
            pTarget->st.retries = pAdapter->props.DefaultRetries;

        pRouter->targets.push_back(pTarget);

        pTarget = NULL;
        targetTimeoutFound = false;
        targetRetriesFound = false;
    }

    return true;
} //end_target_def()

bool Library::end_router_def (
    Adapter *&pAdapter,
    Router *&pRouter
)
{
    if (pRouter)
    {
        if (!pAdapter) return false;
        pAdapter->routers.push_back(pRouter);
        pRouter = NULL;
    }

    return true;
} //end_router_def()

bool Library::end_adapter_def(Adapter *&pAdapter)
{
    if (pAdapter)
    {
        adapters[pAdapter->mac] = pAdapter;
        pAdapter = NULL;
    }

    return true;
}  //end_adapter_def()

bool Library::parse_mac_address (
    const char *addr,
    IS_MAC_ADDRESS &MacAddr
)
{
    if (strlen(addr) == 0)
    {
        MacAddr.first = MacAddr.second = 0;
        return true;
    }

    u32 m1, m2, m3, m4, m5, m6;
    char dummy[2];

    if (sscanf(addr,
               "%02lx%02lx%02lx%02lx%02lx%02lx%1s",
               &m1,
               &m2,
               &m3,
               &m4,
               &m5,
               &m6,
               dummy) != 6 &&
        sscanf(addr,
               "%02lx-%02lx-%02lx-%02lx-%02lx-%02lx%1s",
               &m1,
               &m2,
               &m3,
               &m4,
               &m5,
               &m6,
               dummy) != 6)
    {
        return false;
    }
    
    MacAddr.first =  (m1 << 0  & 0x000000FF) |
                     (m2 << 8  & 0x0000FF00) |
                     (m3 << 16 & 0x00FF0000) |
                     (m4 << 24 & 0xFF000000);
    MacAddr.second = (m5 << 0  & 0x000000FF) |
                     (m6 << 8  & 0x0000FF00);
    return true;
} //parse_mac_address()

bool Library::parse_ip_address (
    const char *addr,
    IP_TYPE &IPAddr
)
{
    if (strlen(addr) == 0)
        return true;

    u32 b1, b2, b3, b4;
    char dummy[2];
    if (sscanf(addr,
               "%3lu.%3lu.%3lu.%3lu%1s",
               &b1,
               &b2,
               &b3,
               &b4,
               dummy) != 4)
    {
        return false;
    }
    
    if (b1 > 255 || b2 > 255 || b3 > 255 || b4 > 255)
    {
        return false;
    }
    
    IPAddr = (b1 << 24) | (b2 << 16) | (b3 << 8) | b4;
    return true;
} //parse_ip_address()

bool Library::parse_bool (
    const char *str,
    u16 &b
)
{
    if (strcasecmp(str, "y") && strcasecmp(str, "yes") &&
        strcasecmp(str, "n") && strcasecmp(str, "no"))
        return false;

    b = static_cast<u16> (!strcasecmp(str, "y") || !strcasecmp(str, "yes"));
    return true;
} // parse_bool()

bool Library::parse_bool (
    const char *str,
    u32 &b
)
{
    u16 tmp;
    if(!parse_bool(str, tmp))
    {
        return false;
    }
    else
    {
        b = tmp;
        return true;
    }
} //parse_bool()

bool Library::parse_hex (
    const char *str,
    u32 &h
)
{
    return sscanf(str, "%lx", &h) == 1 ||
        sscanf(str, "0x%lx", &h) == 1 ||
        sscanf(str, "0X%lx", &h) == 1;
} // parse_hex()

bool Library::parse_int (
    const char *str,
    u32 &h
)
{
    return sscanf(str, "%ld", &h) == 1;
} //parse_int()

bool Library::parse_int (
    const char *str,
    u8 &h
)
{
    u32 tmp;
    if(sscanf(str, "%ld", &tmp) == 1)
    {
        if(tmp <= 0xff)
        {
            h = (u8)tmp;
            return true;
        }
    }

    return false;
} //parse_int()

bool Library::parse_int (
    const char *str,
    u16 &h
)
{
    u32 tmp;
    if(sscanf(str, "%ld", &tmp) == 1)
    {
        if(tmp <= 0xffff)
        {
            h = (u16)tmp;
            return true;
        }
    }

    return false;
} //parse_int()

bool Library::parse_string (
    const char *str,
    char *buf,
    size_t size
)
{
    if (strlen(str) > size)
        return false;

    strcpy(buf, str);
    return true;
} //parse_string()

bool Library::parse_pair (
    const char *str,
    string &newkey,
    string &newval
)
{
    char *p = strchr(str, '=');
    if (!p)
    {
        return false;
    }
    
    *p = 0;
    newkey = str;
    newval = p+1;
    return true;
} //parse_pair()

string Library::render_mac_address(const IS_MAC_ADDRESS &MacAddr)
{
    if (MacAddr.first == 0 && MacAddr.second == 0)
        return "";

    char buf[13];
    sprintf(buf, "%02lX%02lX%02lX%02lX%02lX%02lX",
        MacAddr.first >> 0  & 0xFF,
        MacAddr.first >> 8  & 0xFF,
        MacAddr.first >> 16 & 0xFF,
        MacAddr.first >> 24 & 0xFF,
        MacAddr.second >> 0 & 0xFF,
        MacAddr.second >> 8 & 0xFF);
    return buf;
} //render_mac_address()

string Library::render_ip_address(const IP_TYPE &ip)
{
    char buf[16];
    sprintf(buf, "%ld.%ld.%ld.%ld",
        ip >> 24 & 0xFF,
        ip >> 16 & 0xFF,
        ip >> 8  & 0xFF,
        ip >> 0  & 0xFF);
    return buf;

}  //render_ip_address()

string Library::render_bool(u32 b)
{
    return b ? "Yes" : "No";
} //render_bool()

void Library::set_bit_flag (
    u32 flag,
    u32 &bits,
    u32 mask
)
{
    if (flag)
    {
        bits |= mask;
    }
    else
    {
        bits &= ~mask;
    }
} //set_bit_flag()

bool Library::allocate_shared_memory(void)
{
    shared_mod_count = static_cast<u32 *>(shmat(shmid, NULL, 0));
    if (shared_mod_count == reinterpret_cast<u32 *>(-1))
    {
        shared_mod_count = NULL;
        return is_valid = false;
    }

    return true;
}  //allocate_shared_memory()

bool Library::allocate_semaphore()
{
    // Allocate semaphore
    semid = semget(static_cast<key_t>(SEMID), 1, 0666 | IPC_CREAT);
    if (semid == -1)
        return is_valid = false;

    // Initialize semaphore first time in
    union semun sem_union;
    sem_union.val = 1;    // initial value = 1;
    if (semctl(semid, 0, SETVAL, sem_union) == -1)
        return is_valid = false;

    return true;
} //allocate_semaphore()

bool Library::acquire_semaphore()
{
    struct sembuf sem_b;
    sem_b.sem_num = 0;
    sem_b.sem_op = -1;                            /* P() */
    sem_b.sem_flg = SEM_UNDO;
    return semop(semid, &sem_b, 1) != -1;
}  //acquire_semaphore()

bool Library::release_semaphore()
{
    struct sembuf sem_b;
    sem_b.sem_num = 0;
    sem_b.sem_op = 1;                             /* V() */
    sem_b.sem_flg = SEM_UNDO;
    return semop(semid, &sem_b, 1) != -1;
}  //release_semaphore()

void Library::Unload()
{
    ADAPTER_LIST::iterator i;
    for(i = adapters.begin(); i != adapters.end(); i++)
        delete i->second;
    adapters.clear();
    if(pDrvAdapters != NULL)
    {
        delete pDrvAdapters;
        pDrvAdapters = NULL;
    }
}  //Unload()

IS_RESULT Library::Lock()
{
    // Lock the library
    if (!acquire_semaphore())
        return IS_SYNC_FAILURE;

    IS_RESULT is = IS_SUCCESS;

    // Check to see if another process modified the library
    if (private_mod_count != *shared_mod_count)
    {
        cout << "Detected changes by another process.  Reloading config" << endl;
        Unload();
        if (is_valid = Load())
            private_mod_count = *shared_mod_count;
        else
            is = IS_CONFIG_ERROR;
    }

    if (is != IS_SUCCESS)
    {
        if (!release_semaphore())
            is = IS_SYNC_FAILURE;
    }

    return is;
} //Lock()

IS_RESULT Library::Unlock()
{
    IS_RESULT is = IS_SUCCESS;

    if (is_valid && is_modified)
    {
        // try to save the new configuration
        if (Save())
        {
            // Increment the mod count to inform other processes
            private_mod_count = ++*shared_mod_count;
        }
        else
        {
            is = IS_IO_ERROR;
        }

        // Now that it's saved, reset modified flag
        is_modified = false;
    }

    // and finally... release the semaphore
    if (!release_semaphore())
        is = IS_SYNC_FAILURE;

    // Close the driver if it was opened
    CloseDriver();

    return is;
} //Unlock()

bool Library::get_shmid()
{
    // Allocate shared memory "modified_count" that will monitor whether changes
    // have been made by other instances of the library
    shmid = shmget(static_cast<key_t>(SHMID),
                   sizeof(u32),
                   0666 | IPC_CREAT | IPC_EXCL);

    if (shmid == -1)
    {
        if(errno == EEXIST)
        {
            first_invocation = false;
            //it already exists.  just attach to it
            shmid = shmget(static_cast<key_t>(SHMID), sizeof(u32), 0666);
            if(shmid == -1)   //this is a real error
            {
                return is_valid = false;
            }

            return true;
        }
        else
        {
            return is_valid = false;
        }
    }
    else
    {
        first_invocation = true;
        return true;
    }
} //get_shmid()

Adapter *Library::GetAdapter(const u32 index)
{
    if(pDrvAdapters == NULL)
    {
        return NULL;
    }

    if(index >= pDrvAdapters->nNumberAdapters)
    {
        return NULL;
    }
    else
    {
        return GetAdapter(&pDrvAdapters->iSCSIAdapters[index].AdapterMACAddress);
    }
    
}  //GetAdapter()

Adapter *Library::GetAdapter(const IS_MAC_ADDRESS *mac)
{
    ADAPTER_LIST::iterator i = adapters.find(*mac);
    if(i == adapters.end()) // it wasnt found
    {
        return NULL;
    }
    else
    {
        return i->second;
    }
} //GetAdapter()

u32 Library::GetAdapterCount(void)
{
    if(pDrvAdapters == NULL)
    {
        return 0;
    }

    return pDrvAdapters->nNumberAdapters;
}  //GetAdapterCount()

static bool compare_hbanums(Adapter *pa1, Adapter *pa2)
{
    int hba1 = pa1->hostNumber;
    int hba2 = pa2->hostNumber;
    return hba1 < hba2;
} //compare_hbanums()

IS_RESULT Library::InitializeDriver(void)
{
    int pollcount = MAX_PER_TARGET_LOGIN_POLL;
    IS_RESULT result = IS_SUCCESS;
    IS_RESULT tmpresult = IS_SUCCESS;
    IS_RESULT syncresult = IS_SUCCESS;
    IS_RESULT pollresult = IS_SUCCESS;
    list<Adapter *> pollList;
    vector<Adapter *> initlist;  //ordered by HBA number

    // for each adapter that we have configuration for, and that was
    // detected by the driver, set its configuration data
    // Note that it is important that we do this in SCSI order from lowest to
    // highest in order to make persistent device mappings work.

    //egads there must be a better way to get these in hba order...
    ADAPTER_LIST::iterator j;
    for(j = adapters.begin(); j != adapters.end(); j++)
    {
        Adapter *pAdapter = j->second;
        initlist.push_back(pAdapter);
    }
    sort(initlist.begin(), initlist.end(), compare_hbanums);

    vector<Adapter *>::iterator i;
    for(i = initlist.begin(); i != initlist.end(); i++)
    {
        Adapter *pAdapter = *i;
        if((tmpresult = pAdapter->syncToDriver()) != IS_SUCCESS)
        {
            cerr << "Unable to configure adapter " <<
                render_mac_address(pAdapter->mac);
            if(tmpresult == IS_DHCP_FAILURE)
                cerr << ":  DHCP Failed.";
            cerr << endl;

            //save this error as our result, but continue to process adapters
            syncresult = tmpresult;
        }
        else
        {
            u32 targetCount;
            if(pAdapter->getTargetCount(&targetCount) == IS_SUCCESS)
            {
                if(targetCount > 0)
                {
                    //since we are initializing, reset all target status
                    pAdapter->clearTargetStatus();

                    //regardless of the result, we are gonna poll everyone
                    //anyway.
                    pAdapter->initiateLogins();
                    pollList.push_back(pAdapter);
                }
            }
        }
    }

    bool alldone = pollList.empty() ? true:false;
    while(!alldone && --pollcount > 0)
    {
        alldone = true;
        sleep(SECONDS_PER_POLL);
        list<Adapter *>::iterator a;
        for(a = pollList.begin(); a != pollList.end(); a++)
        {
            Adapter *pAdapter = *a;

            if(!pAdapter->login_complete)
            {
                alldone = false;
                tmpresult = pAdapter->loginPoll();
                if( tmpresult != IS_LOGIN_IN_PROGRESS &&
                    tmpresult != IS_SUCCESS)
                {
                    //if anyone had a failure, let that be our overall result
                    pollresult = tmpresult;
                }
            }
        }
    }

    for(i = initlist.begin(); i != initlist.end(); i++)
    {
        u32 tcount;
        Adapter *pAdapter = *i;

        //any targets left in a 'login in progress' state need to be moved to a
        //'login failure' state
        if((tmpresult = pAdapter->getTargetCount(&tcount)) != IS_SUCCESS)
        {
            cerr << "Unable to get target count: " << tmpresult << endl;
            //continue anyway
        }
        else
        {
            for(u32 tidx = 0; tidx < tcount; tidx++)
            {
                IS_STORAGE_TARGET tmpTarget;
                if((tmpresult = pAdapter->getTarget(tidx,
                                                    &tmpTarget)) != IS_SUCCESS)
                {
                    cerr << "Unable to get target #" << tidx <<": " <<
                        tmpresult << endl;
                    //continue anyway
                }
                else
                {
                    if(tmpTarget.status == TARGET_STATUS_LOGIN_IN_PROGRESS)
                    {
                        pAdapter->setTargetStatus(tmpTarget.address.ip,
                                                  tmpTarget.address.port,
                                                  tmpTarget.iscsi_nodename,
                                                  TARGET_STATUS_LOGIN_FAILURE);
                    }
                }
            }
        }

        if((tmpresult = pAdapter->mapLUNS()) != IS_SUCCESS)
        {
            cerr << "Unable to map LUNS" << endl;
        }
    }

    //which error will we report?  There is too much stuff going on in this
    //routine for one result code. For now, login failures are lowest priority,
    //adapter misconfigurations are highest followed by lun errors
    if(syncresult != IS_SUCCESS)
        result = syncresult;
    else if(tmpresult != IS_SUCCESS)
        result = tmpresult;
    else
        result = pollresult;

    is_modified = true; // make sure we save the new status

    return result;
} //InitializeDriver()

IS_RESULT Library::GetDriverVersion(u32 *version)
{
    IS_RESULT result = IS_SUCCESS;

    ISCSI_GET_DRIVER_VERSION version_request;
    memset(&version_request,0,sizeof(version_request));

    if((result = CallDriver(ISCSI_IOCTL_GET_DRIVER_VERSION,
                            &version_request)) == IS_SUCCESS)
    {
        *version = version_request.version;
    }

    return result;
} //GetDriverVersion()

void Library::GetTicNumber(char *tic)
{
    if(tic == NULL)
        return;

    ifstream in(ticfile.c_str());
    if (!in.is_open())
    {
        sprintf(tic, "verfile.tic not found");
        return;
    }

    string id;
    string colon;

    in >> id;    //dummy
    in >> colon; //dummy
    in >> tic;

    in.close();
} //GetTicNumber()

bool OpenDriver()
{
    if(driverfd >= 0)
        return true;   //its already open

    string devname = "/dev/";
    devname.append(CHARDEVICENAME);
    driverfd = open(devname.c_str(), O_RDWR);

    return (driverfd >= 0);
} //OpenDriver()

void CloseDriver()
{
    if(driverfd >=0)
        close(driverfd);

    driverfd = -1;
} //CloseDriver()

IS_RESULT CallDriver(int cmd, void *arg)
{
    int rc;

    if(!OpenDriver())
    {
        syslog(LOG_ERR, "is1000-library: Unable to open is1000 device driver.");
        return IS_NO_DRIVER;
    }

    rc = ioctl(driverfd, cmd, arg);
    if(rc != 0)
    {
        syslog(LOG_ERR,
               "is1000-library: is1000 command %d failed.",
               _IOC_NR(cmd));
        return IS_IO_ERROR;
    }

    return IS_SUCCESS;
} //CallDriver()
