/*******************************************************************************
* Copyright (C) 2002-2003 Intel Corporation                                    *
*                                                                              *
* This program is free software; you can redistribute it and/or                *
* modify it under the terms of the GNU General Public License                  *
* as published by the Free Software Foundation; either version 2               *
* of the License, or (at your option) any later version.                       *
*                                                                              *
* This program is distributed in the hope that it will be useful,              *
* but WITHOUT ANY WARRANTY; without even the implied warranty of               *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                *
* GNU General Public License for more details.                                 *
*                                                                              *
* You should have received a copy of the GNU General Public License            *
* along with this program; if not, write to the Free Software                  *
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA   *
*******************************************************************************/
/*******************************************************************************
*                                                                              *
* File Name:    library.h                                                      *
*                                                                              *
* Abstract:     This file is the interface to the module that manages the      *
*               configuration file and synchronization to it.                  *
*                                                                              *
* Environment:  This file should only be used by the library internals.        *
*                                                                              *
* Author:       Mikal N. Hart                                                  *
*               Thomas C. Lea                                                  *
*                                                                              *
* Origination:  April 21, 2003                                                 *
*******************************************************************************/

#ifndef __LIBRARY_H__
#define __LIBRARY_H__

// Include Files
#include <sys/shm.h>  // shared memory header
#include <sys/sem.h>  // semaphore header
#include <syslog.h>   //syslog access

// X/OPEN says we have to define this union ourselves
union semun
{
    int val;
    struct semid_ds *buf;
    unsigned short int *array;
    struct seminfo *__buf;
};

// Global Defines
#if _DEBUG
#define DEBUGPRINT(x) cout << "DBG: " << x << endl;
#else
#define DEBUGPRINT(x)
#endif

#define IS_CONFIGURATION          "/etc/is1000.conf"

// Global items
#define GLOBALSETTINGS_STR        "[GLOBALSETTINGS]"
#define VERSION_STR               "VERSION"
#define TICFILE_STR               "TICFILE"

// Adapter items
#define ADAPTER_STR               "[ADAPTER]"
#define MACADDRESS_STR            "MACADDRESS"
#define LAA_ADDRESS_STR           "LAA"
#define IPADDRESS_STR             "IPADDRESS"
#define SUBNETMASK_STR            "SUBNETMASK"
#define GATEWAY_STR               "GATEWAY"
#define DHCP_STR                  "DHCP"
#define INITIATOR_NAME_STR        "INITIATORNAME"
#define FRAMESIZE_STR             "FRAMESIZE"
#define DEFAULTRETRIES_STR        "DEFAULTRETRIES"
#define DEFAULTTIMEOUT_STR        "DEFAULTTIMEOUT"
#define AUTONEG_STR               "AUTONEG"
#define FLOWCTRL_STR              "FLOWCONTROL"
#define DIS10HALF_STR             "DISALLOW10HALF"
#define DIS10FULL_STR             "DISALLOW10FULL"
#define DIS100HALF_STR            "DISALLOW100HALF"
#define DIS100FULL_STR            "DISALLOW100FULL"
#define DIS1000FULL_STR           "DISALLOW1000FULL"
#define FORCEDSPEED_STR           "FORCEDSPEED"
#define FORCEFULLDUPLEX_STR       "FORCEFULLDUPLEX"

// Router items
#define ROUTER_STR                "[ROUTER]"
#define ROUTERIP_STR              "IP"
#define ROUTERPORT_STR            "PORT"

// Target items
#define TARGET_STR                "[TARGET]"
#define TARGETNAME_STR            "NAME"
#define PORTALGROUP_STR           "PORTALGROUP"
#define TARGETID_STR              "TARGETID"
#define TARGETCRC_STR             "TARGETCRC"
#define TARGET_STATUS_STR         "TARGETSTATUS"
#define KEYVALUE_STR              "KEYVALUE"
#define RETRIES_STR               "RETRIES"
#define TIMEOUT_STR               "TIMEOUT"
#define USE_CHAP_STR              "USECHAP"
#define CHAP_USERNAME_STR         "CHAPUSERNAME"
#define CHAP_PASSWORD_STR         "CHAPPASSWORD"
#define USE_INBOUND_CHAP_STR      "USEINBOUNDCHAP"
#define INBOUND_CHAP_USER_STR     "INBOUNDCHAPUSERNAME"
#define INBOUND_CHAP_PASS_STR     "INBOUNDCHAPPASSWORD"

#include <stdio.h>
#include <map>
#include <string>
#include <fstream>
#include <list>
#include <utility>
#include "is1000ioctl.h"
#include "is1000privateapi.h"

#define MAX_PER_TARGET_LOGIN_POLL 30
#define SECONDS_PER_POLL 1

using namespace std;

// Prototypes
class Target;
class Router;
class Adapter;
IS_RESULT CallDriver(int cmd, void *arg);

// Global typedefs
typedef pair<string, string> KEY_VALUE_PAIR;
typedef map<IS_MAC_ADDRESS, Adapter *> ADAPTER_LIST;
typedef list<Router *> ROUTER_LIST;
typedef list<Target *> TARGET_LIST;
typedef list<KEY_VALUE_PAIR> PAIR_LIST;

class Library
{
    public:

/** ****************************************************************************
 * Procedure:   Constructor
 *
 * Description: This is the constructor for the Library class
 ******************************************************************************/
        Library();

/** ****************************************************************************
 * Procedure:   Destructor
 *
 * Description: This is the constructor for the Library class
 ******************************************************************************/
        ~Library();

/** ****************************************************************************
 * Procedure:   Lock
 *
 * Description: This routine locks the global semaphore to ensure safe
 *              synchronization
 *
 * @return  IS_SUCCESS if successful or other codes on error
 ******************************************************************************/
        IS_RESULT Lock();

/** ****************************************************************************
 * Procedure:   Unlock
 *
 * Description: This routine unlocks the global semaphore and saves any changes
 *              to the configuration file.
 *
 * @return  IS_SUCCESS if successful or other codes on error
 ******************************************************************************/
        IS_RESULT Unlock();

/** ****************************************************************************
 * Procedure:   GetAdapter
 *
 * Description: This routine retrieves the adapter instance matching the
 *              specified index.
 *
 * @param   index the index of the desired adapter instance
 *
 * @return  a pointer to the desired instance, or NULL if not found
 ******************************************************************************/
        Adapter *GetAdapter(const u32 index);

/** ****************************************************************************
 * Procedure:   GetAdapter
 *
 * Description: This routine retrieves the adapter instance matching the
 *              specified MAC address.
 *
 * @param   mac the MAC address of the desired adapter instance
 *
 * @return  a pointer to the desired instance, or NULL if not found
 ******************************************************************************/
        Adapter *GetAdapter(const IS_MAC_ADDRESS *mac);
        
/** ****************************************************************************
 * Procedure:   GetAdapterCount
 *
 * Description: This routine retrieves the number of adapters in the system
 *
 * @return  the number of adapters
 ******************************************************************************/
        u32 GetAdapterCount(void);
        
/** ****************************************************************************
 * Procedure:   InitializeDriver
 *
 * Description: Sends configuration from the config file (default
 *              /etc/is1000.conf) to the device driver.  After adapters and
 *              targets are configured, login will be initiated for each target.
 *              Must be performed exactly once each time the driver is loaded.
 *
 * @return  IS_SUCCESS on success, or other IS_* result codes.
 ******************************************************************************/
        IS_RESULT InitializeDriver();

/** ****************************************************************************
 * Procedure:   GetDriverVersion
 *
 * Description: Retrieves the version of the device driver
 *
 * @param   version on success receives the version of the driver
 *
 * @return  IS_SUCCESS on success, or other IS_* result codes.
 ******************************************************************************/
        IS_RESULT GetDriverVersion(u32 *version);

/** ****************************************************************************
 * Procedure:   GetTicNumber
 *
 * Description: Retrieves the TIC version number from
 *              /usr/share/is1000-2.0/verfile.tic
 *
 * @param   tic receives the TIC version string on success or
 *          "verfile.tic not found" on failure
 ******************************************************************************/
        void GetTicNumber(char * tic);

        //Public data members
        u32 *library_ref_count;
        u32 *shared_mod_count;
        u32 private_mod_count;
        int shmid;    // shared memory id
        int semid;    // semaphore id
        bool is_modified;
        bool is_valid;
        u32 file_version;
        string ticfile;

    private:

/** ****************************************************************************
 * Procedure:   Load
 *
 * Description: Reads and parses the configuration file in order to build
 *              the object model.
 *
 * @return  true on success, false on failure
 ******************************************************************************/
        bool Load();

/** ****************************************************************************
 * Procedure:   Save
 *
 * Description: Writes the object model to the configuration file 
 *
 * @return  true on success, false on failure
 ******************************************************************************/
        bool Save();

/** ****************************************************************************
 * Procedure:   Unload
 *
 * Description: Clears the object model
 ******************************************************************************/
        void Unload();

/** ****************************************************************************
 * Procedure:   GetAdaptersFromDriver
 *
 * Description: Queries the driver for the list of detected adapters.  For
 *              each adapter found, an Adapter object is created.
 *
 * @return  true on success, false on failure
 ******************************************************************************/
        bool GetAdaptersFromDriver();

        //configuration file helpers

        
/** ****************************************************************************
 * Procedure:   parse_line
 *
 * Description: separates a line from the configuration file into a key and
 *              value pair
 *
 * @param   buf the line from the configuration file
 *
 * @param   key the parsed key
 *
 * @param   val the parsed value
 ******************************************************************************/
        void parse_line(char *buf, string &key, string &val);

/** ****************************************************************************
 * Procedure:   parse_mac_address
 *
 * Description: converts a string into a MAC address structure
 *
 * @param   addr the string to be parsed
 *
 * @param   MacAddr the resulting MAC address structure
 *
 * @return  true on success, false on failure
 ******************************************************************************/
        bool parse_mac_address(const char *addr, IS_MAC_ADDRESS &MacAddr);

/** ****************************************************************************
 * Procedure:   parse_ip_address
 *
 * Description: converts a string into an IP address structure
 *
 * @param   addr the string to be parsed
 *
 * @param   IPAddr the resulting IP address structure
 *
 * @return  true on success, false on failure
 ******************************************************************************/
        bool parse_ip_address(const char *addr, IP_TYPE &IPAddr);

/** ****************************************************************************
 * Procedure:   parse_bool
 *
 * Description: converts a string into numerical boolean
 *
 * @param   str the string to be parsed
 *
 * @param   b the resulting boolean value (16 bit unsigned)
 *
 * @return  true on success, false on failure
 ******************************************************************************/
        bool parse_bool(const char *str, u16 &b);

/** ****************************************************************************
 * Procedure:   parse_bool
 *
 * Description: converts a string into numerical boolean
 *
 * @param   str the string to be parsed
 *
 * @param   b the resulting boolean value (32 bit unsigned)
 *
 * @return  true on success, false on failure
 ******************************************************************************/
        bool parse_bool(const char *str, u32 &b);

/** ****************************************************************************
 * Procedure:   parse_hex
 *
 * Description: converts a string representing a hexadecimal number into a
 *              native integer
 *
 * @param   str the string to be parsed
 *
 * @param   h the resulting integer value (32 bit unsigned)
 *
 * @return  true on success, false on failure
 ******************************************************************************/
        bool parse_hex(const char *str, u32 &h);

/** ****************************************************************************
 * Procedure:   parse_int
 *
 * Description: converts a string representing a decimal number into a
 *              native integer
 *
 * @param   str the string to be parsed
 *
 * @param   h the resulting integer value (32 bit unsigned)
 *
 * @return  true on success, false on failure
 ******************************************************************************/
        bool parse_int(const char *str, u32 &h);

/** ****************************************************************************
 * Procedure:   parse_int
 *
 * Description: converts a string representing a decimal number into a
 *              native integer
 *
 * @param   str the string to be parsed
 *
 * @param   h the resulting integer value (16 bit unsigned)
 *
 * @return  true on success, false on failure
 ******************************************************************************/
        bool parse_int(const char *str, u16 &h);

/** ****************************************************************************
 * Procedure:   parse_int
 *
 * Description: converts a string representing a decimal number into a
 *              native integer
 *
 * @param   str the string to be parsed
 *
 * @param   h the resulting integer value (8 bit unsigned)
 *
 * @return  true on success, false on failure
 ******************************************************************************/
        bool parse_int(const char *str, u8 &h);

/** ****************************************************************************
 * Procedure:   parse_string
 *
 * Description: extracts a substring from the input
 *
 * @param   str the string to be parsed
 *
 * @param   buf the buffer to contain the substring
 *
 * @param   size the number of characters to copy from str into buf
 *
 * @return  true on success, false on failure
 ******************************************************************************/
        bool parse_string(const char *str, char *buf, size_t size);
        
/** ****************************************************************************
 * Procedure:   parse_pair
 *
 * Description: extracts a name value pair from an input string
 *
 * @param   str the string to be parsed
 *
 * @param   newkey the parsed key
 *
 * @param   newval the parsed value
 *
 * @return  true on success, false on failure
 ******************************************************************************/
        bool parse_pair(const char *str, string &newkey, string &newval);
        
/** ****************************************************************************
 * Procedure:   render_mac_address
 *
 * Description: converts a MAC address to a string
 *
 * @param   MacAddr the MAC address to convert
 *
 * @return  the converted MAC address string
 ******************************************************************************/
        string render_mac_address(const IS_MAC_ADDRESS &MacAddr);

/** ****************************************************************************
 * Procedure:   render_ip_address
 *
 * Description: converts an IP address to a string
 *
 * @param   ip the IP address to convert
 *
 * @return  the converted IP address string
 ******************************************************************************/
        string render_ip_address(const IP_TYPE &ip);
        
/** ****************************************************************************
 * Procedure:   render_bool
 *
 * Description: converts a boolean to a string
 *
 * @param   b the boolean value to convert
 *
 * @return  the converted boolean string
 ******************************************************************************/
        string render_bool(u32 b);
        
/** ****************************************************************************
 * Procedure:   end_target_def
 *
 * Description: checks to see if the parser is outside of a target definition
 *
 * @param   pRouter the current router object
 *
 * @param   pTarget the current target object
 *
 * @return  true if the parser is outside of a target definition
 ******************************************************************************/
        bool end_target_def(Router *&pRouter, Target *&pTarget);

/** ****************************************************************************
 * Procedure:   end_router_def
 *
 * Description: checks to see if the parser is outside of a router definition
 *
 * @param   pAdapter the current adapter object
 *
 * @param   pRouter the current router object
 *
 * @return  true if the parser is outside of a router definition
 ******************************************************************************/
        bool end_router_def(Adapter *&pAdapter, Router *&pRouter);

/** ****************************************************************************
 * Procedure:   end_adapter_def
 *
 * Description: checks to see if the parser is outside of an adapter definition
 *
 * @param   pAdapter the current adapter object
 *
 * @return  true if the parser is outside of an adapter definition
 ******************************************************************************/
        bool end_adapter_def(Adapter *&pAdapter);

/** ****************************************************************************
 * Procedure:   set_bit_flag
 *
 * Description: manipulates a 32 bit bitfield
 *
 * @param   flag if nonzero it applies the mask to the bits
 *
 * @param   bits the bitfield to manipulate
 *
 * @param   mask the bitmask to apply to bits
 ******************************************************************************/
        void set_bit_flag(u32 flag, u32 &bits, u32 mask);

/** ****************************************************************************
 * Procedure:   is_global_context
 *
 * Description: determines if the current parser context is global
 *
 * @return  true if the parser is in a global context
 ******************************************************************************/
        bool is_global_context() { return !pAdapter && !pRouter && !pTarget; }

/** ****************************************************************************
 * Procedure:   is_adapter_context
 *
 * Description: determines if the current parser context is adapter
 *
 * @return  true if the parser is in an adapter context
 ******************************************************************************/
        bool is_adapter_context() { return pAdapter && !pRouter && !pTarget; }

/** ****************************************************************************
 * Procedure:   is_router_context
 *
 * Description: determines if the current parser context is router
 *
 * @return  true if the parser is in a router context
 ******************************************************************************/
        bool is_router_context() { return pAdapter && pRouter && !pTarget; }

/** ****************************************************************************
 * Procedure:   is_target_context
 *
 * Description: determines if the current parser context is target
 *
 * @return  true if the parser is in a target context
 ******************************************************************************/
        bool is_target_context() { return pAdapter && pRouter && pTarget; }

/** ****************************************************************************
 * Procedure:   syntax_error
 *
 * Description: logs a syntax error to syslog
 *
 * @return  false always
 ******************************************************************************/
        bool syntax_error(const string &key, const string &val)
        {
            string msg = string("is1000-library: Syntax error: ") +
                            key + "=" + val;
            syslog(LOG_ERR, msg.c_str());
            return false;
        }

        //Data Members
        bool first_invocation;

        //Information retrieved from the config file
        ADAPTER_LIST adapters;

        Adapter *pAdapter;
        Router  *pRouter;
        Target  *pTarget;
        int     hCtrl;
        bool        targetTimeoutFound;
        bool        targetRetriesFound;

        //Information retrieved from the driver
        PISCSI_ENUMERATE_BOARDS pDrvAdapters;

/** ****************************************************************************
 * Procedure:   get_shmid
 *
 * Description: get the global shared memory id.  Create it if necessary.
 *
 * @return  true on success, false on error
 ******************************************************************************/
        bool get_shmid();

/** ****************************************************************************
 * Procedure:   allocate_shared_memory
 *
 * Description: allocate the shared memory region
 *
 * @return  true on success, false on error
 ******************************************************************************/
        bool allocate_shared_memory();

/** ****************************************************************************
 * Procedure:   allocate_semaphore
 *
 * Description: allocate the semaphore
 *
 * @return  true on success, false on error
 ******************************************************************************/
        bool allocate_semaphore();

/** ****************************************************************************
 * Procedure:   acquire_semaphore
 *
 * Description: acquire the semaphore (blocking if necessary)
 *
 * @return  true on success, false on error
 ******************************************************************************/
        bool acquire_semaphore();

/** ****************************************************************************
 * Procedure:   release_semaphore
 *
 * Description: release the semaphore
 *
 * @return  true on success, false on error
 ******************************************************************************/
        bool release_semaphore();
};

/** ****************************************************************************
 * this helper class ensures that the Library destructor gets called at process
 * cleanup time. It enables the library to be instantiated at InitializeLibrary
 * time instead of process startup.
 ******************************************************************************/
class LibraryHelper
{
    public:
        LibraryHelper() : lib(NULL) {};
        ~LibraryHelper()
        {
            if(lib)
            {
                delete lib; lib = NULL;
            }
        };

        Library *lib;
};
#endif                                            // __LIBRARY_H__
