/*******************************************************************************
* Copyright (C) 2002-2003 Intel Corporation                                    *
*                                                                              *
* This program is free software; you can redistribute it and/or                *
* modify it under the terms of the GNU General Public License                  *
* as published by the Free Software Foundation; either version 2               *
* of the License, or (at your option) any later version.                       *
*                                                                              *
* This program is distributed in the hope that it will be useful,              *
* but WITHOUT ANY WARRANTY; without even the implied warranty of               *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                *
* GNU General Public License for more details.                                 *
*                                                                              *
* You should have received a copy of the GNU General Public License            *
* along with this program; if not, write to the Free Software                  *
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA   *
*******************************************************************************/
/*******************************************************************************
*                                                                              *
* File Name:    router.cpp                                                     *
*                                                                              *
* Abstract:     This module provides a class that represents an iSCSI router.  *
*               It manages a list of Targets that are all present on the same  *
*               router (ipaddress and port).                                   *
*                                                                              *
* Environment:  This file should only be used by the library internals.        *
*                                                                              *
* Author:       Thomas C. Lea                                                  *
*                                                                              *
* Origination:  April 21, 2003                                                 *
*******************************************************************************/

// Include Files
#include <iostream>
#include <unistd.h>
#include <sys/ioctl.h>
#include "is1000ioctl.h"
#include "router.h"
#include "target.h"
#include "adapter.h"

using namespace std;

Router::Router()
{
    ip = 0;
    port = 0;
    memset(&discovery_chap, 0, sizeof(discovery_chap));
    login_complete = false;
} //Router()

bool Router::is_valid()
{
    if (ip == 0 || port == 0)
        return false;

    if (discovery_chap.useCHAP &&
        (!strlen(discovery_chap.CHAPusername) ||
         !strlen(discovery_chap.CHAPpassword)))
        return false;

    if (discovery_chap.useInboundCHAP &&
        (!strlen(discovery_chap.inboundCHAPusername) ||
         !strlen(discovery_chap.inboundCHAPpassword)))
        return false;

    for (TARGET_LIST::iterator p = targets.begin(); p != targets.end(); ++p)
        if (!(*p)->is_valid())
            return false;

    return true;
} //is_valid()

IS_RESULT Router::syncToDriver()
{
    IS_RESULT result = IS_SUCCESS;

    if(is_valid())
    {
        TARGET_LIST::iterator i;
        for(i = targets.begin(); i != targets.end(); i++)
        {
            Target *pTarget = *i;
            IS_RESULT tmpresult = pTarget->syncToDriver();
            if(tmpresult != IS_SUCCESS)
                result = tmpresult;
        }
    }
    else
    {
        return IS_CONFIG_ERROR;
        login_complete = true;  // no need to go through login process
    }

    return result;
} //syncToDriver()

IS_RESULT Router::initiateLogins()
{
    IS_RESULT result = IS_SUCCESS;

    if(is_valid())
    {
        TARGET_LIST::iterator i;
        for(i = targets.begin(); i != targets.end(); i++)
        {
            Target *pTarget = *i;
            result = pTarget->login();
        }
    }
    else
    {
        return IS_CONFIG_ERROR;
    }

    return result;
} //initiateLogins()

IS_RESULT Router::loginPoll()
{
    int pendingLogins = 0;
    IS_RESULT result = IS_SUCCESS;

    if(is_valid())
    {
        TARGET_LIST::iterator i;
        for(i = targets.begin(); i != targets.end(); i++)
        {
            Target *pTarget = *i;
            result = pTarget->loginPoll();
            if(result == IS_LOGIN_IN_PROGRESS)
                pendingLogins++;
        }
    }
    else
    {
        login_complete = true;
        return IS_CONFIG_ERROR;
    }

    if(pendingLogins == 0)
        login_complete = true;

    return result;
} //loginPoll

IS_RESULT Router::mapLUNS()
{
    IS_RESULT result = IS_SUCCESS;
    TARGET_LIST::iterator i;
    for(i = targets.begin(); i != targets.end(); i++)
    {
        Target *pTarget = *i;
        pTarget->mapLUNS();
    }

    return result;
} //mapLUNS()

IS_RESULT Router::getTarget (
    const char *targetName,
    Target **pTarget
)
{
    if(pTarget == NULL)
        return IS_BAD_PARAMETER;

    *pTarget = NULL;

    TARGET_LIST::iterator i;
    for(i = targets.begin(); i != targets.end(); i++)
    {
        Target *pTmpTarget = *i;
        if(strcmp(pTmpTarget->st.iscsi_nodename, targetName) == 0)
        {
            *pTarget = pTmpTarget;
            return IS_SUCCESS;
        }
    }

    return IS_TARGET_NOT_FOUND;
} //getTarget()

IS_RESULT Router::getTarget (
    u32 target_index,
    PIS_STORAGE_TARGET st
)
{
    u32 count = 0;
    TARGET_LIST::iterator i;
    for(i = targets.begin(); i != targets.end(); i++)
    {
        Target *pTmpTarget = *i;
        if(count++ == target_index)
        {
            memcpy(st, pTmpTarget, sizeof(IS_STORAGE_TARGET));
            return IS_SUCCESS;
        }
    }

    return IS_TARGET_NOT_FOUND;
} //getTarget()

IS_RESULT Router::addTarget (
    const char *targetName,
    PIS_CHAP_SETTINGS pchap_settings,
    u32 timeoutSeconds,
    u32 retries,
    IS_MAC_ADDRESS parentMAC,
    int parentHostNumber,
    IS_CRC_MODE crcmode
)
{
    IS_RESULT result = IS_SUCCESS;

    Target *pTarget = new Target;
    if(pTarget == NULL)
        return IS_OUT_OF_MEMORY;

    pTarget->st.address.ip = ip;
    pTarget->st.address.port = port;
    pTarget->st.crcMode = crcmode;
    memcpy(&pTarget->parentMAC, &parentMAC, sizeof(pTarget->parentMAC));
    pTarget->parentHostNumber = parentHostNumber;
    strncpy(pTarget->st.iscsi_nodename, targetName, IS_MAX_NODENAME_STRING-1);
    memcpy(&(pTarget->st.chap), pchap_settings, sizeof(IS_CHAP_SETTINGS));
    pTarget->st.timeout = timeoutSeconds;
    pTarget->st.retries = retries;
    targets.push_back(pTarget);

    int pollcount = MAX_PER_TARGET_LOGIN_POLL;
    if((result = pTarget->syncToDriver()) != IS_SUCCESS)
    {
        cerr << "Unable to send new target to driver." << endl;
    }
    else
    {
        result = pTarget->login();
        if(result != IS_SUCCESS && result != IS_LOGIN_IN_PROGRESS)
            cerr << "Unable to login new target." << endl;
        else
        {
            while(--pollcount > 0 &&
                  ((result = pTarget->loginPoll()) == IS_LOGIN_IN_PROGRESS))
            {
                sleep(SECONDS_PER_POLL);
            }

            if(result != IS_SUCCESS)
                pTarget->st.status = TARGET_STATUS_LOGIN_FAILURE;
        }
    }

    return result;
} //addTarget()

IS_RESULT Router::removeTarget(const char *targetName)
{
    //cant use getTarget() because we need the iterator for the remove
    TARGET_LIST::iterator i;
    for(i = targets.begin(); i != targets.end(); i++)
    {
        Target *pTmpTarget = *i;
        if(strcmp(pTmpTarget->st.iscsi_nodename, targetName) == 0)
        {
            //Remove the target from the driver,
            //dont worry if it fails (it may not exist)
            pTmpTarget->remove();
            targets.erase(i);
            delete pTmpTarget;

            return IS_SUCCESS;
        }
    }

    return IS_TARGET_NOT_FOUND;
} //removeTarget()

IS_RESULT Router::setTargetParameters (
    const char *targetName,
    u32 timeoutSeconds,
    u32 retries
)
{
    IS_RESULT result = IS_SUCCESS;
    Target *pTmpTarget;
    if(getTarget(targetName, &pTmpTarget) == IS_SUCCESS)
    {
        result = pTmpTarget->setParameters(timeoutSeconds,retries);
    }
    else
    {
        result = IS_TARGET_NOT_FOUND;
    }

    return result;
}

IS_RESULT Router::setTargetSecurity (
    const char *targetName,
    PIS_CHAP_SETTINGS pchap_settings
)
{
    IS_RESULT result = IS_SUCCESS;
    Target *pTmpTarget;
    if(getTarget(targetName, &pTmpTarget) == IS_SUCCESS)
    {
        result = pTmpTarget->setSecurity(pchap_settings);
    }
    else
    {
        result = IS_TARGET_NOT_FOUND;
    }

    return result;
} // setTargetSecurity()

IS_RESULT Router::clearTargetStatus()
{
    IS_RESULT result = IS_SUCCESS;

    TARGET_LIST::iterator i;
    for(i = targets.begin(); i != targets.end(); i++)
    {
        Target *pTarget = *i;
        pTarget->clearStatus();
    }

    return result;
} //clearTargetStatus()

IS_RESULT Router::setTargetStatus (
    const char *targetName,
    TARGET_STATUS newStatus
)
{
    IS_RESULT result = IS_SUCCESS;
    Target *pTarget = NULL;

    if(getTarget(targetName, &pTarget) == IS_SUCCESS)
    {
        if(newStatus == TARGET_STATUS_CONNECTION_LOST)
        {
            pTarget->disable();
        }

        //if we get a connection faulted event followed by a failed event, leave
        //it in the faulted/connection_lost state
        if(newStatus == TARGET_STATUS_RECOVERING &&
            pTarget->st.status == TARGET_STATUS_CONNECTION_LOST)
        {
            //fall through
        }
        else
        {
            pTarget->st.status = newStatus;
        }
    }
    else
    {
        result = IS_TARGET_NOT_FOUND;
    }

    return result;
} //setTargetStatus()

IS_RESULT Router::getTargetStatus (
    const char *targetName,
    TARGET_STATUS *targetStatus
)
{
    IS_RESULT result = IS_SUCCESS;
    Target *pTarget = NULL;

    if(getTarget(targetName, &pTarget) == IS_SUCCESS)
    {
        *targetStatus = pTarget->st.status;
    }
    else
    {
        result = IS_TARGET_NOT_FOUND;
    }

    return result;
} //getTargetStatus()

void Router::setHostNumber(int hn)
{
    TARGET_LIST::iterator i;
    for(i = targets.begin(); i != targets.end(); i++)
    {
        Target *pTarget = *i;
        pTarget->parentHostNumber = hn;
    }
} //setHostNumber()

IS_RESULT Router::targetPropertiesChanged(const char *nodename)
{
    IS_RESULT result = IS_SUCCESS;

    Target *pTarget = NULL;
    if(getTarget(nodename, &pTarget) == IS_SUCCESS)
    {
        result = pTarget->propertiesChanged();
    }
    else
    {
        result = IS_TARGET_NOT_FOUND;
    }

    return result;
} //targetPropertiesChanged()

Target *Router::getTargetByTID(int tid)
{
    if(tid < 0)
        return NULL;

    TARGET_LIST::iterator i;
    for(i = targets.begin(); i != targets.end(); i++)
    {
        Target *pTarget = *i;
        if(pTarget->st.targetId == (u32)tid)
            return pTarget;
    }

    return NULL;
} //getTargetByTID()
