/*******************************************************************************
* Copyright (C) 2002-2003 Intel Corporation                                    *
*                                                                              *
* This program is free software; you can redistribute it and/or                *
* modify it under the terms of the GNU General Public License                  *
* as published by the Free Software Foundation; either version 2               *
* of the License, or (at your option) any later version.                       *
*                                                                              *
* This program is distributed in the hope that it will be useful,              *
* but WITHOUT ANY WARRANTY; without even the implied warranty of               *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                *
* GNU General Public License for more details.                                 *
*                                                                              *
* You should have received a copy of the GNU General Public License            *
* along with this program; if not, write to the Free Software                  *
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA   *
*******************************************************************************/
/*******************************************************************************
*                                                                              *
* File Name:    router.h                                                       *
*                                                                              *
* Abstract:     This file represents the interface to the Router class, which  *
*               manages a list of Targets that share the same ip address and   *
*               port.                                                          *
*                                                                              *
* Environment:  This file should only be used by the library internals.        *
*                                                                              *
* Author:       Thomas C. Lea                                                  *
*                                                                              *
* Origination:  April 21, 2003                                                 *
*******************************************************************************/

#ifndef __ROUTER_H__
#define __ROUTER_H__

#include "is1000common.h"
#include "library.h"

class Router
{
    public:
/** ****************************************************************************
 * Procedure:   Constructor
 *
 * Description: This is the constructor for the Router class
 ******************************************************************************/
        Router();
        
/** ****************************************************************************
 * Procedure:   is_valid
 *
 * Description: This routine checks the validity of the router instance
 *
 * @return  true if this instance is valid, false otherwise
 ******************************************************************************/
        bool is_valid();
        
/** ****************************************************************************
 * Procedure:   syncToDriver
 *
 * Description: This routine causes a heirarchical synchronization of properties
 *              and configuration from this router down to any targets it may
 *              have with the driver and firmware.
 *
 * @return  IS_SUCCESS if successful or other codes on error
 ******************************************************************************/
        IS_RESULT syncToDriver();

/** ****************************************************************************
 * Procedure:   initiateLogins
 *
 * Description: This routine causes a heirarchical target login of all targets
 *              managed by this router.  After this call, loginPoll is used
 *              to check on the progress of the logins.
 *
 * @return  IS_SUCCESS if successful or other codes on error
 ******************************************************************************/
        IS_RESULT initiateLogins();
        
/** ****************************************************************************
 * Procedure:   loginPoll
 *
 * Description: This routine perfoms one complete login poll on all targets
 *
 * @return  IS_SUCCESS if successful or other codes on error
 ******************************************************************************/
        IS_RESULT loginPoll();

/** ****************************************************************************
 * Procedure:   mapLUNS
 *
 * Description: This routine instructs each managed target to map its LUNs
 *
 * @return  IS_SUCCESS if successful or other codes on error
 ******************************************************************************/
        IS_RESULT mapLUNS();
        
/** ****************************************************************************
 * Procedure:   getTarget
 *
 * Description: This routine retrieves a pointer to the specified target
 *              structure
 *
 * @param   targetName the iSCSI nodename of the desired target
 *
 * @param   pTarget a pointer that will be populated with the address to the
 *          corresponding target structure if found.
 *
 * @return  IS_SUCCESS if successful or other codes on error
 ******************************************************************************/
        IS_RESULT getTarget(const char *targetName,
                            Target **pTarget);

/** ****************************************************************************
 * Procedure:   getTarget
 *
 * Description: This routine retrieves the specified target structure
 *
 * @param   index the index of the desired target
 *
 * @param   st a pointer to the structure that is to receive the data
 *
 * @return  IS_SUCCESS if successful or other codes on error
 ******************************************************************************/
        IS_RESULT getTarget(u32 index,
                            PIS_STORAGE_TARGET st);

/** ****************************************************************************
 * Procedure:   addTarget
 *
 * Description: This routine adds a new target
 *
 * @param   targetName the name of the target
 *
 * @param   pchap_settings the new CHAP configuration to use
 *
 * @param   timeoutSeconds the default number of seconds to wait between retry
 *          attempts during recovery
 *
 * @param   retries the default number of times to retry a failed connection
 *          (zero means infinite) before giving up.
 *
 * @param   parentMAC the MAC address of the managing adapter
 *
 * @param   parentHostNumber the SCSI host number of the managing HBA
 *
 * @param   crcmode the CRC mode to use for the target
 *
 * @return  IS_SUCCESS if successful or other codes on error
 ******************************************************************************/
        IS_RESULT addTarget(const char *targetName,
                            PIS_CHAP_SETTINGS pchap_settings,
                            u32 timeoutSeconds,
                            u32 retries,
                            IS_MAC_ADDRESS parentMAC,
                            int parentHostNumber,
                            IS_CRC_MODE crcmode);
                            
/** ****************************************************************************
 * Procedure:   removeTarget
 *
 * Description: This routine removes a target and tears down its connection
 *
 * @param   targetName the iSCSI nodename of the target
 *
 * @return  IS_SUCCESS if successful or other codes on error
 ******************************************************************************/
        IS_RESULT removeTarget(const char *targetName);
        
/** ****************************************************************************
 * Procedure:   setTargetParameters
 *
 * Description: This routine updates a target's timeout and retry values.
 *
 * @param   targetName the iSCSI nodename of the target
 *
 * @param   timeoutSeconds the number of seconds to wait between retries
 *
 * @param   retries the number of times to retry a connection recovery
 *
 * @return  IS_SUCCESS if successful or other codes on error
 ******************************************************************************/
        IS_RESULT setTargetParameters(const char *targetName,
                                      u32 timeoutSeconds,
                                      u32 retries);
                                      
/** ****************************************************************************
 * Procedure:   setTargetSecurity
 *
 * Description: This routine updates a target's CHAP configuration
 *
 * @param   targetName the iSCSI nodename of the target
 *
 * @param   pchap_settings the new CHAP configuration to use
 *
 * @return  IS_SUCCESS if successful or other codes on error
 ******************************************************************************/
        IS_RESULT setTargetSecurity(const char *targetName,
                                    PIS_CHAP_SETTINGS pchap_settings);
        
/** ****************************************************************************
 * Procedure:   getTargetStatus
 *
 * Description: This routine retrieves the status of the specified target
 *
 * @param   targetName the iSCSI nodename of the target
 *
 * @param   targetStatus the status of the target
 *
 * @return  IS_SUCCESS if successful or other codes on error
 ******************************************************************************/
        IS_RESULT getTargetStatus(const char *targetName,
                                  TARGET_STATUS *targetStatus);
                                  
/** ****************************************************************************
 * Procedure:   setTargetStatus
 *
 * Description: This routine sets the status of the specified target
 *
 * @param   targetName the iSCSI nodename of the target
 *
 * @param   newStatus the status for the target
 *
 * @return  IS_SUCCESS if successful or other codes on error
 ******************************************************************************/
        IS_RESULT setTargetStatus(const char *targetName,
                                  TARGET_STATUS newStatus);
                                  
/** ****************************************************************************
 * Procedure:   targetPropertiesChanged
 *
 * Description: This routine notifies us that the properties have changed.
 *
 * @param   targetName the iSCSI nodename of the target
 *
 * @return  IS_SUCCESS if successful or other codes on error
 ******************************************************************************/
        IS_RESULT targetPropertiesChanged(const char *targetName);
        
/** ****************************************************************************
 * Procedure:   clearTargetStatus
 *
 * Description: This routine marks all managed targets with an uninitialized
 *              status.  Primarily used during initialization.
 *
 * @return  IS_SUCCESS if successful or other codes on error
 ******************************************************************************/
        IS_RESULT clearTargetStatus();
        
/** ****************************************************************************
 * Procedure:   setHostNumber
 *
 * Description: This routine sets the SCSI HBA number for the managing
 *              adapter.  The HBA number is selected by the kernel.  It needs
 *              to bubble down so that targets can map their LUNs properly
 *
 * @param   hn the SCSI host number
 *
 * @return  IS_SUCCESS if successful or other codes on error
 ******************************************************************************/
        void setHostNumber(int hn);
        
/** ****************************************************************************
 * Procedure:   getTargetByTID
 *
 * Description: This routine returns the target associated with the provided
 *              target id.
 *
 * @param   tid the target id of the desired target
 *
 * @return  IS_SUCCESS if successful or other codes on error
 ******************************************************************************/
        Target *getTargetByTID(int tid);

        
        u32 ip;
        u16 port;
        IS_CHAP_SETTINGS discovery_chap;
        TARGET_LIST targets;
        bool login_complete;
};

#endif  // __ROUTER_H__
