/*******************************************************************************
* Copyright (C) 2002-2003 Intel Corporation                                    *
*                                                                              *
* This program is free software; you can redistribute it and/or                *
* modify it under the terms of the GNU General Public License                  *
* as published by the Free Software Foundation; either version 2               *
* of the License, or (at your option) any later version.                       *
*                                                                              *
* This program is distributed in the hope that it will be useful,              *
* but WITHOUT ANY WARRANTY; without even the implied warranty of               *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                *
* GNU General Public License for more details.                                 *
*                                                                              *
* You should have received a copy of the GNU General Public License            *
* along with this program; if not, write to the Free Software                  *
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA   *
*******************************************************************************/
/*******************************************************************************
*                                                                              *
* File Name:    target.cpp                                                     *
*                                                                              *
* Abstract:     This module provides a class that represents an iSCSI target.  *
*                                                                              *
* Environment:  This file should only be used by the library internals.        *
*                                                                              *
* Author:       Thomas C. Lea                                                  *
*                                                                              *
* Origination:  April 21, 2003                                                 *
*******************************************************************************/

// Include Files
#include <unistd.h>
#include <sys/ioctl.h>
#include <iostream>
#include "is1000api.h"
#include "is1000ioctl.h"
#include "target.h"
#include "adapter.h"

extern LibraryHelper libhelper;

Target::Target()
{
    memset(&st, 0, sizeof(st));
    st.targetId = IS_TARGETID_INVALID;
    memset(&parentMAC, 0, sizeof(parentMAC));
    parentHostNumber = -1;
} //Target()

bool Target::is_valid()
{
    if (st.chap.useCHAP &&
        (strlen(st.chap.CHAPusername) == 0 ||
         strlen(st.chap.CHAPpassword) == 0))
        return false;

    if (st.chap.useCHAP &&
        (strlen(st.chap.CHAPusername) == 0 ||
         strlen(st.chap.CHAPpassword) == 0))
        return false;

    if (st.address.ip == 0 ||
        st.address.port == 0 ||
        strlen(st.iscsi_nodename) == 0)
        return false;

    return true;
} //is_valid()

IS_RESULT Target::syncToDriver()
{
    IS_RESULT result = IS_SUCCESS;

    if(is_valid())
    {
        ISCSI_STORAGE_TARGET_INFO target_request;

        memset(&target_request,
               0,
               sizeof(target_request));
        memcpy(&target_request.MACAddr,
               &parentMAC,
               sizeof(target_request.MACAddr));
        memcpy(&target_request.sTarget,
               &st,
               sizeof(target_request.sTarget));
        if((result = CallDriver(ISCSI_IOCTL_ADD_TARGET,
                                &target_request)) == IS_SUCCESS)
        {
            //we are ready for a login attempt
            st.status = TARGET_STATUS_LOGIN_REQUIRED;
            //TODO: once the driver returns a failure code, check it here.
#if 0
            if(target_request.dwStatus == )
            {
                //invalid.  perhaps we already have a connection to this portal group?
                st.status = TARGET_STATUS_SYNC_REQUIRED;
            }
            else
            {
                                                  //we are ready for a login attempt
                st.status = TARGET_STATUS_LOGIN_REQUIRED;
            }
#endif
        }
    }
    else
    {
        return IS_CONFIG_ERROR;
    }

    return result;
} //syncToDriver()

IS_RESULT Target::login()
{
    IS_RESULT result = IS_SUCCESS;

    if(is_valid())
    {
        ISCSI_STORAGE_TARGET_INFO login_request;

        memset(&login_request,
               0,
               sizeof(login_request));
        memcpy(&login_request.MACAddr,
               &parentMAC,
               sizeof(login_request.MACAddr));
        memcpy(&login_request.sTarget,
               &st,
               sizeof(login_request.sTarget));

        if((result = CallDriver(ISCSI_IOCTL_TARGET_INITIATE_LOGIN,
                                &login_request)) == IS_SUCCESS)
        {
            if(login_request.dwStatus == CONNECTION_OK)
            {
                st.status = TARGET_STATUS_CONNECTED;
                result = propertiesChanged();
            }
            else if(login_request.dwStatus == CONNECTION_PENDING)
            {
                result = IS_LOGIN_IN_PROGRESS;
                st.status = TARGET_STATUS_LOGIN_IN_PROGRESS;
            }
            else if(login_request.dwStatus == CONNECTION_FAILED_LOGIN)
            {
                result = IS_LOGIN_FAILURE;
                st.status = TARGET_STATUS_LOGIN_FAILURE;
            }
            else
            {
                result = IS_GENERAL_FAILURE;
                st.status = TARGET_STATUS_LOGIN_FAILURE;
            }
        }
        else
        {
            //fairly accurate... can we do better?
            st.status = TARGET_STATUS_LOGIN_FAILURE;
        }
    }
    else
    {
        return IS_CONFIG_ERROR;
    }

    return result;
} // login()

IS_RESULT Target::loginPoll()
{
    IS_RESULT result = IS_SUCCESS;

    if(is_valid())
    {
        u32 status;

        if((result = getStatusFromDriver(status)) != IS_SUCCESS)
        {
            return result;
        }
        else
        {
            if(status == CONNECTION_OK)
            {
                st.status = TARGET_STATUS_CONNECTED;
                result = propertiesChanged();
            }
            else if(status == CONNECTION_PENDING)
            {
                result = IS_LOGIN_IN_PROGRESS;
                st.status = TARGET_STATUS_LOGIN_IN_PROGRESS;
            }
            else if(status == CONNECTION_FAILED_LOGIN          ||
                //stack,connection,memory failure
                status == CONNECTION_DISRUPTED             ||
                //initiate login didnt work
                status == CONNECTION_NOT_ENABLED           ||
                status == CONNECTION_NEED_SCSI_ENUMERATION ||
                status == CONNECTION_DISABLED)
            {
                //try to get more details from the driver
                ISCSI_TARGET_ERROR_STATUS error_request;
                memset(&error_request,
                       0,
                       sizeof(error_request));
                memcpy(&error_request.MACAddr,
                       &parentMAC,
                       sizeof(error_request.MACAddr));
                error_request.targetId = st.targetId;
                IS_RESULT errresult;
                if((errresult = CallDriver(ISCSI_IOCTL_TARGET_ERROR_STATUS,
                                           &error_request)) == IS_SUCCESS)
                {
                    switch (error_request.sessionState & 0xffff)
                    {
                        case ISCSI_ISM_ERROR_LOGIN_FAIL:
                            result = IS_LOGIN_FAILURE;
                            st.status = TARGET_STATUS_LOGIN_FAILURE;
                            break;

                        case ISCSI_ISM_ERROR_AUTH_FAIL:
                            result = IS_AUTH_FAILURE;
                            st.status = TARGET_STATUS_LOGIN_FAILURE;
                            break;

                        case ISCSI_ISM_ERROR_CONNECT_FAIL:
                            result = IS_TCP_FAILURE;
                            st.status = TARGET_STATUS_LOGIN_FAILURE;
                            break;

                        case ISCSI_ISM_ERROR_MAX_REDIRECT:
                            result = IS_LOGIN_FAILURE;
                            st.status = TARGET_STATUS_LOGIN_FAILURE;
                            break;

                        case ISCSI_ISM_ERROR_LOGIN_TCP_FAIL:
                            result = IS_TCP_FAILURE;
                            st.status = TARGET_STATUS_LOGIN_FAILURE;
                            break;

                        case ISCSI_ISM_ERROR_LOGIN_BAD_STATE:
                            result = IS_LOGIN_FAILURE;
                            st.status = TARGET_STATUS_LOGIN_FAILURE;
                            break;

                        case ISCSI_ISM_ERROR_LOGIN_BAD_PDU:
                            result = IS_LOGIN_FAILURE;
                            st.status = TARGET_STATUS_LOGIN_FAILURE;
                            break;

                        case ISCSI_ISM_ERROR_LOGIN_BAD_USER:
                            result = IS_LOGIN_FAILURE;
                            st.status = TARGET_STATUS_LOGIN_FAILURE;
                            break;

                        case ISCSI_ISM_ERROR_LOGIN_CANT_AGREE:
                            result = IS_LOGIN_FAILURE;
                            st.status = TARGET_STATUS_LOGIN_FAILURE;
                            break;

                        case ISCSI_ISM_ERROR_LOGIN_BAD_REDIRECT:
                            result = IS_LOGIN_FAILURE;
                            st.status = TARGET_STATUS_LOGIN_FAILURE;
                            break;

                        default:
                            result = IS_GENERAL_FAILURE;
                            st.status = TARGET_STATUS_LOGIN_FAILURE;
                            break;
                    }
                }
                else
                {
                    result = errresult;
                    st.status = TARGET_STATUS_LOGIN_FAILURE;
                }
            }
            else if(status == CONNECTION_NO_MATCH)//target add didnt succeed
            {
                result = IS_GENERAL_FAILURE;
                st.status = TARGET_STATUS_LOGIN_FAILURE;
            }
            else
            {
                result = IS_GENERAL_FAILURE;
                st.status = TARGET_STATUS_LOGIN_FAILURE;
            }

            //since we updated the persistent status field, we need to save
            //to disk
            libhelper.lib->is_modified = true;
        }
    }
    else
    {
        return IS_CONFIG_ERROR;
    }

    return result;
} // loginPoll()

IS_RESULT Target::remove()
{
    IS_RESULT result = IS_SUCCESS;

    if(is_valid())
    {
        ISCSI_STORAGE_TARGET_INFO remove_request;
        memset(&remove_request,
               0,
               sizeof(remove_request));
        memcpy(&remove_request.MACAddr,
               &parentMAC,
               sizeof(remove_request.MACAddr));
        memcpy(&remove_request.sTarget,
               &st,
               sizeof(remove_request.sTarget));

        if((result = CallDriver(ISCSI_IOCTL_REMOVE_TARGET,
                                &remove_request)) == IS_SUCCESS)
            st.status = TARGET_STATUS_SYNC_REQUIRED;

        unmapLUNS();
    }
    else
    {
        return IS_CONFIG_ERROR;
    }

    return result;
} //remove()

IS_RESULT Target::disable()
{
    IS_RESULT result = IS_SUCCESS;

    if(is_valid())
    {
        ISCSI_STORAGE_TARGET_INFO disable_request;
        memset(&disable_request,
               0,
               sizeof(disable_request));
        memcpy(&disable_request.MACAddr,
               &parentMAC,
               sizeof(disable_request.MACAddr));
        memcpy(&disable_request.sTarget,
               &st,
               sizeof(disable_request.sTarget));

        if((result = CallDriver(ISCSI_IOCTL_TARGET_DISABLE,
                                &disable_request)) == IS_SUCCESS)
            st.status = TARGET_STATUS_LOGIN_REQUIRED;
    }
    else
    {
        return IS_CONFIG_ERROR;
    }

    return result;
} //disable()

IS_RESULT Target::getStatusFromDriver(u32 &status)
{
    IS_RESULT result = IS_SUCCESS;

    if(is_valid())
    {
        ISCSI_STORAGE_TARGET_INFO status_request;

        memset(&status_request,
               0,
               sizeof(status_request));
        memcpy(&status_request.MACAddr,
               &parentMAC,
               sizeof(status_request.MACAddr));
        memcpy(&status_request.sTarget,
               &st,
               sizeof(status_request.sTarget));

        if((result = CallDriver(ISCSI_IOCTL_STORAGE_TARGET_STATUS,
                                &status_request)) == IS_SUCCESS)
        {
            status = status_request.dwStatus;

            if(status_request.sTarget.targetId != IS_TARGETID_INVALID &&
                status_request.sTarget.targetId != st.targetId)
            {
                //we got a new targetId
                libhelper.lib->is_modified = true;
                st.targetId = status_request.sTarget.targetId;
            }
        }
    }
    else
    {
        return IS_CONFIG_ERROR;
    }

    return result;
} //getStatusFromDriver()

IS_RESULT Target::setParameters (
    u32 timeoutSeconds,
    u32 retries
)
{
    IS_RESULT result = IS_SUCCESS;

    st.timeout = timeoutSeconds;
    st.retries = retries;

    if(is_valid())
    {
        ISCSI_TARGET_RECOVERY_DATA target_request;
        memset(&target_request, 0, sizeof(target_request));
        memcpy(&target_request.MACAddr, &parentMAC, sizeof(target_request.MACAddr));
        target_request.targetId = st.targetId;
        target_request.dwStatus = ISCSI_RECOVERY_DATA_SET;
        target_request.timeOut = st.timeout;
        target_request.maxRetries = st.retries;

        result = CallDriver(ISCSI_IOCTL_TARGET_RECOVERY_DATA, &target_request);
    }
    else
    {
        return IS_CONFIG_ERROR;
    }

    return result;
} //setParameters()

IS_RESULT Target::setSecurity(PIS_CHAP_SETTINGS pchap_settings)
{
    IS_RESULT result = IS_SUCCESS;

    memcpy(&st.chap, pchap_settings, sizeof(IS_CHAP_SETTINGS));

    if(is_valid())
    {
        ISCSI_STORAGE_TARGET_INFO target_request;

        memset(&target_request,
               0,
               sizeof(target_request));
        memcpy(&target_request.MACAddr,
               &parentMAC,
               sizeof(target_request.MACAddr));
        memcpy(&target_request.sTarget,
               &st,
               sizeof(target_request.sTarget));

        result = CallDriver(ISCSI_IOCTL_SET_TARGET_CHAP, &target_request);
    }
    else
    {
        return IS_CONFIG_ERROR;
    }

    return result;
} //setSecurity()

IS_RESULT Target::resetConnection()
{
    IS_RESULT result = IS_SUCCESS;

    if(is_valid())
    {
        // connected targets can be reset with a RESET ioctl
        if(st.status == TARGET_STATUS_CONNECTED)  
        {
            ISCSI_TARGET_RESET reset_request;

            memset(&reset_request,
                   0,
                   sizeof(reset_request));
            memcpy(&reset_request.MACAddr,
                   &parentMAC,
                   sizeof(reset_request.MACAddr));
            memcpy(reset_request.targetname,
                   st.iscsi_nodename,
                   sizeof(reset_request.targetname));
            reset_request.targetId = st.targetId;

            if((result = CallDriver(ISCSI_IOCTL_RESET_TARGET,
                                    &reset_request)) == IS_SUCCESS)
            {
                //since the reset was issued successfully, go ahead and
                // update our status now
                st.status = TARGET_STATUS_RECOVERING;
            }
        }
        else  //targets in every other state have to be reset the hard way
        {
            if((result = remove()) == IS_SUCCESS)
            {
                if((result = syncToDriver()) == IS_SUCCESS)
                {
                    if((result = login()) == IS_LOGIN_IN_PROGRESS)
                    {
                        u32 pollcount = MAX_PER_TARGET_LOGIN_POLL;
                        while(pollcount-- > 0)
                        {
                            //is this indented enough for you?
                            if((result = loginPoll()) != IS_LOGIN_IN_PROGRESS)
                                break;
                            sleep(SECONDS_PER_POLL);
                        }

                        //if we still arent connected, set our status to login
                        //failure
                        if(result == IS_LOGIN_IN_PROGRESS)
                        {
                            st.status = TARGET_STATUS_LOGIN_FAILURE;
                            libhelper.lib->is_modified = true;
                        }
                    }
                }
            }
        }
    }
    else
    {
        return IS_CONFIG_ERROR;
    }

    return result;
} //resetConnection()

IS_RESULT Target::getRecoveryStatus (
    u32 *timeout,
    u32 *maxretries,
    u32 *currentretries
)
{
    IS_RESULT result = IS_SUCCESS;

    *timeout = 0;
    *maxretries = 0;
    *currentretries = 0;

    if(is_valid())
    {
        ISCSI_TARGET_RECOVERY_DATA target_request;
        memset(&target_request,
               0,
               sizeof(target_request));
        memcpy(&target_request.MACAddr,
               &parentMAC,
               sizeof(target_request.MACAddr));
        target_request.targetId = st.targetId;
        target_request.dwStatus = ISCSI_RECOVERY_DATA_GET;
        target_request.timeOut = st.timeout;
        target_request.maxRetries = st.retries;

        if((result = CallDriver(ISCSI_IOCTL_TARGET_RECOVERY_DATA,
                                &target_request)) == IS_SUCCESS)
        {
            *timeout = target_request.timeOut;
            *maxretries = target_request.maxRetries;
            *currentretries = target_request.currentRetries;
        }
    }
    else
    {
        return IS_CONFIG_ERROR;
    }

    return result;
} //getRecoveryStatus()

IS_RESULT Target::getStatistics(PIS_SESSION_STATS pstats)
{
    IS_RESULT result = IS_SUCCESS;

    memset(pstats, 0, sizeof(IS_SESSION_STATS));

    if(is_valid())
    {
        ISCSI_GET_TARGET_STATISTICS target_request;
        memset(&target_request,
               0,
               sizeof(target_request));
        memcpy(&target_request.MACAddr,
               &parentMAC,
               sizeof(target_request.MACAddr));
        target_request.targetId = st.targetId;

        if((result = CallDriver(ISCSI_IOCTL_GET_TARGET_STATISTICS,
                                &target_request)) == IS_SUCCESS)
        {
            memcpy(pstats, &target_request.stats, sizeof(IS_SESSION_STATS));
        }
    }
    else
    {
        return IS_CONFIG_ERROR;
    }

    return result;
} // getStatistics()

IS_RESULT Target::clearStatus()
{
    st.status = TARGET_STATUS_SYNC_REQUIRED;
    return IS_SUCCESS;
} //clearStatus()

IS_RESULT Target::mapLUNS()
{
    if(st.status != TARGET_STATUS_CONNECTED)
    {
        //for now just return success.. if there was a login error we would
        //know about it elsewhere
        return IS_SUCCESS;
    }

    for(unsigned int lun = 0; lun < IS_MAX_LUNS; lun++)
    {
        char outbuf[128];
        sprintf(outbuf,
                "echo \"scsi add-single-device %d %d %d %d\" > /proc/scsi/scsi\n",
                parentHostNumber,
                0,
                st.targetId,
                lun);
        system(outbuf);
    }

    return IS_SUCCESS;
} //mapLUNS()

IS_RESULT Target::unmapLUNS()
{
    for(unsigned int lun = 0; lun < IS_MAX_LUNS; lun++)
    {
        char outbuf[128];
        sprintf(outbuf,
                "echo \"scsi remove-single-device %d %d %d %d\" > /proc/scsi/scsi\n",
                parentHostNumber,
                0,
                st.targetId,
                lun);
        system(outbuf);
    }

    return IS_SUCCESS;
} //unmapLUNS()

IS_RESULT Target::propertiesChanged()
{
    IS_RESULT result = IS_SUCCESS;

    if(is_valid())
    {
        ISCSI_STORAGE_TARGET_INFO target_request;

        memset(&target_request,
               0,
               sizeof(target_request));
        memcpy(&target_request.MACAddr,
               &parentMAC,
               sizeof(target_request.MACAddr));
        memcpy(&target_request.sTarget,
               &st,
               sizeof(target_request.sTarget));

        if((result = CallDriver(ISCSI_IOCTL_GET_TARGET_PROPERTIES,
                                &target_request)) == IS_SUCCESS)
        {
            //the only thing that can change in the driver that we would need
            //to save is ip/port information for redirection
            st.address.ip = target_request.sTarget.address.ip;
            st.address.port = target_request.sTarget.address.port;
            st.address.portalgroup = target_request.sTarget.address.portalgroup;

            //since we have updated some fields, we need to persist.
            libhelper.lib->is_modified = true;
        }
    }
    else
    {
        return IS_CONFIG_ERROR;
    }

    return result;
} //propertiesChanged()
