/*******************************************************************************
* Copyright (C) 2002-2003 Intel Corporation                                    *
*                                                                              *
* This program is free software; you can redistribute it and/or                *
* modify it under the terms of the GNU General Public License                  *
* as published by the Free Software Foundation; either version 2               *
* of the License, or (at your option) any later version.                       *
*                                                                              *
* This program is distributed in the hope that it will be useful,              *
* but WITHOUT ANY WARRANTY; without even the implied warranty of               *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                *
* GNU General Public License for more details.                                 *
*                                                                              *
* You should have received a copy of the GNU General Public License            *
* along with this program; if not, write to the Free Software                  *
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA   *
*******************************************************************************/
/*******************************************************************************
*                                                                              *
* File Name:    target.h                                                       *
*                                                                              *
* Abstract:     This file contains the interface to the Target class, which    *
*               represents an iSCSI target.                                    *
*                                                                              *
* Environment:  This file should only be used by the library internals.        *
*                                                                              *
* Author:       Thomas C. Lea                                                  *
*                                                                              *
* Origination:  April 21, 2003                                                 *
*******************************************************************************/

#ifndef __TARGET_H__
#define __TARGET_H__

// Include Files
#include "is1000common.h"
#include "library.h"

class Target
{
    public:
    
/** ****************************************************************************
 * Procedure:   Constructor
 *
 * Description: This is the constructor for the Target class
 ******************************************************************************/
        Target();
        
/** ****************************************************************************
 * Procedure:   is_valid
 *
 * Description: This routine checks the validity of the target instance
 *
 * @return  true if this instance is valid, false otherwise
 ******************************************************************************/
        bool is_valid();
        
/** ****************************************************************************
 * Procedure:   syncToDriver
 *
 * Description: This routine causes a heirarchical synchronization of properties
 *              and configuration from this target to the driver and firmware.
 *
 * @return  IS_SUCCESS if successful or other codes on error
 ******************************************************************************/
        IS_RESULT syncToDriver();

/** ****************************************************************************
 * Procedure:   login
 *
 * Description: This routine causes a  target to initiate a login
 *
 * @return  IS_SUCCESS if successful or other codes on error
 ******************************************************************************/
        IS_RESULT login();

/** ****************************************************************************
 * Procedure:   loginPoll
 *
 * Description: This routine polls the target to check the progress of a login
 *              operation
 *
 * @return  IS_SUCCESS if successful or other codes on error
 ******************************************************************************/
        IS_RESULT loginPoll();

/** ****************************************************************************
 * Procedure:   remove
 *
 * Description: This routine removes the target and tears down its connection
 *
 * @return  IS_SUCCESS if successful or other codes on error
 ******************************************************************************/
        IS_RESULT remove();

/** ****************************************************************************
 * Procedure:   disable
 *
 * Description: This routine disables the target in the driver/firmware without
 *              removing its configuration
 *
 * @return  IS_SUCCESS if successful or other codes on error
 ******************************************************************************/
        IS_RESULT disable();

/** ****************************************************************************
 * Procedure:   getStatusFromDriver
 *
 * Description: This routine retrieves the current status of the target from
 *              the driver
 *
 * @param   status the retrieved target status
 *
 * @return  IS_SUCCESS if successful or other codes on error
 ******************************************************************************/
        IS_RESULT getStatusFromDriver(u32 &status);

/** ****************************************************************************
 * Procedure:   setParameters
 *
 * Description: This routine updates the target's timeout and retry values.
 *
 * @param   timeoutSeconds the number of seconds to wait between retries
 *
 * @param   retries the number of times to retry a connection recovery
 *
 * @return  IS_SUCCESS if successful or other codes on error
 ******************************************************************************/
        IS_RESULT setParameters(u32 timeoutSeconds,
                                u32 retries);

/** ****************************************************************************
 * Procedure:   setSecurity
 *
 * Description: This routine updates the target's CHAP configuration
 *
 * @param   pchap_settings the new CHAP configuration to use
 *
 * @return  IS_SUCCESS if successful or other codes on error
 ******************************************************************************/
        IS_RESULT setSecurity(PIS_CHAP_SETTINGS pchap_settings);

/** ****************************************************************************
 * Procedure:   resetConnection
 *
 * Description: This routine performs a reset of a target, removing it and
 *              re-adding it if necessary.
 *
 * @return  IS_SUCCESS if successful or other codes on error
 ******************************************************************************/
        IS_RESULT resetConnection();

/** ****************************************************************************
 * Procedure: getRecoveryStatus
 *
 * Description: When a target is in recovery (TARGET_STATUS_RECOVERING), this
 *              function provides information about the progress of
 *              reestablishing the connection.
 *
 * @param   timeout the time (in seconds) between retry attempts
 *
 * @param   maxretries the maximum number of retries that will be attempted
 *          (0 for infinite)
 *
 * @param   currentretries the current retry count
 *
 * @return  IS_SUCCESS on success, or other IS_* result codes.
 ******************************************************************************/
        IS_RESULT getRecoveryStatus(u32 *timeout,
                                    u32 *maxretries,
                                    u32 *currentretries);

/** ****************************************************************************
 * Procedure:   getStatistics
 *
 * Description: This routine retrieves the statistics of a target
 *
 * @param   pstats a pointer to a structure that will receive the data
 *
 * @return  IS_SUCCESS if successful or other codes on error
 ******************************************************************************/
        IS_RESULT getStatistics(PIS_SESSION_STATS pstats);

/** ****************************************************************************
 * Procedure:   clearStatus
 *
 * Description: This routine marks the target with an uninitialized
 *              status.  Primarily used during initialization.
 *
 * @return  IS_SUCCESS if successful or other codes on error
 ******************************************************************************/
        IS_RESULT clearStatus();

/** ****************************************************************************
 * Procedure:   mapLUNS
 *
 * Description: This routine instructs the target to map its LUNs to the SCSI
 *              layer.
 *
 * @return  IS_SUCCESS if successful or other codes on error
 ******************************************************************************/
        IS_RESULT mapLUNS();

/** ****************************************************************************
 * Procedure:   unmapLUNS
 *
 * Description: This routine instructs the target to remove its LUNs from the
 *              SCSI layer.
 *
 * @return  IS_SUCCESS if successful or other codes on error
 ******************************************************************************/
        IS_RESULT unmapLUNS();

/** ****************************************************************************
 * Procedure:   propertiesChanged
 *
 * Description: This routine is used by the isdaemon to notify the target that
 *              its properties have changed.
 *
 * @return  IS_SUCCESS if successful or other codes on error
 ******************************************************************************/
        IS_RESULT propertiesChanged();

        
        IS_STORAGE_TARGET st;
        PAIR_LIST kv_pairs;

        //non-persistent
        IS_MAC_ADDRESS  parentMAC;
        int             parentHostNumber;
};

#endif   // __TARGET_H__
